#include "object.h"
#include "calc.h"

#include <oors_internal.h>

#include <list>
#include <iostream>

#include <cstdlib>

/**
 * Debugger for the polish notation pattern matcher.
 * The debugger is used to count the number of attempted and successful matches.
 */
class Debugger : public oors::Debugger<Object, Object>
{
  public:
    Debugger (void)
      : m_matches (0)
      , m_cost (0)
    {
    }

  public:
    inline int matches (void) const { return m_matches; }
    inline int cost (void) const { return m_cost; }

  public:
    void currentRule (const oors::Rule<Object, Object> *)                                                      { ++m_matches; }
    void finishRule (const oors::Rule<Object, Object> *, bool)                                                 {}
    void applyRule (const oors::Rule<Object, Object> *)                                                        {}
    void newAlternative (const oors::Alternative<Object, Object> *, const oors::Alternative<Object, Object> *) {}
    void currentAlternative (const oors::Alternative<Object, Object> *)                                        {}
    void deleteAlternative (const oors::Alternative<Object, Object> *)                                         {}
    void matchObjectPattern (const oors::Alternative<Object, Object> *, unsigned int, const Object *)          {}
    void matchWildcardPattern (const oors::Alternative<Object, Object> *, unsigned int, const Object *)        {}
    void finishWildcardPattern (const oors::Alternative<Object, Object> *, unsigned int)                       {}
    void finishBasicBlock (int cost)                                                                           { m_cost = cost; }

  private:
    /// number of attempted matches
    int m_matches;

    /// number of successful matches
    int m_cost;
};

/**
 * Calc main routine.
 * @param argc Argument count.
 * @param argv Argument list.
 */
int main (int argc, char **argv)
{
  // no expression given
  if (argc < 3)
  {
    std::cout << "Usage: " << argv[0] << " <Polish|ReversePolish> [expression]" << std::endl
              << "The expression must be in polish or reverse polish notation." << std::endl
              << "Use blanks to separate operators (+, -, x, /) and numbers from each other." << std::endl
              << std::endl
              << "Example:" << std::endl
              << "To compute (2+((2*4.5)/0.5))/(3-1.5) in polish notation, call the program as follows:" << std::endl
              << "" << argv[0] << " Polish / + 2 / x 2 4.5 0.5 - 3 1.5" << std::endl;
    return 1;
  }

  // read calculator queue
  std::list<Object *> queue;
  for (int i = 2; i < argc; ++i)
  {
    // get operator or number
    std::string input = argv[i];
    if (!input.length())
      continue;

    // get operator
    switch (input[0])
    {
      case '+': queue.push_back (new Add ()); break;
      case '-': queue.push_back (new Sub ()); break;
      case 'x': queue.push_back (new Mul ()); break;
      case '/': queue.push_back (new Div ()); break;
      default:  queue.push_back (new Number (atof(input.c_str())));
    }
  }

  // process
  try
  {
    Debugger debugger;
    calc::PatternMatcher matcher (argv[1]);
    matcher.setFirstMatch (true);
    matcher.setLocalMinimum (true);
    matcher.setDebugger (&debugger);
    matcher.process (queue);
    if ((queue.size() == 1)  && dynamic_cast<Number *>(queue.front()))
      std::cout << "The result is: " << static_cast<const Number *>(queue.front())->value() << " (matches: " << debugger.matches() << ", cost: " << debugger.cost() << ")" << std::endl;
    else
      std::cerr << "There was an error calculating the result. Please check your expression." << std::endl;
  }

  // something bad has happened
  catch (oors::Exception exception)
  {
    std::cerr << exception.msg() << std::endl;
  }

  // cleanup
  for (std::list<Object *>::iterator i = queue.begin(), end = queue.end(); i != end; ++i)
    delete *i;

  // done
  return 0;
}
