#include <oors_internal.h>

#include <list>
#include <vector>
#include <iostream>

/**
 * Compute the factorial of n.
 * @param n Natural number.
 * @return Factorial of n.
 */
unsigned int fac (unsigned int n)
{
  unsigned int result = 1;
  for (unsigned int i = 1; i <= n; ++i)
    result *= i;
  return result;
}

/**
 * Recursive function that represents an upper bound for the runtime of a match with the worst-case search pattern.
 * @param n Amount of wildcard patterns.
 * @param m Amount of items.
 * @return Upper bound for the number of required steps.
 */
unsigned int p (unsigned int n, unsigned int m)
{
  return ((n > 0) && (m > 0))?(2 + p(n-1, m) + p(n, m-1)):n;
}

/**
 * Other representation of p.
 * @param n Amount of wildcard patterns.
 * @param m Amount of items.
 * @return Upper bound for the number of required steps.
 */
unsigned int P (unsigned int n, unsigned int m)
{
  if (n == 0)
    return 0;

  unsigned int sum = 0;
  for (unsigned int i = 1; i <= m; ++i)
    sum += P(n-1, i);

  return 2*m+n + sum;
}

/**
 * Abstract object.
 */
class Object {};

/**
 * Debugger that counts the number of steps.
 */
class Debugger : public oors::Debugger<Object, Object>
{
  public:
    /**
     * Initialse the debugger.
     */
    Debugger (void)
      : m_steps (0)
    {
    }

  public:
    /**
     * Access the number of steps.
     * @return The number of steps.
     */
    inline unsigned int steps (void) { return m_steps; }

  public:
    void matchWildcardPattern (const oors::Alternative<Object, Object> *, unsigned int, const Object *) { ++m_steps; }
    void finishWildcardPattern (const oors::Alternative<Object, Object> *, unsigned int)                { ++m_steps; }

  private:
    /// counted number of steps
    unsigned int m_steps;
};

/**
 * Rule that contains the worst-case search pattern.
 */
class Rule : public oors::Rule<Object, Object>
{
  public:
    /**
     * Initialise the rule.
     * Because wildcard pattern may match items only if at least one item pattern has already
     * matched an item and wildcard patterns will be stopped, if all items patterns have matched,
     * the worst case pattern must be embedded into a sequence pattern of the form:
     * [item, [*, ..., *], item] or [item, {*, ..., *} item].
     * @param n Amount of wildcard patterns.
     * @param ordered Indicates whether the sequence pattern should be ordered.
     */
    Rule (unsigned int n, bool ordered)
      : oors::Rule<Object, Object> ("Rule", 0, false)
    {
      m_pattern = new oors::SequencePattern<Object> (NULL, 0, 3, true);
      m_pattern->insert (new oors::ItemPattern<Object> (m_pattern, 0), 0);
      oors::SequencePattern<Object> *sequence = new oors::SequencePattern<Object> (m_pattern, 0, n, ordered);
      for (unsigned int i = 1; i <= n; ++i)
        sequence->insert (new oors::WildcardPattern<Object> (sequence, i), i-1);
      m_pattern->insert (sequence, 1);
      m_pattern->insert (new oors::ItemPattern<Object> (m_pattern, n+1), 2);
      m_size = 2+n;
      m_items = 2;
    }

  public:
    /**
     * Unused cost function.
     * @return Always zero.
     */
    int cost (void) const { return 0; }

    /**
     * Unused apply function.
     */
    void apply (const oors::BasePatternMatcher &, std::list<Object *> &, bool) const {}
};

/**
 * Matcher that computes the number of required steps and compares them with the upper bound.
 */
class PatternMatcher : public oors::PatternMatcher<Object, Object>
{
  public:
    /**
     * Initialise the test matcher.
     * @param n The number of wildcard patterns.
     * @param m The number of available items.
     * @param ordered Indicates whether the sequence pattern should be ordered.
     */
    PatternMatcher (unsigned int n, unsigned int m, bool ordered)
      : oors::PatternMatcher<Object, Object> ("")
      , m_rule (n, ordered)
      , m_n (n)
      , m_m (m)
      , m_ordered (ordered)
    {
      m_rules.push_back (&m_rule);
    }

  public:
    /**
     * Process the given basic block.
     * @param objects The objects to process.
     */
    void process (std::list<Object *> &objects)
    {
      // initialise debugger
      Debugger debugger;
      setDebugger (&debugger);

      // initialise match
      oors::Match<Object> match (objects.begin(), objects.end(), objects.size());
      std::vector<oors::Alternative<Object, Object> *> alternativees;

      // apply the rule
      PatternMatcher::match (NULL, match, alternativees);

      // compare the results
      unsigned int steps = debugger.steps(), border = (m_ordered?1:fac(m_n))*P(m_n, m_m);
      std::cout << (m_ordered?"(Ordered)  ":"(Unordered)") << " Steps: "
                << steps << " <= " << border << " = " << ((steps <= border)?"true":"false")
                << " (" << alternativees.size() << " alternatives)" << std::endl;

      // clear alternativees
      for (unsigned int i = 0, size = alternativees.size(); i < size; ++i)
        delete alternativees[i];
    }

  private:
    /// the rule with the evil pattern
    Rule m_rule;

    /// number of wildcard patterns
    unsigned int m_n;

    /// number of items
    unsigned int m_m;

    /// indicates wether the sequence pattern of the rule is ordered
    bool m_ordered;
};

/**
 * Main routine.
 * @param argc Argument count.
 * @param argv Argument list.
 */
int main (int argc, char **argv)
{
  // not enough arguments.
  if (argc < 3)
  {
    std::cout << "Usage: " << argv[0] << " <n> <m>" << std::endl
              << "Parameters:" << std::endl
              << "  n: Number of wildcard patterns." << std::endl
              << "  m: Number of items to match." << std::endl
              << std::endl
              << "Determines the number of required steps to match [*, ..., *] and {*, ..., *}" << std::endl
              << "and checks whether the function p represents an upper bound for the number of steps." << std::endl;
    return 1;
  }

  // allocate objects
  unsigned int n = atoi(argv[1]), m = atoi(argv[2]);
  std::list<Object *> objects;
  for (unsigned int i = 0, size = m+2; i < size; ++i)
    objects.push_back (new Object ());

  // ordered sequence pattern
  {
    PatternMatcher matcher (n, m, true);
    matcher.process (objects);
  }

  // unordered sequence pattern
  {
    PatternMatcher matcher (n, m, false);
    matcher.process (objects);
  }

  // cleanup
  for (std::list<Object *>::iterator i = objects.begin(), end = objects.end(); i != end; ++i)
    delete *i;

  // done
  return 0;
}
