/* Object-Oriented Rewrite System (oors) template library
   Copyright (C) 2006  Gernot Gebhard

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA */

#ifndef __OORS_H__
#define __OORS_H__

#include <list>
#include <vector>
#include <string>
#include <algorithm>

/// contains shared pattern matcher stuff
namespace oors
{
  // forward declarations
  template<class S, class T> class Alternative;
  template<class S, class T> class Rule;

  /**
   * Purge the given block.
   * @param block The block to purge.
   */
  template<class S> static void purge (std::list<S *> &block)
  {
    for (typename std::list<S *>::iterator i = block.begin(), end = block.end(); i != end; ++i)
      delete *i;
    block.clear ();
  }

  /**
   * Basic exception.
   */
  class Exception
  {
    public:
      /**
       * Constructor.
       * @param msg Exception message.
       */
      Exception (const std::string &msg)
        : m_msg (msg)
      {
      }

    public:
      /**
       * Access exception message.
       * @return Exception message.
       */
      inline const std::string &msg (void) const { return m_msg; }

    private:
      /// exception message
      std::string m_msg;
  };

  /**
   * Exception that is raised whenever a given block could not be matched.
   */
  template<class S> class MatchException : public Exception
  {
    public:
      /**
       * Initialise exception.
       * @param msg Exception message.
       * @param item The item that could not be matched.
       */
      MatchException (const std::string &msg, const S *item)
        : Exception (msg)
        , m_item (item)
      {
      }

    public:
      /**
       * Access the item that could not be matched.
       * @return The who.
       */
      inline const S *item (void) const { return m_item; }

    private:
      /// the unmatchable item
      const S *m_item;
  };

  /**
   * Exception that is raised if a unavailable pattern is accessed.
   */
  class InvalidAccessException : public Exception
  {
    public:
      /**
       * Initialise exception.
       * @param rule The rule in which it happened.
       * @param index The index of the rule that could not be accessed.
       * @param line The line number.
       */
      InvalidAccessException (const std::string &rule, unsigned int index, unsigned int line)
        : Exception ("Invalid pattern access.")
        , m_rule (rule)
        , m_index (index)
        , m_line (line)
      {
      }

    public:
      /**
       * Access the rule's name.
       * @return The rule's name.
       */
      inline const std::string &rule (void) const { return m_rule; }

      /**
       * Access the index of the unavailable pattern.
       * This will not be the "real" index, but the index as it would appear in the matcher generator.
       * @return Pattern index.
       */
      inline unsigned int index (void) const { return m_index; }

      /**
       * The line number where the error has occured.
       * @return Line number.
       */
      inline unsigned int line (void) const { return m_line; }

    private:
      /// the rule in which it happened
      std::string m_rule;

      /// the index of the unavailable pattern
      unsigned int m_index;

      /// the line number where the error has occured
      unsigned int m_line;
  };

#ifdef OORS_DEBUG
  /**
   * Debugger interface class.
   */
  template<class S, class T> class Debugger
  {
    public:
      /**
       * Virtual destructor that does nothing.
       */
      virtual ~Debugger (void) {}

    public:
      /**
       * Tells the debugger that a new rule is being used.
       * @param rule New rule.
       */
      virtual void currentRule (const Rule<S, T> *) {}

      /**
       * Tell the debugger that the given rule is finished.
       * @param rule Finished rule.
       * @param matched True if the rule has matched successfully.
       */
      virtual void finishRule (const Rule<S, T> *, bool) {}

      /**
       * Tell debugger that given rule is being applied.
       * @param rule Applied rule.
       */
      virtual void applyRule (const Rule<S, T> *) {}

      /**
       * Inform debugger about a newly created alternative.
       * @param alternative Newly created alternative.
       * @param parent The alternative this rule was derived from (may be NULL).
       */
      virtual void newAlternative (const Alternative<S, T> *, const Alternative<S, T> *) {}

      /**
       * Tell debugger which alternative is being used.
       * @param alternative The current alternative that is being used.
       */
      virtual void currentAlternative (const Alternative<S, T> *) {}

      /**
       * Indicates that the given alternative can not be investigated further.
       * @param alternative The alternative that has been deleted.
       */
      virtual void deleteAlternative (const Alternative<S, T> *) {}

      /**
       * Report the result of the item pattern side condition.
       * @param alternative The alternative in which an item pattern should be matched.
       * @param index The item pattern index.
       * @param item The item to match.
       * @param result The result of the side condition.
       */
      virtual void checkItemPattern (const Alternative<S, T> *, unsigned int, const S *, bool) {}

      /**
       * An item pattern has been matched.
       * @param alternative The alternative in which it has been matched.
       * @param index The item pattern index.
       * @param item The item that has been matched.
       */
      virtual void matchItemPattern (const Alternative<S, T> *, unsigned int, const S *) {}

      /**
       * A wildcard pattern has been matched.
       * @param alternative The alternative in which it has been matched.
       * @param index The wildcard pattern index.
       * @param item The item that has been matched.
       */
      virtual void matchWildcardPattern (const Alternative<S, T> *, unsigned int, const S *) {}

      /**
       * Indicates that a wildcard pattern is no longer being matched.
       * @param alternative The alternative in which it has been matched.
       * @param index The wildcard pattern index.
       */
      virtual void finishWildcardPattern (const Alternative<S, T> *, unsigned int) {}

      /**
       * Indicates that the pattern matcher has finished processing the given basic block.
       * @param cost The total cost.
       */
      virtual void finishBasicBlock (int) {}
  };
#endif

  /**
   * Class storing data of a match.
   * A match is used to specify where a rule match starts and ends.
   */
  template<class S> class Match
  {
    public:
      /**
       * Initialise match.
       * @param begin Where to begin matching.
       * @param end Where the matching should end.
       * @param items The amount of available items.
       */
      Match (typename std::list<S *>::iterator begin, typename std::list<S *>::iterator end, unsigned int items)
        : m_level (0)
        , m_begin (begin)
        , m_end (end)
        , m_items (items)
        , m_active (false)
      {
      }

      /**
       * Copy the match and adjust it to the new block.
       * @param match The match to copy from.
       * @param block The block to adjust the match to.
       * @param skip The number of items to skip.
       */
      Match (const Match<S> &match, std::list<S *> &block, unsigned int skip)
        : m_level (match.m_level)
        , m_begin (block.begin())
        , m_end (block.end())
        , m_items (match.m_items-skip)
        , m_active (true)
      {
        for (unsigned int i = 0, skip = block.size()-m_items; i < skip; ++i)
          ++m_begin;
      }

    public:
      /**
       * Set new beginning.
       * @param begin Where to begin matching.
       */
      inline void setBegin (typename std::list<S *>::iterator begin) { m_begin = begin; }

      /**
       * Set new end.
       * @param end Where to end matching.
       */
      inline void setEnd (typename std::list<S *>::iterator end) { m_end = end; }

      /**
       * Set active flag.
       * @param active New active state.
       */
      inline void setActive (bool active) { m_active = active; }

      /**
       * Reset the match.
       * Each reset will increase the internal level that is used to determine the amount of block loops.
       * @param begin Where to begin matching.
       * @param end Where the matching should end.
       * @param items The amount of available items.
       */
      inline void reset (typename std::list<S *>::iterator begin, typename std::list<S *>::iterator end, unsigned int items)
      {
        m_begin  = begin;
        m_end    = end;
        m_items  = items;
        m_active = false;
        ++m_level;
      }

      /**
       * Skip to next item.
       */
      inline void skip (void) { ++m_begin; --m_items; }

      /**
       * Skip a certain amount of items.
       * This will mark the match as active.
       * @param items The amount of items to skip.
       */
      inline void skip (unsigned int items) { m_items -= items; m_active = true; }

    public:
      /**
       * Get current level.
       * @return Current level.
       */
      inline unsigned int level (void) const { return m_level; }

      /**
       * Get first item.
       * @return First item to match.
       */
      inline typename std::list<S *>::iterator begin (void) const { return m_begin; }

      /**
       * Get last item.
       * @return Last item to match (exlusive).
       */
      inline typename std::list<S *>::iterator end (void) const { return m_end; }

      /**
       * Get amount of items.
       * @return Amount of available items.
       */
      inline unsigned int items (void) const { return m_items; }

      /**
       * Check if match is still active.
       * @return True if it is still active.
       */
      inline bool active (void) const { return m_active; }

    private:
      /// how often the block has matched anew
      unsigned int m_level;

      /// start of block
      typename std::list<S *>::iterator m_begin;

      /// end of block
      typename std::list<S *>::iterator m_end;

      /// amount of available items
      unsigned int m_items;

      /// remember if match is active
      bool m_active;
  };

  /**
   * Base pattern matcher class.
   * This class comes without functionality. It only stores variables steering
   * the control flow of the real pattern matcher.
   */
  class BasePatternMatcher
  {
    public:
      /**
       * Initialize pattern matcher.
       * @param profile The profile to use.
       */
      BasePatternMatcher (const std::string &profile)
        : m_profile (profile)
        , m_first_match (true)
        , m_immediate_match (false)
        , m_local_minimum (true)
        , m_max_level (0)
        , m_mask (0)
      {
      }

      /**
       * Initialize pattern matcher.
       * @param profile The profile to use.
       * @param first Enable first-match mode.
       * @param immediate Enable immediate-match mode.
       * @param local Enable local-cost-minimum mode.
       * @param level Maximum recursion level.
       * @param mask Rule selection mask.
       */
      BasePatternMatcher (const std::string &profile, bool first, bool immediate, bool local, unsigned int level, int mask)
        : m_profile (profile)
        , m_first_match (first)
        , m_immediate_match (immediate)
        , m_local_minimum (local)
        , m_max_level (level)
        , m_mask (mask)
      {
      }

    public:
      /**
       * Whether to take the first matching rule or not.
       * @param first True if the first matching rule should be taken.
       */
      inline void setFirstMatch (bool first) { m_first_match = first; }

      /**
       * Whether to immediately apply matched rules.
       * @param immediate True if rules should be applied immediately.
       */
      inline void setImmediateMatch (bool immediate) { m_immediate_match = immediate; }

      /**
       * Whether to determine the global or local minimum.
       * @param local True if the local minimum is to be determined.
       */
      inline void setLocalMinimum (bool local) { m_local_minimum = local; }

      /**
       * Set maximum recursion level.
       * @param level New maximum recursion level.
       */
      inline void setMaxLevel (unsigned int level) { m_max_level = level; }

      /**
       * Set rule mask.
       * @param mask New rule mask.
       */
      inline void setMask (int mask) { m_mask = mask; }

    public:
      /**
       * Access the name of the current profile.
       * @return Name of the current profile.
       */
      inline const std::string &profile (void) const { return m_profile; }

      /**
       * Check whether the pattern matcher works in first-match mode.
       * @return True if in first-match mode.
       */
      inline bool firstMatch (void) const { return m_first_match; }

      /**
       * Check whether the pattern matcher immediately applies matched rules.
       * @return True if in immediate-match mode.
       */
      inline bool immediateMatch (void) const { return m_immediate_match; }

      /**
       * Check whether the pattern matcher determines the local cost minimum.
       * @return True if determining the local cost minimum.
       */
      inline bool localMinimum (void) const { return m_local_minimum; }

      /**
       * Set maximum recursion level.
       * @return The maximum recursion level.
       */
      inline unsigned int maxLevel (void) const { return m_max_level; }

      /**
       * Access rule selection mask.
       * @return The rule selection mask.
       */
      inline int mask (void) const { return m_mask; }

    protected:
      /// the used profile
      std::string m_profile;

      /// whether to work in first-match mode
      bool m_first_match;

      /// whether to immediately apply matched rules
      bool m_immediate_match;

      /// whether to determine the global or local minimum
      bool m_local_minimum;

      /// the maximum recursion level
      unsigned int m_max_level;

      /// the mask to use
      int m_mask;
  };

  /**
   * Abstract pattern matcher.
   */
  template<class S, class T> class PatternMatcher : public BasePatternMatcher
  {
    public:
      /**
       * Initialize pattern matcher.
       * @param profile The profile to use.
       */
      PatternMatcher (const std::string &profile)
        : BasePatternMatcher (profile)
#ifdef OORS_DEBUG
        , m_debugger (NULL)
#endif
      {
      }

      /**
       * Initialize pattern matcher.
       * @param profile The profile to use.
       * @param first Enable first-match mode.
       * @param immediate Enable immediate-match mode.
       * @param local Enable local-cost-minimum mode.
       * @param level Maximum recursion level.
       * @param mask Rule selection mask.
       * @param debugger Optional debugger.
       */
      PatternMatcher (const std::string &profile, bool first, bool immediate, bool local, unsigned int level, int mask
#ifdef OORS_DEBUG
                      , Debugger<S, T> *debugger = NULL
#endif
                     )
        : BasePatternMatcher (profile, first, immediate, local, level, mask)
#ifdef OORS_DEBUG
        , m_debugger (debugger)
#endif
      {
      }

      /**
       * Copy constructor.
       * @param matcher The matcher to copy.
       */
      PatternMatcher (const BasePatternMatcher &matcher)
        : BasePatternMatcher (matcher)
#ifdef OORS_DEBUG
        , m_debugger (NULL)
#endif
      {
      }

      /**
       * Empty virtual destructor.
       */
      virtual ~PatternMatcher (void)
      {
      }

#ifdef OORS_DEBUG
    public:
      /**
       * Set the debugger interface.
       * @param debugger The debugger to use.
       */
      void setDebugger (Debugger<S, T> *debugger)
      {
        m_debugger = debugger;
        for (unsigned int i = 0, size = m_rules.size(); i < size; ++i)
          m_rules[i]->setDebugger (debugger);
      }
#endif

    protected:
      /**
       * Reset given match and decide whether to stop processing or not.
       * @param match The match to reset.
       * @param source The block to match.
       * @return True if processing the block should be stopped.
       */
      inline bool reset (Match<S> &match, std::list<S *> &source) const
      {
        // no need to restart
        if (!match.active() || !source.size())
          return true;

        // reset
        match.reset (source.begin(), source.end(), source.size());
        return m_max_level?(match.level() >= m_max_level):false;
      }

    protected:
      /**
       * How to copy an object.
       * @return Nothing.
       * @throws Exception Because the user has to specify how to duplicate objects.
       */
      virtual S *copy (const S *) const
      {
        throw Exception ("Cannot duplicate object. You need to specify the copy function.");
      }

      /**
       * Copy given block.
       * @param source The block to copy.
       * @param target Where to put the contents.
       * @throws Exception If an item could not be copied.
       */
      void copy (const std::list<S *> &source, std::list<S *> &target)
      {
        for (typename std::list<S *>::const_iterator i = source.begin(), end = source.end(); i != end; ++i)
          target.push_back (copy (*i));
      }

    protected:
      /**
       * Process given block recursively determining the local minimum.
       * @param source The block to process.
       */
      void processLocal (std::list<S *> &source)
      {
        // nothing to do
        if (source.empty())
          return;

        // preparations
        Match<S> match (source.begin(), source.end(), source.size());
        const Alternative<S, T> *alternative = NULL;
        std::vector<Alternative<S, T> *> alternatives;
#ifdef OORS_DEBUG
        int cost = 0;
#endif

        // do the thing
        do
        {
          // at least one rule has matched
          PatternMatcher::match (alternative, match, alternatives);
          if (!alternatives.empty())
          {
            // delete old alternative
           if (alternative)
              delete alternative;

            // get cheapest alternative
            typename Alternative<S, T>::Comparator comparator;
            std::sort (alternatives.begin(), alternatives.end(), comparator);
            alternative = alternatives.front ();

            // get rule and apply it
            Rule<S, T> *rule = alternative->rule ();
            rule->setAlternative (const_cast<Alternative<S, T> *>(alternative));
#ifdef OORS_DEBUG
            if (m_debugger)
            {
              m_debugger->applyRule (rule);
              cost += alternative->cost ();
            }
#endif
            rule->apply (*this, source, true);

            // get rid of all unused alternatives
            for (unsigned int i = 1, size = alternatives.size(); i < size; ++i)
              delete alternatives[i];
            alternatives.clear ();

            // adjust control vars
            match.skip (alternative->skip());
          }

          // no match possible
          else
          {
            // get the first available item
            if (alternative)
            {
              match.setBegin (alternative->items().front());
              delete alternative;
              alternative = NULL;
            }

            // skip to next item since there was no match possible starting with this one
            match.skip ();
          }

          // all items have been processed
          if (!match.items())
          {
            // remove old alternative
            if (alternative)
            {
              delete alternative;
              alternative = NULL;
            }

            // restart
            if (reset (match, source))
              break;
          }
        } while (true);

#ifdef OORS_DEBUG
        // done
        if (m_debugger)
          m_debugger->finishBasicBlock (cost);
#endif
      }

      /**
       * Process given block always using the local minimum.
       * @param source The block to process.
       * @param target The resulting block.
       * @throws Exception If input block cannot be matched.
       */
      void processLocal (const std::list<S *> &source, std::list<T *> &target)
      {
        // nothing to do
        if (source.empty())
          return;

        // preparations
        Match<S> match (const_cast<std::list<S *> &>(source).begin(), const_cast<std::list<S *> &>(source).end(),  source.size());
        const Alternative<S, T> *alternative = NULL;
        std::vector< std::vector<Alternative<S, T> *> > trace;
        trace.push_back (std::vector<Alternative<S, T> *> ());

        // do the thing
        do
        {
          // let every rule do it's thing
          std::vector<Alternative<S, T> *> &alternatives = trace.back ();
          PatternMatcher::match (alternative, match, alternatives);

          // at least one rule has been matched
          // order alternatives and select cheapest one
          if (!alternatives.empty())
          {
            typename Alternative<S, T>::Comparator comparator;
            std::sort (alternatives.begin(), alternatives.end(), comparator);
            alternative = alternatives.front ();
            if (!alternative->empty())
              trace.push_back (std::vector<Alternative<S, T> *> ());
          }

          // no rule has matched
          else
          {
            // remember the item for it could be the one that cannot be matched
            const S *item = alternative?*alternative->items().front():source.front();

            // find an alternative that could be used to go on
            do
            {
              // pop current level and bail out if there are no more alternatives
              unsigned int size = trace.size ();
              if (size == 1)
                throw MatchException<S> ("Given block cannot be matched.", item);

              // select previous level and remove first alternative
              std::vector<Alternative<S, T> *> &previous = trace[size-2];
              delete previous.front ();
              previous.erase (previous.begin());

              // select next alternative if not empty
              if (!previous.empty())
              {
                alternative = previous.front ();
                if (alternative->empty())
                  trace.pop_back ();
                break;
              }
              else
                trace.pop_back ();
            } while (true);
          }
        } while (!alternative->empty());

        // after all items have been matched apply the rules and get rid of all alternatives
#ifdef OORS_DEBUG
        int cost = 0;
#endif
        for (unsigned int i = 0, size = trace.size(); i < size; ++i)
        {
          // preparations
          std::vector<Alternative<S, T> *> &alternatives = trace[i];
          Alternative<S, T> *alternative = alternatives.front ();

          // do the thing
          Rule<S, T> *rule = alternative->rule ();
          rule->setAlternative (alternative);
#ifdef OORS_DEBUG
          if (m_debugger)
          {
            m_debugger->applyRule (rule);
            cost += alternative->cost ();
          }
#endif
          rule->apply (*this, target, false);
          for (unsigned int i = 0, size = alternatives.size(); i < size; ++i)
            delete alternatives[i];
          alternatives.clear ();
        }
        trace.clear ();

#ifdef OORS_DEBUG
        // done
        if (m_debugger)
          m_debugger->finishBasicBlock (cost);
#endif
      }

      /**
       * Process given block immediately applying the rule with the lowest cost (aiming at a local cost minimum).
       * @param source The block to process.
       * @param target The resulting block.
       * @throws Exception If input block cannot be matched.
       */
      void processLocalImmediate (const std::list<S *> &source, std::list<T *> &target)
      {
        // nothing to do
        if (source.empty())
          return;

        // preparations
#ifdef OORS_DEBUG
        int cost = 0;
#endif
        Match<S> match (const_cast<std::list<S *> &>(source).begin(), const_cast<std::list<S *> &>(source).end(),  source.size());
        Alternative<S, T> *alternative = NULL;
        typename std::list<T *>::iterator target_end = target.end ();
        std::vector< std::pair< std::vector<Alternative<S, T> *>, std::list<typename std::list<T *>::iterator> > > trace;
        trace.push_back (std::pair< std::vector<Alternative<S, T> *>, std::list<typename std::list<T *>::iterator> > ());

        // do the thing
        do
        {
          // let every rule do it's thing
          std::pair< std::vector<Alternative<S, T> *>, std::list<typename std::list<T *>::iterator> > &current = trace.back ();
          std::vector<Alternative<S, T> *> &alternatives = current.first;
          PatternMatcher::match (alternative, match, alternatives);

          // at least one rule has been matched
          // order alternatives and select cheapest one
          if (!alternatives.empty())
          {
            // select locally cheapest alternative
            typename Alternative<S, T>::Comparator comparator;
            std::sort (alternatives.begin(), alternatives.end(), comparator);
            alternative = alternatives.front ();

            // inform debugger (if applicable)
            Rule<S, T> *rule = alternative->rule ();
            rule->setAlternative (alternative);
#ifdef OORS_DEBUG
            if (m_debugger)
            {
              m_debugger->applyRule (rule);
              cost += alternative->cost ();
            }
#endif

            // apply rule locally
            std::list<T *> local_target;
            rule->apply (*this, local_target, false);

            // remember generated instructions to enable proper cleanup when backtracking
            std::list<typename std::list<T *>::iterator> &instructions = current.second;
            for (typename std::list<T *>::iterator i = local_target.begin(), end = local_target.end(); i != end; ++i)
              instructions.push_back (target.insert (target_end, *i));

            // it seems we are not yet done processing the input
            if (!alternative->empty())
              trace.push_back (std::pair< std::vector<Alternative<S, T> *>, std::list<typename std::list<T *>::iterator> > ());
          }

          // no rule has matched
          else
          {
            // remember the item for it could be the one that cannot be matched
            const S *item = alternative?*alternative->items().front():source.front();

            // find an alternative that could be used to go on
            do
            {
              // there are no alternatives left to explore, bail out
              unsigned int size = trace.size ();
              if (size == 1)
                throw MatchException<S> ("Given block cannot be matched.", item);

              // remove the alternative that failed
              std::pair< std::vector<Alternative<S, T> *>, std::list<typename std::list<T *>::iterator> > &previous = trace[size-2];
              std::vector<Alternative<S, T> *> &previous_alternatives = previous.first;
#ifdef OORS_DEBUG
              cost -= previous_alternatives.front()->cost();
#endif
              delete previous_alternatives.front ();
              previous_alternatives.erase (previous_alternatives.begin());

              // and get rid of the corresponding generated instructions
              std::list<typename std::list<T *>::iterator> &instructions = previous.second;
              for (typename std::list<typename std::list<T *>::iterator>::iterator i = instructions.begin(), end = instructions.end(); i != end; ++i)
              {
                delete **i;
                target.erase (*i);
              }
              instructions.clear ();

              // if there are alternatives left from the previous run, select the next
              if (!previous_alternatives.empty())
              {
                // inform debugger (if applicable)
                alternative = previous_alternatives.front ();
                Rule<S, T> *rule = alternative->rule ();
                rule->setAlternative (alternative);
#ifdef OORS_DEBUG
                if (m_debugger)
                {
                  m_debugger->applyRule (rule);
                  cost += alternative->cost ();
                }
#endif

                // apply rule locally
                std::list<T *> local_target;
                rule->apply (*this, local_target, false);

                // remember generated instructions to enable proper cleanup when backtracking
                for (typename std::list<T *>::iterator i = local_target.begin(), end = local_target.end(); i != end; ++i)
                  instructions.push_back (target.insert (target_end, *i));

                // seems we are done with that input block
                if (alternative->empty())
                  trace.pop_back ();
                break;
              }

              // if there are no alternatives left, backtrack even further
              else
                trace.pop_back ();
            } while (true);
          }
        } while (!alternative->empty());

        // get rid of all alternatives
        for (unsigned int i = 0, size = trace.size(); i < size; ++i)
        {
          std::vector<Alternative<S, T> *> &alternatives = trace[i].first;
          for (unsigned int i = 0, size = alternatives.size(); i < size; ++i)
            delete alternatives[i];
          alternatives.clear ();
        }
        trace.clear ();

#ifdef OORS_DEBUG
        // done
        if (m_debugger)
          m_debugger->finishBasicBlock (cost);
#endif
      }

      /**
       * Process the given block recursively determining the global minimum.
       * @param source The block to process.
       */
      void processGlobal (std::list<S *> &source)
      {
        // nothing to do
        if (source.empty())
          return;

        // determinal global minimum recursively
        Alternative<S, T> *alternative = NULL;
        Match<S> match (source.begin(), source.end(), source.size());
#ifdef OORS_DEBUG
        int cost = processGlobal (alternative, match, source);
        if (m_debugger)
          m_debugger->finishBasicBlock (cost);
#else
        processGlobal (alternative, match, source);
#endif
      }

      /**
       * Process given block determining the global minimum.
       * @param source The block to process.
       * @param target The resulting block.
       * @throws Exception If input block cannot be matched.
       */
      void processGlobal (const std::list<S *> &source, std::list<T *> &target)
      {
        // nothing to do
        if (source.empty())
          return;

        // determine global minimum
        Match<S> match (const_cast<std::list<S *> &>(source).begin(), const_cast<std::list<S *> &>(source).end(),  source.size());
        std::vector<Alternative<S, T> *> trace;
        processGlobal (NULL, match, trace);

        // apply matched rules
#ifdef OORS_DEBUG
        int cost = 0;
#endif
        for (unsigned int i = 0, size = trace.size(); i < size; ++i)
        {
          Alternative<S, T> *alternative = trace[i];
          Rule<S, T> *rule = alternative->rule ();
          rule->setAlternative (alternative);
#ifdef OORS_DEBUG
          if (m_debugger)
          {
            m_debugger->applyRule (rule);
            cost += alternative->cost ();
          }
#endif
          rule->apply (*this, target, false);
          delete alternative;
        }
        trace.clear ();

#ifdef OORS_DEBUG
        if (m_debugger)
          m_debugger->finishBasicBlock (cost);
#endif
      }

    private:
      /**
       * One iteration of the recursive global minimum determination.
       * @param alternative The alternative to start with.
       * @param match The current match.
       * @param source The block to modify.
       */
      int processGlobal (Alternative<S, T> *alternative, Match<S> &match, std::list<S *> &source)
      {
        // reset match if alternative is empty (= no more available items)
        if (alternative && alternative->empty())
        {
          // reset alternative to make the next rule create a new one
          delete alternative;
          alternative = NULL;

          // well, there is nothing more to do
          if (reset (match, source))
            return 0;
        }

        // apply rules until a match is found
        std::vector<Alternative<S, T> *> alternatives;
        while (true)
        {
          // apply, if a rule has matched all is fine
          PatternMatcher::match (alternative, match, alternatives);
          if (!alternatives.empty())
            break;

          // select the first item
          if (alternative)
          {
            match.setBegin (alternative->items().front());
            delete alternative;
            alternative = NULL;
          }

          // skip to next item and restart this process eventually
          match.skip ();

          // all items processed, then check if restart is possible
          if (!match.items() && reset(match, source))
            return 0;
        }

        // get rid of old alternative
        if (alternative)
        {
          delete alternative;
          alternative = NULL;
        }

        // do not copy the basc block if there is only one alternative
        // simply apply the rule
        unsigned int size = alternatives.size ();
        if (size == 1)
        {
          // get alternative and rule and apply it
          Alternative<S, T> *alternative = alternatives.front();
          Rule<S, T> *rule = alternative->rule ();
          rule->setAlternative (alternative);
#ifdef OORS_DEBUG
          if (m_debugger)
            m_debugger->applyRule (rule);
#endif
          rule->apply (*this, source, true);

          // mark that this match is active
          match.setActive (true);

          // then do the recursion
          return alternative->cost()+processGlobal (alternative, match, source);
        }

        // now comes the bad thing:
        // each alternative is a potential candidate for the best matching
        // thus each alternative must get its own basic block and match (and that is expensive!)
        int cost = 0;
        std::list<S *> a, b, *current = NULL;
        for (unsigned int i = 0; i < size; ++i)
        {
          // get alternative and current active block
          alternative = alternatives[i];
          std::list<S *> &block = (current == &a)?b:a;

          // copy block and match then adjust alternative and copied match to new block
          copy (source, block);
          alternative->adjustTo (source, block);
          Match<S> copy (match, block, alternative->skip());

          // get rule and apply it
          Rule<S, T> *rule = alternative->rule ();
          rule->setAlternative (alternative);
#ifdef OORS_DEBUG
          if (m_debugger)
            m_debugger->applyRule (rule);
#endif
          rule->apply (*this, block, true);

          // then do the recursion and get rid of the alternative since it is not used anymore
          int new_cost = alternative->cost()+processGlobal (alternative, copy, block);

          // do nothing on first run
          if (!current)
          {
            cost = new_cost;
            current = &block;
          }

          // good, the new block is cheaper
          else if (new_cost < cost)
          {
            purge<S> (*current);
            cost = new_cost;
            current = &block;
          }

          // boy, forget that thing, its too expensive
          else
            purge<S> (block);
        }

        // get rid of the old block and assign the actual one
        if (current)
        {
          purge<S> (source);
          source = *current;
        }

        // done
        return cost;
      }

      /**
       * One iteration of the global minimum determination.
       * @param alternative The alternative to start with.
       * @param match The current match.
       * @param trace Where to put the results.
       * @return The cost of the produced trace.
       * @throws Exception If the block could not be matched beginning at the given alternative.
       */
      int processGlobal (Alternative<S, T> *alternative, Match<S> &match, std::vector<Alternative<S, T> *> &trace)
      {
        // do one pass
        std::vector<Alternative<S, T> *> alternatives;
        PatternMatcher::match (alternative, match, alternatives);

        // remember the item that could not be matched
        const S *item = NULL;

        // behold the magicks of the global minimum
        if (!alternatives.empty())
        {
          // preparations
          int cost = 0;
          Alternative<S, T> *previous = NULL;
          std::vector<Alternative<S, T> *> a, b, *subtrace = NULL;

          // match beginning at each subalternative
          for (unsigned int i = 0, size = alternatives.size(); i < size; ++i)
            try
            {
              // reset unmatched item and get current alternative
              item = NULL;
              alternative = alternatives[i];

              // first run
              if (!subtrace)
              {
                cost     = alternative->cost()+(alternative->empty()?0:processGlobal(alternative, match, a));
                previous = alternative;
                subtrace = &a;
              }

              // consecutive run
              else
              {
                std::vector<Alternative<S, T> *> &old_trace = *subtrace, &new_trace = (subtrace == &a)?b:a;
                int new_cost = alternative->cost()+(alternative->empty()?0:processGlobal(alternative, match, new_trace));
                if (new_cost < cost)
                {
                  // get rid of old stuff
                  delete previous;
                  for (unsigned int i = 0, size = old_trace.size(); i < size; ++i)
                    delete old_trace[i];
                  old_trace.clear ();

                  // set new stuff
                  cost     = new_cost;
                  previous = alternative;
                  subtrace = (subtrace == &a)?&b:&a;
                }
                else
                {
                  // get rid of the new stuff since it is not cheaper
                  delete alternative;
                  for (unsigned int i = 0, size = new_trace.size(); i < size; ++i)
                    delete new_trace[i];
                  new_trace.clear ();

                  // make sure alternative is set to the head of the cheapest trace
                  alternative = previous;
                }
              }
            }

            // the given alternative did not lead to an end
            catch (MatchException<S> exception)
            {
              item = exception.item ();
              delete alternative;
              alternative = previous;
            }

          // be done with this alternative
          if (subtrace)
          {
            trace.push_back (alternative);
            std::vector<Alternative<S, T> *> &final = *subtrace;
            for (unsigned int i = 0, size = final.size(); i < size; ++i)
              trace.push_back (final[i]);
            return cost;
          }
        }

        // no way
        throw MatchException<S> ("Given block cannot be matched.", item?item:alternative?*alternative->items().front():*match.begin());
      }

    protected:
      /**
       * Apply the set of rules on the current block.
       * @param alternative The current active alternative.
       * @param match The current match.
       * @param alternatives Where to put the results.
       */
      void match (const Alternative<S, T> *alternative, const Match<S> &match, std::vector<Alternative<S, T> *> &alternatives)
      {
        for (unsigned int i = 0, size = m_rules.size(); i < size; ++i)
        {
          // get rule and try match if allowed to
          Rule<S, T> *rule = m_rules[i];
          if (rule->enabled(m_mask))
            rule->match (alternative, match, alternatives);

          // stop now if rule has matched when being in first-match mode
          if (m_first_match && !alternatives.empty())
            return;
        }
      }

    protected:
      /// list of available rules
      std::vector<Rule<S, T> *> m_rules;

#ifdef OORS_DEBUG
      /// the used debugger
      Debugger<S, T> *m_debugger;
#endif
  };
}

#endif
