/* Object-Oriented Rewrite System (oors) template library
   Copyright (C) 2006  Gernot Gebhard

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA */

#ifndef __OORS_INTERNAL_H__
#define __OORS_INTERNAL_H__

#include "oors.h"

#include <list>
#include <vector>
#include <string>

#include <iostream>

/// contains shared pattern matcher stuff
namespace oors
{
  // forward declarations
  template<class S> class SequencePattern;
  template<class S, class T> class Rule;

  /**
   * Abstract pattern.
   */
  template<class S> class Pattern
  {
    public:
      /**
       * Initialize pattern.
       * @param parent Parent pattern.
       * @param index Pattern index.
       */
      Pattern (SequencePattern<S> *parent, unsigned int index)
        : m_parent (parent)
        , m_index (index)
        , m_pos (0)
      {
      }

      /**
       * Empty virtual destructor.
       */
      virtual ~Pattern (void)
      {
      }

    public:
      /**
       * Set the pattern position.
       * @param pos The new position.
       */
      inline void setPos (unsigned int pos) { m_pos = pos; }

    public:
      /**
       * Access the parent pattern.
       * @return Parent pattern.
       */
      inline SequencePattern<S> *parent (void) const { return m_parent; }

      /**
       * Access the pattern index.
       * @return Pattern index.
       */
      inline unsigned int index (void) const { return m_index; }

      /**
       * Access the pattern pos.
       * @return Pattern position.
       */
      inline unsigned int pos (void) const { return m_pos; }

    public:
      /**
       * Insert given item.
       * @param item The item that is to be inserted.
       */
      virtual void insert (typename std::list<S *>::iterator &item) = 0;

      /**
       * Tell the pattern that it is done.
       */
      virtual void done (void) = 0;

    public:
      /**
       * Clone the pattern.
       * @param parent The parent of the cloned pattern.
       * @param patterns Store accessible patterns here.
       * @return A copy of the current pattern.
       */
      virtual Pattern<S> *clone (SequencePattern<S> *parent, Pattern<S> **patterns) const = 0;

      /**
       * Check if the pattern has been matched.
       * @return True if the pattern has been matched.
       */
      virtual bool matched (void) const = 0;

    public:
      /**
       * Compute FIRST set.
       * @param set Where to store the patterns.
       */
      void first (std::list<Pattern<S> *> &set) const
      {
        // compute FIRST for sequence pattern
        const SequencePattern<S> *sequence = dynamic_cast<const SequencePattern<S> *>(this);
        if (sequence)
          if (sequence->ordered())
          {
            if (sequence->size())
              sequence->pattern(0)->first (set);
          }
          else
            for (unsigned int i = 0, size = sequence->size(); i < size; ++i)
              sequence->pattern(i)->first (set);

        // FIRST for non-sequence patterns is simple
        else
          set.push_back (const_cast<Pattern<S> *>(this));
      }

      /**
       * Compute FOLLOW set.
       * Only non-matched patterns will be inserted.
       * @param set Where to store the patterns.
       */
      void follow (std::list<Pattern<S> *> &set) const
      {
        // no parent no successor
        if (!m_parent)
          return;

        // select direct successor in ordered sequence
        unsigned int size = m_parent->size ();
        if (m_parent->ordered())
        {
          unsigned int index = m_pos+1;
          if (index < size)
          {
            m_parent->pattern(index)->first (set);
            return;
          }
        }

        // all patterns within unordered sequence are of interest except the current one
        else
        {
          bool done = false;
          for (unsigned int i = 0; i < size; ++i)
          {
            Pattern<S> *pattern = m_parent->pattern (i);
            if (!pattern->matched() && (pattern != this))
            {
              pattern->first (set);
              done = true;
            }
          }
          if (done)
            return;
        }

        // get parent's follow set
        m_parent->follow (set);
      }

    protected:
      /// parent pattern
      SequencePattern<S> *m_parent;

      /// pattern index
      unsigned int m_index;

      /// pattern position
      unsigned int m_pos;
  };

  /**
   * Sequence pattern.
   */
  template<class S> class SequencePattern : public Pattern<S>
  {
    public:
      /**
       * Create new sequence pattern.
       * @param parent Parent pattern.
       * @param index Pattern index.
       * @param size Sequence size.
       * @param ordered Indicates whether this is an ordered sequence or not.
       */
      SequencePattern (SequencePattern<S> *parent, unsigned int index, unsigned int size, bool ordered)
        : Pattern<S> (parent, index)
        , m_current (0)
        , m_sequence (new Pattern<S>* [size])
        , m_size (size)
        , m_ordered (ordered)
      {
      }

      /**
       * Dispose the sequence pattern.
       */
      ~SequencePattern (void)
      {
        for (unsigned int i = 0; i < m_size; ++i)
          delete m_sequence[i];
        delete[] m_sequence;
      }

    public:
      /**
       * Insert the given pattern.
       * @param pattern The pattern that is to be inserted.
       * @param index Where to put the pattern.
       */
      inline void insert (Pattern<S> *pattern, unsigned int index) { m_sequence[index] = pattern; pattern->setPos(index); }

    public:
      /**
       * Access pattern size.
       * @return Pattern size.
       */
      inline unsigned int size (void) const { return m_size; }

      /**
       * Access sequence type.
       * @return True if ordered, false else.
       */
      inline bool ordered (void) const { return m_ordered; }

      /**
       * Access pattern at given index.
       * @param index Pattern index.
       * @return Required pattern.
       */
      inline Pattern<S> *pattern (unsigned int index) const { return m_sequence[index]; }

      /**
       * Determine which pattern is currently to be matched within the given alternative.
       * @return The pattern that is to be matched next, or NULL if there is none.
       */
      inline Pattern<S> *current (void) const { return (m_current < m_size)?m_sequence[m_current]:NULL; }

    public:
      void insert (typename std::list<S *>::iterator &item)
      {
        if (m_current < m_size)
          m_sequence[m_current]->insert (item);
      }

      void done (void)
      {
        if ((m_current < m_size) && (++m_current >= m_size) && Pattern<S>::m_parent)
          Pattern<S>::m_parent->done ();
      }

    public:
      bool matched (void) const { return m_current == m_size; }

      Pattern<S> *clone (SequencePattern<S> *parent, Pattern<S> **patterns) const
      {
        SequencePattern<S> *pattern = new SequencePattern<S> (parent, Pattern<S>::m_index, m_size, m_ordered);
        for (unsigned int i = 0; i < m_size; ++i)
          pattern->m_sequence[i] = m_sequence[i]->clone (pattern, patterns);
        pattern->m_pos     = Pattern<S>::m_pos;
        pattern->m_current = m_current;
        return pattern;
      }

    private:
      /// position within the sequence
      unsigned int m_current;

      /// pattern sequence
      Pattern<S> **m_sequence;

      /// sequence size
      unsigned int m_size;

      /// sequecne type
      bool m_ordered;
  };

  /**
   * Wildcard pattern.
   */
  template<class S> class WildcardPattern : public Pattern<S>
  {
    public:
      /**
       * Initialize new wildcard pattern.
       * @param parent Parent pattern.
       * @param index Pattern index.
       */
      WildcardPattern (SequencePattern<S> *parent, unsigned int index)
        : Pattern<S> (parent, index)
        , m_matched (false)
      {
      }

    public:
      /**
       * Replace stored item with a copy.
       * @param index Index of the item to replace.
       * @param item Copy of the item.
       */
      inline void replace (unsigned int index, S *item) { m_items[index] = item; }

    public:
      /**
       * Access the matched items.
       * @return A list of matched items.
       */
      inline const std::vector<S *> &items (void) const { return m_items; }

      /**
       * Access the amount of matched items within the given alternative.
       * @return The amount of matched patterns.
       */
      inline unsigned int amount (void) const { return m_items.size (); }

    public:
      void insert (typename std::list<S *>::iterator &item) { m_items.push_back (*item); }

      void done (void)
      {
        if (Pattern<S>::m_parent && !m_matched)
          Pattern<S>::m_parent->done ();
        m_matched = true;
      }

    public:
      Pattern<S> *clone (SequencePattern<S> *parent, Pattern<S> **patterns) const
      {
        WildcardPattern<S> *pattern = new WildcardPattern<S> (parent, Pattern<S>::m_index);
        pattern->m_pos     = Pattern<S>::m_pos;
        pattern->m_items   = m_items;
        pattern->m_matched = m_matched;
        return patterns[Pattern<S>::m_index] = pattern;
      }

      bool matched (void) const { return m_matched; }

    private:
      /// matched items
      std::vector<S *> m_items;

      /// indicates whether the wildcard pattern is done or not
      bool m_matched;
  };

  /**
   * Item pattern.
   */
  template<class S> class ItemPattern : public Pattern<S>
  {
    public:
      /**
       * Initialize new item pattern.
       * @param parent Parent pattern.
       * @param index Pattern index.
       * @param remove Whether to remove the matched item from the basic block.
       */
      ItemPattern (SequencePattern<S> *parent, unsigned int index, bool remove = true)
        : Pattern<S> (parent, index)
        , m_item (NULL)
        , m_dispose (true)
        , m_remove (remove)
      {
      }

    public:
      /**
       * Access the matched item while setting the dispose flag.
       * @param dispose New dispose state.
       * @return Matched item.
       */
      inline S *item (bool dispose) { return (m_dispose = dispose),m_item; }

      /**
       * Replace stored item by a copy.
       * @param item The copy of the item.
       */
      inline void replace (typename std::list<S *>::iterator &item) { m_item = *(m_iterator = item); }

    public:
      /**
       * Access the matched item.
       * @return The matched item.
       */
      inline S *item (void) const { return m_item; }

      /**
       * Access the iterator belonging to the matched item.
       * @return Associated iterator.
       */
      inline const typename std::list<S *>::iterator &iterator (void) const { return m_iterator; }

      /**
       * Check whether the matched item should be disposed.
       * @return True if so.
       */
      inline bool dispose (void) const { return m_dispose && m_remove; }

      /**
       * Check whether the matched item should be removed from the basic block.
       * @return True if so.
       */
      inline bool remove (void) const { return m_remove; }

    public:
      void insert (typename std::list<S *>::iterator &item)
      {
        m_item = *(m_iterator = item);
        if (Pattern<S>::m_parent)
          Pattern<S>::m_parent->done ();
      }

      void done (void) { /* the pattern knows when it's done */ }

    public:
      Pattern<S> *clone (SequencePattern<S> *parent, Pattern<S> **patterns) const
      {
        ItemPattern<S> *pattern = new ItemPattern<S> (parent, Pattern<S>::m_index);
        pattern->m_pos      = Pattern<S>::m_pos;
        pattern->m_item     = m_item;
        pattern->m_iterator = m_iterator;
        pattern->m_dispose  = m_dispose;
        pattern->m_remove   = m_remove;
        return patterns[Pattern<S>::m_index] = pattern;
      }

      bool matched (void) const { return m_item; }

    private:
      /// matched item
      S *m_item;

      /// associated iterator
      typename std::list<S *>::iterator m_iterator;

      /// indicates whether the matched item should be disposed
      bool m_dispose;

      /// inidicates whether the matched item should be removed from the basic block
      bool m_remove;
  };

  /**
   * An alternative stores available items and a (partially) matched pattern.
   */
  template<class S, class T> class Alternative
  {
    public:
      /**
       * An alternative comparator.
       */
      class Comparator
      {
        public:
          /**
           * Compare to alternatives according to their cost.
           * @param a First alternative.
           * @param b Second alternative.
           * @return True if first alternative is cheaper than the second one.
           */
          inline bool operator() (const Alternative<S, T> *a, const Alternative<S, T> *b) { return a->cost() < b->cost(); }
      };

    public:
      /**
       * Initialize alternative.
       * @param alternative The alternative to derive from.
       * @param match Data concerning the current match.
       * @param rule The rule this alternative belongs to.
       * @param sequence The used search pattern.
       * @param size Amount of accessible patterns.
       * @param debugger The used debugger.
       */
      Alternative (const Alternative<S, T> *alternative, const Match<S> &match, Rule<S, T> *rule, SequencePattern<S> *sequence, unsigned int size
#ifdef OORS_DEBUG
              , Debugger<S, T> *debugger
#endif
             )
        : m_matched (0)
        , m_skip (0)
        , m_rule (rule)
        , m_patterns (new Pattern<S> * [size])
        , m_sequence (static_cast<SequencePattern<S> *>(sequence->clone(NULL, m_patterns)))
        , m_pattern (NULL)
        , m_size (size)
        , m_cost (0)
#ifdef OORS_DEBUG
        , m_debugger (debugger)
#endif
      {
        // copy available items
        if (alternative)
          m_available = alternative->m_available;

        // copy all items
        else
          for (typename std::list<S *>::iterator i = match.begin(), end = match.end(); i != end; ++i)
            m_available.push_back (i);

        // init missing vars
        m_current   = m_available.begin ();
        m_remaining = m_available.size ();

#ifdef OORS_DEBUG
        // inform debugger
        if (m_debugger)
          m_debugger->newAlternative (this, alternative);
#endif
      }

      /**
       * Copy alternative.
       * @param alternative The alternative to copy from.
       */
      Alternative (const Alternative<S, T> *alternative)
        : m_available (alternative->m_available)
        , m_remaining (alternative->m_remaining)
        , m_matched (alternative->m_matched)
        , m_skip (alternative->m_skip)
        , m_rule (alternative->m_rule)
        , m_patterns (new Pattern<S>* [alternative->m_size])
        , m_sequence (static_cast<SequencePattern<S> *>(alternative->m_sequence->clone(NULL, m_patterns)))
        , m_pattern (alternative->m_pattern?m_patterns[alternative->m_pattern->index()]:NULL)
        , m_size (alternative->m_size)
        , m_cost (0)
#ifdef OORS_DEBUG
        , m_debugger (alternative->m_debugger)
#endif
      {
        // select current item
        m_current = m_available.begin();
        for (unsigned int i = m_available.size(); i != m_remaining; --i)
          ++m_current;

        // copy pattern trace
        const std::list<const Pattern<S> *> &trace = alternative->m_trace;
        for (typename std::list<const Pattern<S> *>::const_iterator i = trace.begin(), end = trace.end(); i != end; ++i)
          m_trace.push_back (m_patterns[(*i)->index()]);

#ifdef OORS_DEBUG
        // inform debugger
        if (m_debugger)
          m_debugger->newAlternative (this, alternative);
#endif
      }

      /**
       * Dispose alternative.
       */
      ~Alternative (void)
      {
#ifdef OORS_DEBUG
        // inform debugger
        if (m_debugger)
          m_debugger->deleteAlternative (this);
#endif

        // delete patterns and stuff
        delete m_sequence;
        delete[] m_patterns;
      }

    public:
      /**
       * Set current active pattern.
       * @param index The index of the active pattern.
       */
      inline void setPattern (unsigned int index) { m_pattern = m_patterns[index]; }

      /**
       * Set the cost of the alternative.
       * @param cost The alternative's cost.
       */
      inline void setCost (int cost) { m_cost = cost; }

    public:
      /**
       * Access the alternative's rule.
       * @return The rule the alternative belongs to.
       */
      inline Rule<S, T> *rule (void) const { return m_rule; }

      /**
       * Access the main pattern.
       * @return Main pattern.
       */
      inline SequencePattern<S> *sequence (void) const { return m_sequence; }

      /**
       * Access the current active pattern.
       * @return Current active pattern.
       */
      inline Pattern<S> *pattern (void) const { return m_pattern; }

      /**
       * Access the ith accessible pattern.
       * @param index Pattern index.
       * @return The required pattern.
       */
      inline Pattern<S> *pattern (unsigned int index) const { return m_patterns[index]; }

      /**
       * Access the amount of matched items in wildcard pattern with the given index.
       * @param index The index of the wildcard pattern.
       * @return The amount of matched items.
       */
      inline unsigned int wildcard (unsigned int index) const { return static_cast<WildcardPattern<S> *>(m_patterns[index])->amount(); }

      /**
       * Access the matched item of the item pattern with the given index.
       * @param index The index of the item pattern.
       * @return The matched item.
       */
      inline S *item (unsigned int index) const { return static_cast<ItemPattern<S> *>(m_patterns[index])->item(); }

      /**
       * Access the current item.
       * @return The current item that is to be matched.
       */
      inline S *current (void) const { return **m_current; }

      /**
       * Access the list of available items.
       * @return All available items.
       */
      inline const std::list<typename std::list<S *>::iterator> &items (void) const { return m_available; }

      /**
       * Access how many items are available at all.
       * @return Amount of available items including the ones matched by a wildcard pattern.
       */
      inline unsigned int available (void) const { return m_available.size(); }

      /**
       * Check how many items have not been visited yet.
       * @return Remaining items.
       */
      inline unsigned int remaining (void) const { return m_remaining; }

      /**
       * Checks whether there is still an available item.
       * @return False if there is at least one available item left.
       */
      inline bool done (void) const { return m_current == m_available.end(); }

      /**
       * Check whether the list of available items is empty.
       * @return True if there are no more items left.
       */
      inline bool empty (void) const { return m_available.empty(); }

      /**
       * Get the amount of matched items.
       * @return Amount of matched items.
       */
      inline unsigned int matched (void) const { return m_matched; }

      /**
       * Get the amount of items to skip.
       * @return Amount of items to skip.
       */
      inline unsigned int skip (void) const { return m_skip; }

      /**
       * Access the pattern trace.
       * @return A list of the matched patterns in order of appearance.
       */
      inline const std::list<const Pattern<S> *> &trace (void) const { return m_trace; }

      /**
       * Access the alternative's cost.
       * @return Cost of the alternative.
       */
      inline int cost (void) const { return m_cost; }

    public:
      /**
       * Remember current item as being matched by an wildcard pattern.
       * @param index The index of the wildcard pattern.
       */
      void setWildcard (unsigned int index)
      {
        // insert into wildcard pattern
        typename std::list<S *>::iterator &current = *m_current;
        (m_pattern = m_patterns[index])->insert (current);
        ++m_current;
        --m_remaining;

        // remember pattern in trace
        if (!m_trace.size() || (m_trace.back() != m_pattern))
          m_trace.push_back (m_pattern);

#ifdef OORS_DEBUG
        // inform debugger
        if (m_debugger)
          m_debugger->matchWildcardPattern (this, index+1, *current);
#endif
      }

      /**
       * Finish given wildcard pattern.
       * @param index Wildcard pattern index.
       */
      void finishWildcard (unsigned int index)
      {
        (m_pattern = m_patterns[index])->done ();
#ifdef OORS_DEBUG
        if (m_debugger)
          m_debugger->finishWildcardPattern (this, index+1);
#endif
      }

      /**
       * Mark current item as being matched by an item pattern.
       * @param index The index of the item pattern.
       */
      void setItem (unsigned int index)
      {
        // insert into item pattern
        typename std::list<S *>::iterator &current = *m_current;
#ifdef OORS_DEBUG
        S *item = *current;
#endif
        (m_pattern = m_patterns[index])->insert (current);
        if (static_cast<ItemPattern<S> *>(m_pattern)->remove())
        {
          m_current = m_available.erase (m_current);
          ++m_skip;
        }
        else
          ++m_current;
        --m_remaining;
        ++m_matched;

        // remember pattern in trace
        m_trace.push_back (m_pattern);

#ifdef OORS_DEBUG
        // inform debugger
        if (m_debugger)
          m_debugger->matchItemPattern (this, index+1, item);
#endif
      }

      /**
       * Finish unfinished wildcard patterns.
       */
      void finish (void)
      {
        for (unsigned int i = 0; i < m_size; ++i)
        {
          Pattern<S> *pattern = m_patterns[i];
          if (dynamic_cast<WildcardPattern<S> *>(pattern) && !pattern->matched())
          {
            pattern->done ();
#ifdef OORS_DEBUG
            if (m_debugger)
              m_debugger->finishWildcardPattern (this, i+1);
#endif
          }
        }
      }

      /**
       * Adjust alternative and given match from source to target block.
       * This function assumes a successfully matched rule.
       * @param source The original block.
       * @param target The new block.
       */
      void adjustTo (const std::list<S *> &source, std::list<S *> &target)
      {
        // first adjust the list of remaining items
        {
          typename std::list<typename std::list<S *>::iterator>::iterator avail = m_available.begin (), avail_end = m_available.end ();
          typename std::list<S *>::iterator item = target.begin ();
          for (typename std::list<S *>::const_iterator i = source.begin(), end = source.end(); i != end; ++i, ++item)
          {
            // nothing more to do
            if (avail == avail_end)
              break;

            // replace the iterator
            if (*i == **avail)
              *(avail++) = item;
          }
        }

        // then adjust the trace
        {
          typename std::list<const Pattern<S> *>::iterator trace = m_trace.begin (), trace_end = m_trace.end ();
          typename std::list<S *>::iterator item = target.begin ();
          for (typename std::list<S *>::const_iterator i = source.begin(), end = source.end(); i != end;)
          {
            // nothing more to do
            if (trace == trace_end)
              break;

            // get item and pattern
            S *current = *i;
            Pattern<S> *pattern = const_cast<Pattern<S> *>(*trace);

            // modify item pattern
            ItemPattern<S> *ipattern = dynamic_cast<ItemPattern<S> *>(pattern);
            if (ipattern)
            {
              // replace the item pattern content
              if (ipattern->item() == current)
              {
                ipattern->replace (item);
                ++trace;
              }
            }

            // modify wildcard pattern
            else
            {
              WildcardPattern<S> *wpattern = static_cast<WildcardPattern<S> *>(pattern);
              unsigned int amount = wpattern->amount ();
              if (amount && (wpattern->items()[0] == current))
              {
                for (unsigned int j = 0; j < amount; ++j)
                {
                  wpattern->replace (j, *(item++));
                  ++i;
                }
                ++trace;
                continue;
              }
            }

            // next
            ++i;
            ++item;
          }
        }
      }

    private:
      /// the list of available items
      std::list<typename std::list<S *>::iterator> m_available;

      /// the current item
      typename std::list<typename std::list<S *>::iterator>::iterator m_current;

      /// amount of items that have not yet been visited
      unsigned int m_remaining;

      /// matched items
      unsigned int m_matched;

      /// items to skip
      unsigned int m_skip;

      /// rule of the alternative
      Rule<S, T> *m_rule;

      /// a list of accessible patterns
      Pattern<S> **m_patterns;

      /// a copy of the rule's search pattern
      SequencePattern<S> *m_sequence;

      /// the current active pattern
      Pattern<S> *m_pattern;

      /// amount of available patterns
      unsigned int m_size;

      /// alternative trace
      std::list<const Pattern<S> *> m_trace;

      /// cost of the alternative
      int m_cost;

#ifdef OORS_DEBUG
      /// the used debugger
      Debugger<S, T> *m_debugger;
#endif
  };

  /**
   * Abstract rule.
   */
  template<class S, class T> class Rule
  {
    public:
      /**
       * Create an empty rule.
       * @param name The rule's name.
       * @param mask The mask.
       * @param has_mask Indicates if the rule has a mask.
       */
      Rule (const std::string &name, int mask, bool has_mask)
        : m_name (name)
        , m_mask (mask)
        , m_has_mask (has_mask)
        , m_pattern (NULL)
        , m_alternative (NULL)
        , m_size (0)
        , m_items (0)
#ifdef OORS_DEBUG
        , m_debugger (NULL)
#endif
      {
      }

      /**
       * Dispose rule.
       */
      virtual ~Rule (void)
      {
        if (m_pattern)
        {
          delete m_pattern;
          m_pattern = NULL;
        }
      }

    public:
      /**
       * Sets the alternative that is to be used.
       * @param alternative New alternative.
       */
      inline void setAlternative (Alternative<S, T> *alternative) { m_alternative = alternative; }

#ifdef OORS_DEBUG
      /**
       * Set the debugger interface.
       * @param debugger The used debugger.
       */
      inline void setDebugger (Debugger<S, T> *debugger) { m_debugger = debugger; }
#endif

    public:
      /**
       * Access the rule's name.
       * @return The rule's name.
       */
      inline const std::string &name (void) const { return m_name; }

      /**
       * Check whether the rule is currently enabled.
       * @param mask The current used mask.
       */
      inline bool enabled (int mask) const { return !m_has_mask || (m_mask & mask); }

      /**
       * Access the current set of alternatives.
       * @return Current set of alternatives.
       */
      inline const std::vector<Alternative<S, T> *> &alternatives (void) const { return m_alternatives; }

      /**
       * Check whether access to item is valid,
       * @param item The accessed item.
       * @param index The index of the item.
       * @param line The line of the original file.
       * @return The item.
       * @throws InvalidAccessException If the accessed pattern is NULL.
       */
      inline T *access (T *item, unsigned int index, unsigned int line) const
      {
        // invalid access detected
        if (!item)
          throw InvalidAccessException (m_name, index, line);

        // all fine
        return item;
      }

    public:
      /**
       * Check rule side condition.
       * By default this will always return true.
       * @return True if the condition is fulfilled.
       */
      virtual bool condition (void) const { return true; }

      /**
       * Compute the cost of this rule.
       * @return Rule cost.
       */
      virtual int cost (void) const = 0;

      /**
       * Fills the target block using the rule's replace function.
       * @param matcher The base pattern matcher that is used if the replace pattern is intermediate.
       * @param target The target block.
       * @param insert True to insert the replace pattern at the first matched item.
       */
      virtual void apply (const BasePatternMatcher &matcher, std::list<T *> &target, bool insert) const = 0;

    public:
      /**
       * Match the given block.
       * @param alternative The initial alternative.
       * @param current Data of the current match.
       * @param result Where to store completely matched alternatives.
       */
      void match (const Alternative<S, T> *alternative, const Match<S> &current, std::vector<Alternative<S, T> *> &result)
      {
        // nothing to do
        if ((alternative?alternative->available():current.items()) < m_items)
          return;

#ifdef OORS_DEBUG
        // inform debugger
        if (m_debugger)
          m_debugger->currentRule (this);
#endif

        // init first alternative
        Alternative<S, T> *initial;
        m_alternatives.push_back (initial = new Alternative<S, T> (alternative, current, this, m_pattern, m_size
#ifdef OORS_DEBUG
                              , m_debugger
#endif
                             ));

        // do the thing
        bool done = false;
        while (!done)
        {
          done = true;
          for (unsigned int alternative = 0, alternatives = m_alternatives.size(); alternative < alternatives; ++alternative)
          {
            // get current alternative and pattern
            m_alternative = m_alternatives[alternative];
            SequencePattern<S> *sequence = m_alternative->sequence ();

            // only match current pattern if some items are available and it is not matched
            if (!m_alternative->done() && !sequence->matched())
            {
#ifdef OORS_DEBUG
              if (m_debugger)
                m_debugger->currentAlternative (m_alternative);
#endif
              if (match(alternative, alternatives))
                done &= sequence->matched();
              done &= m_alternatives.size() == alternatives;
            }
          }
        }

        // clear useless alternatives
        for (unsigned int alternative = 0, alternatives = m_alternatives.size(); alternative < alternatives; ++alternative)
        {
          // get alternative and finish unfinished wildcard patterns
          m_alternative = m_alternatives[alternative];
          m_alternative->finish ();

          // check whether alternative is matched and condition is fulfilled
          if (!m_alternative->sequence()->matched() || !condition())
            clear (alternative, alternatives);
          else
          {
            m_alternative->setCost(cost());
            result.push_back (m_alternative);
          }
        }

#ifdef OORS_DEBUG
        // inform debugger
        if (m_debugger)
          m_debugger->finishRule (this, !m_alternatives.empty());
#endif

        // we are done here
        m_alternatives.clear ();
        m_alternative = NULL;
      }

    protected:
      /**
       * Tests whether a wildcard pattern should be finished.
       * @return True if it no longer makes sense to let a wildcard pattern eat up items.
       */
      inline bool finishWildcard (void) const { return m_alternative->remaining() <= (m_items-m_alternative->matched()); }

    protected:
      /**
       * Check the side condition of pattern with given index.
       * @return True if the patterns side condition is fulfilled, false else.
       */
      virtual bool condition (unsigned int) const { return true; }

    protected:
      /**
       * Match the rule's pattern in the current alternative.
       * @param alternative Alternative index.
       * @param alternatives Amount of alternatives.
       * @return False if the current alternative could not be matched any further.
       */
      bool match (unsigned int &alternative, unsigned int &alternatives)
      {
        // get last active pattern and stay with it if it is an unmatched wildcard pattern
        bool continued = false;
        Pattern<S> *pattern = m_alternative->pattern ();
        if (dynamic_cast<WildcardPattern<S> *>(pattern) && !pattern->matched())
          continued = match (static_cast<WildcardPattern<S> *>(pattern), true);

        // select all possible candidates
        else
        {
          std::list<Pattern<S> *> patterns;
          if (pattern) pattern->follow (patterns);
          else m_alternative->sequence()->first (patterns);

          // select all those item patterns whose condition is true
          // do not dare to touch a wildcard pattern if no item has been matched yet
          // and ensure that the wildcard patterns are processed first
          for (typename std::list<Pattern<S> *>::iterator i = patterns.begin(), end = patterns.end(); i != end;)
          {
            // get pattern
            Pattern<S> *pattern = *i;
            bool erase = true;

            // determine wether it is to be erased or not
            if (dynamic_cast<WildcardPattern<S> *>(pattern)) { if (m_alternative->matched()) patterns.push_front (pattern); }
            else
            {
              // check item pattern side condition
              unsigned int index = pattern->index ();
              erase = !condition (index);

#ifdef OORS_DEBUG
              // inform debugger
              if (m_debugger)
                m_debugger->checkItemPattern (m_alternative, index+1, m_alternative->current(), !erase);
#endif
            }

            // erase that pattern then or leave it
            if (erase) i = patterns.erase (i);
            else ++i;
          }

          // process the valid followers of the last pattern
          for (typename std::list<Pattern<S> *>::iterator i = patterns.begin(), end = patterns.end(); i != end;)
          {
            Pattern<S> *pattern = *i;
            bool last = ++i == end;
            if (dynamic_cast<ItemPattern<S> *>(pattern)) continued |= match (static_cast<ItemPattern<S> *>(pattern), last);
            else continued |= match (static_cast<WildcardPattern<S> *>(pattern), last);
          }
        }

        // there are no possible continuations
        if (!continued)
          clear (alternative, alternatives);

        // done
        return continued;
      }

      /**
       * Match item pattern whose condition is fulfilled.
       * @param pattern The item pattern to match.
       * @param last True if it is the last one that is going to be processed in the current alternative.
       * @return True if the current alternative is to be continued.
       */
      bool match (ItemPattern<S> *pattern, bool last)
      {
        Alternative<S, T> *alternative = m_alternative;
        if (!last)
          m_alternatives.push_back (alternative = new Alternative<S, T> (m_alternative));
        alternative->setItem (pattern->index());
        return last;
      }

      /**
       * Match wildcard pattern.
       * @param pattern The wildcard pattern to match.
       * @param last True if it is the last one that is going to be processed in the current alternative.
       * @return True if the current alternative is to be continued.
       */
      bool match (WildcardPattern<S> *pattern, bool last)
      {
        // no more item patterns have to be matched
        unsigned int index = pattern->index ();
        if (m_alternative->matched() == m_items)
        {
          m_alternative->finishWildcard (index);
          return true;
        }

        // flags
        bool continued = false;
        bool finish = finishWildcard ();

        // do lookahead check
        std::list<Pattern<S> *> patterns;
        pattern->follow (patterns);
        if (!patterns.empty())
        {
          // remove all but the item pattern followers
          bool wildcard = false;
          for (typename std::list<Pattern<S> *>::iterator i = patterns.begin(), end = patterns.end(); i != end;)
          {
            if (dynamic_cast<WildcardPattern<S> *>(*i))
            {
              i = patterns.erase (i);
              wildcard = true;
            }
            else
              ++i;
          }

          // if there is a wildcard pattern successor stop this wildcard in either a new or this alternative
          if (wildcard)
          {
            Alternative<S, T> *alternative = m_alternative;
            if (!last || !finish)
              m_alternatives.push_back (alternative = new Alternative<S, T> (m_alternative));
            alternative->finishWildcard (index);
            continued = true;
          }

          // now handle the item pattern successors
          for (typename std::list<Pattern<S> *>::iterator i = patterns.begin(), end = patterns.end(); i != end;)
          {
            // select current pattern
            Pattern<S> *successor = *i;
            bool last = ++i == end;

            // lookahead at item pattern condition
            unsigned int sindex = successor->index ();
            if (condition(sindex))
            {
              Alternative<S, T> *alternative = m_alternative;
              if (!last || !finish)
                m_alternatives.push_back (alternative = new Alternative<S, T> (m_alternative));
              alternative->finishWildcard (index);
              alternative->setItem (sindex);
              continued = true;
            }
          }
        }

        // eat up item if it makes sense
        if (!finish)
        {
          Alternative<S, T> *alternative = m_alternative;
          if (!last)
            m_alternatives.push_back (alternative = new Alternative<S, T> (m_alternative));
          alternative->setWildcard (index);
          continued = true;
        }

        // done
        return continued;
      }

      /**
       * Clear the current active alternative.
       * @param alternative Index of the current alternative.
       * @param alternatives Current amount of used alternatives.
       */
      void clear (unsigned int &alternative, unsigned int &alternatives)
      {
        delete m_alternative;
        m_alternatives.erase (m_alternatives.begin()+alternative);
        m_alternative = NULL;
        --alternative;
        --alternatives;
      }

      /**
       * Truly remove the matched items from the given block.
       * @param block The block to remove the matched items from.
       */
      void remove (std::list<S *> &block) const
      {
        const std::list<const Pattern<S> *> &trace = m_alternative->trace ();
        for (typename std::list<const Pattern<S> *>::const_iterator i = trace.begin(), end = trace.end(); i != end; ++i)
        {
          const ItemPattern<S> *item = dynamic_cast<const ItemPattern<S> *>(*i);
          if (item)
          {
            if (item->dispose())
              delete item->item();
            if (item->remove())
              block.erase (item->iterator());
          }
        }
      }

    protected:
      /// rule name
      std::string m_name;

      /// the rule's mask
      int m_mask;

      /// indicates if the rule has a mask
      bool m_has_mask;

      /// search pattern
      SequencePattern<S> *m_pattern;

      /// the current active alternative
      Alternative<S, T> *m_alternative;

      /// all available alternatives
      std::vector<Alternative<S, T> *> m_alternatives;

      /// amount of accesible patterns
      unsigned int m_size;

      /// amount of items that are to be matched
      unsigned int m_items;

#ifdef OORS_DEBUG
      /// the used debugger
      Debugger<S, T> *m_debugger;
#endif
  };
}

#endif
