/**
 * clparser.cpp
 * 2005-09-05
 */

#include "clparser.h"

#include <exception.h>

#include <fstream>
#include <sstream>
#include <iostream>

CLParameter::CLParameter (unsigned int id, char s, unsigned int args, const std::string &descr)
  : m_id (id)
  , m_has_short (true)
  , m_short (s)
  , m_has_long (false)
  , m_expected_args (args)
  , m_descr (descr)
{
}

CLParameter::CLParameter (unsigned int id, const std::string &l, unsigned int args, const std::string &descr)
  : m_id (id)
  , m_has_short (false)
  , m_has_long (true)
  , m_long (l)
  , m_expected_args (args)
  , m_descr (descr)
{
}

CLParameter::CLParameter (unsigned int id, char s, const std::string &l, unsigned int args, const std::string &descr)
  : m_id (id)
  , m_has_short (true)
  , m_short (s)
  , m_has_long (true)
  , m_long (l)
  , m_expected_args (args)
  , m_descr (descr)
{
}

CLParameter &CLParameter::operator= (const CLParameter &param)
{
  // don't copy myself
  if (this != &param) {
    m_id            = param.m_id;
    m_has_short     = param.m_has_short;
    m_short         = param.m_short;
    m_has_long      = param.m_has_long;
    m_long          = param.m_long;
    m_expected_args = param.m_expected_args;
    m_descr         = param.m_descr;
    m_args.clear();
  }

  // done
  return *this;
}

std::string CLParameter::flag (void) const
{
  // build flag string
  std::stringstream s ("");
  if (m_has_short)
  {
    s << '-' << m_short;
    if (m_has_long) s << ", --" << m_long;
  }
  else s << "--" << m_long;

  // add expected parameters
  for (unsigned int i = 1; i <= m_expected_args; ++i)
    s << " <" << i << ">";

  // done
  return s.str();
}

void CLParameter::help (unsigned int length) const
{
  // print flag first
  std::string f = flag ();
  std::string space ((f.length() >= length)?0:length-f.length(), ' ');
  std::cout << "  " << f << space << "  " << m_descr << std::endl;
}

CLParser::CLParser (int argc, char **argv, const std::string &info, bool ignore_first)
  : m_index (0)
  , m_arg_index (0)
  , m_param (0, '\0', 0, "")
  , m_flag_length (0)
  , m_info (info)
{
  // first true param has index
  int first = (int)ignore_first;

  // init params from file (@ functionality)
  if ((first < argc) && (argv[first][0] == '@'))
  {
    std::string file = argv[first];
    std::ifstream input (file.substr(1, file.length()-1).c_str());
    if (input) {
      char c;
      std::string arg;
      bool within = false, escape = false;
      while (true)
      {
        input.get (c);
        if ((c == '\r') || (c == '\n') || input.eof())
          break;

        if (escape)
        {
          arg += c;
          escape = false;
        }
        else if (c == '\\')
          escape = true;
        else if (c == '"')
          within = !within;
        else if ((c == ' ') || (c == '\t'))
        {
          if (within)
            arg += c;
            else if (arg.length())
            {
              m_args.push_back (arg);
              arg = "";
            }
        }
        else
          arg += c;
      }
      if (arg.length())
        m_args.push_back (arg);
    }
  }

  // don't load from file
  else
    for (int i = first; i < argc; ++i)
      m_args.push_back (argv[i]);

  // set first argument
  if (m_args.size())
    m_arg = m_args[0];
}

CLParser::~CLParser (void)
{
  for (unsigned int i = 0; i < m_params.size(); ++i)
    delete m_params[i];
}

void CLParser::add (unsigned int id, char s, unsigned int args, const std::string &descr)
{
  // add new parameter
  CLParameter *param = new CLParameter (id, s, args, descr);
  m_params.push_back (param);
  m_short[s] = param;

  // update flag length
  unsigned int length = param->flag().length ();
  if (length > m_flag_length) m_flag_length = length;
}

void CLParser::add (unsigned int id, const std::string &l, unsigned int args, const std::string &descr)
{
  // add new parameter
  CLParameter *param = new CLParameter (id, l, args, descr);
  m_params.push_back (param);
  m_long[l] = param;

  // update flag length
  unsigned int length = param->flag().length ();
  if (length > m_flag_length) m_flag_length = length;
}

void CLParser::add (unsigned int id, char s, const std::string &l, unsigned int args, const std::string &descr)
{
  // add new parameter
  CLParameter *param = new CLParameter (id, s, l, args, descr);
  m_params.push_back (param);
  m_short[s] = param;
  m_long[l] = param;

  // update flag length
  unsigned int length = param->flag().length ();
  if (length > m_flag_length) m_flag_length = length;
}

CLParser::State CLParser::next (void)
{
  // select next argument
  if (m_arg_index >= m_arg.length())
  {
    if (++m_index < m_args.size())
    {
      m_arg = m_args[m_index];
      m_arg_index = 0;
    }
    else
      return done;
  }

  // current parameter
  CLParameter *param = NULL;

  // determine argument type and handle long options (if applicable)
  bool is_short = (m_arg_index != 0);
  if (!is_short)
  {
    // this is no parameter
    if ((m_arg[0] != '-') || (m_arg.length() == 1))
    {
       m_arg_index = m_arg.length();
       return argument;
    }

    // might be a long option
    if (m_arg[1] == '-')
    {
      // in any case: we are done with this argument
      m_arg_index = m_arg.length();

      // treat "--" as: stop parsing
      if (m_arg.length() == 2)
      {
        m_index = m_args.size();
        return done;
      }

      // handle long option
      std::map<std::string, CLParameter *>::iterator i = m_long.find(m_arg.substr(2, m_arg.length()-2));
      if (i != m_long.end()) param = i->second;
      else m_unknown_param = m_arg;
    }

    // k, it's a short option (list)
    else
    {
      is_short = true;
      ++m_arg_index;
    }
  }

  // handle short option
  if (is_short)
  {
    char name = m_arg[m_arg_index++];
    std::map<char, CLParameter *>::iterator i = m_short.find (name);
    if (i != m_short.end()) param = i->second;
    else
    {
      m_unknown_param = "-";
      m_unknown_param += name;
    }
  }

  // unknown parameter
  if (param == NULL)
    return unknown;

  // copy parameter and acquire arguments
  m_param = *param;
  if (m_index+m_param.args() >= m_args.size())
    return missing;
  for (unsigned int i = 0; i < m_param.args(); ++i)
    m_param.add (m_args[++m_index]);

  // done
  return parameter;
}

void CLParser::help (void) const
{
  std::cout << "Command line parameters:" << std::endl;
  for (unsigned int i = 0; i < m_params.size(); ++i)
    m_params[i]->help (m_flag_length);
  std::cout << std::endl << m_info << std::endl;
}
