/**
 * clparser.h
 * 2005-09-05
 */

#ifndef __CLPARSER_H__
#define __CLPARSER_H__

#include <map>
#include <vector>
#include <string>

/**
 * Command line parameter.
 * Represents a command line parameter in all its details.
 */
struct CLParameter
{
  public:
    /**
     * Init parameter w/o long flag.
     * @param id Parameter id.
     * @param s Short parameter flag.
     * @param args Expected arguments.
     * @param descr Parameter description.
     */
    CLParameter (unsigned int id, char s, unsigned int args, const std::string &descr);

    /**
     * Init parameter w/o short flag.
     * @param id Parameter id.
     * @param l Long parameter flag.
     * @param args Expected arguments.
     * @param descr Parameter description.
     */
    CLParameter (unsigned int id, const std::string &l, unsigned int args, const std::string &descr);

    /**
     * Init parameter.
     * @param id Parameter id.
     * @param s Short parameter flag.
     * @param l Long parameter flag.
     * @param args Expected arguments.
     * @param descr Parameter description.
     */
    CLParameter (unsigned int id, char s, const std::string &l, unsigned int args, const std::string &descr);

  public:
    /**
     * Assignment operator.
     * @param param Parameter to copy.
     */
    CLParameter &operator= (const CLParameter &param);

    /**
     * Construct parameter flag string.
     * @return Parameter flag string.
     */
    std::string flag (void) const;

    /**
     * Print parameter help
     * @param length The length of the longest parameter flag. Used for formatting the output.
     */
    void help (unsigned int length = 0) const;

  public:
    /**
     * Add argument.
     * @param arg Argument to add.
     */
    inline void add (const std::string &arg) { m_args.push_back(arg); }

  public:
    /**
     * Access parameter id.
     * @return parameter id.
     */
    inline unsigned int id (void) const { return m_id; }

    /**
     * Access expected argument count.
     * @return Amount of expected arguments.
     */
    inline unsigned int args (void) const { return m_expected_args; }

    /**
     * Access arguments.
     * @param i Argument index.
     * @return Argument value.
     */
    inline const std::string &operator[] (unsigned int i) const { return m_args[i]; }

  private:
    /// parameter id
    unsigned int m_id;

    /// has short flag
    bool m_has_short;

    /// short flag
    char m_short;

    /// has long flag
    bool m_has_long;

    /// long flag
    std::string m_long;

    /// expected arguments
    unsigned int m_expected_args;

    /// parameter arguments
    std::vector<std::string> m_args;

    /// parameter description
    std::string m_descr;
};

/**
 * Command line parser.
 */
struct CLParser
{
  public:
    /**
     * Possible parse states.
     */
    typedef enum
    {
      /// done parsing
      done,

      /// current thing is a simple argument not belonging to any parameter
      argument,

      /// current thing is a parameter (maybe with some arguments)
      parameter,

      /// missing arguments for current given parameter
      missing,

      /// unknown parameter
      unknown
    } State;

  public:
    /**
     * Init command line parser.
     * @param argc Argument count.
     * @param argv Argument list.
     * @param info General information about the command line paramters.
     * @param ignore_first Ignore first argument because it's the program name.
     */
    CLParser (int argc, char **argv, const std::string &info, bool ignore_first = true);

    /**
     * Dispose command line parser.
     */
    ~CLParser (void);

  public:
    /**
     * Add parameter w/o long flag.
     * @param id Parameter id.
     * @param s Short parameter flag.
     * @param args Expected arguments.
     * @param descr Parameter description.
     */
    void add (unsigned int id, char s, unsigned int args, const std::string &descr);

    /**
     * Add parameter w/o short flag.
     * @param id Parameter id.
     * @param l Long parameter flag.
     * @param args Expected arguments.
     * @param descr Parameter description.
     */
    void add (unsigned int id, const std::string &l, unsigned int args, const std::string &descr);

    /**
     * Add parameter.
     * @param id Parameter id.
     * @param s Short parameter flag.
     * @param l Long parameter flag.
     * @param args Expected arguments.
     * @param descr Parameter description.
     */
    void add (unsigned int id, char s, const std::string &l, unsigned int args, const std::string &descr);

    /**
     * Parse next parameter.
     * @return Current parse state.
     */
    State next (void);

    /**
     * Print help.
     */
    void help (void) const;

  public:
    /**
     * Access currently processed argument.
     * The next function must be called at least once, otherwise the result of this function
     * is undefined.
     * @return Current argument.
     */
    inline const std::string &arg (void) const { return m_arg; }

    /**
     * Access current parameter.
     * The next function must be called first. Depending on it's return value the return value
     * of this function may be undefined.
     * @return Currrent parameter.
     */
    inline const CLParameter &param (void) const { return m_param; }

   /**
     * Access current unknown parameter.
     * The return value is undefined if next did not return unknown previously.
     * @return Currrent unknown parameter.
     */
    inline const std::string &unknown_param (void) const { return m_unknown_param; }

   private:
    /// current arguments index
    unsigned int m_index;

    /// current argument index
    unsigned int m_arg_index;

    /// current argument
    std::string m_arg;

    /// current parameter
    CLParameter m_param;

    /// unknown option
    std::string m_unknown_param;

    /// used arguments
    std::vector<std::string> m_args;

    /// known command line parameters
    std::vector<CLParameter *> m_params;

    /// short flag to parameter mapping
    std::map<char, CLParameter *> m_short;

    /// long flag to parameter mapping
    std::map<std::string, CLParameter *> m_long;

    /// length of longest parameter flag
    unsigned int m_flag_length;

    /// general information
    std::string m_info;
};

#endif
