/**
 * indentfilestream.cpp
 * 2005-09-05
 */

#include "indentfilestream.h"
#include "util.h"

#include <exception.h>

#include <sstream>

IndentFileStream::IndentFileStream (const std::string &filename, const std::string &indent)
  : m_filename (sanitize(filename))
  , m_stream (filename.c_str())
  , m_indent (indent)
  , m_stack ("")
  , m_line (1)
  , m_newline (true)
  , m_enabled (true)
{
  // that didn't work
  if (!m_stream)
    throw IOException ("Cannot open '"+filename+"' for writing.");
}

#ifdef _MSC_VER
#pragma warning(push)
#pragma warning(disable:4702)
IndentFileStream::IndentFileStream (const IndentFileStream &)
{
  throw IOException ("Cannot copy indent filestream. Operation not supported.");
}
#pragma warning(pop)
#endif

IndentFileStream::~IndentFileStream (void)
{
  m_stream.close ();
}

#ifdef _MSC_VER
IndentFileStream &IndentFileStream::operator= (const IndentFileStream &)
{
  throw IOException ("Cannot assign indent filestream. Operation not supported.");
}
#endif

IndentFileStream &IndentFileStream::operator<< (const std::string &s)
{
  // last start position of the stream
  unsigned int start = 0, i;

  // split the string into lines
  for (i = 0; i < s.length(); ++i)
  {
    // only indent if not at end of line
    if (s[i] != '\n')
      indent();

    // line break
    if (s[i] == '\n')
    {
      ++m_line;
      m_newline = true;
      m_stream << s.substr(start, i-start+1).c_str ();
      start = i+1;
    }
  }

  // print the rest of the string
  if (start < i)
    m_stream << s.substr(start, i-start).c_str ();

  // done
  return *this;
}

IndentFileStream& IndentFileStream::operator<< (const std::stringstream &s)
{
  return operator<< (s.str());
}

IndentFileStream &IndentFileStream::operator<< (const char *s)
{
  return operator<< (std::string(s));
}

IndentFileStream &IndentFileStream::operator<< (const void *p)
{
  indent ();
  m_stream << p;
  return *this;
}

IndentFileStream &IndentFileStream::operator<< (float f)
{
  indent ();
  m_stream << f;
  return *this;
}

IndentFileStream &IndentFileStream::operator<< (int i)
{
  indent ();
  m_stream << i;
  return *this;
}

IndentFileStream &IndentFileStream::operator<< (unsigned int i)
{
  indent ();
  m_stream << i;
  return *this;
}

IndentFileStream &IndentFileStream::operator<< (char c)
{
  if (c == '\n')
  {
    ++m_line;
    m_newline = true;
  }
  else indent ();
  m_stream << c;
  return *this;
}

IndentFileStream &IndentFileStream::operator<< (unsigned char c)
{
  if (c == '\n')
  {
    ++m_line;
    m_newline = true;
  }
  else indent ();
  m_stream << c;
  return *this;
}

IndentFileStream &IndentFileStream::operator<< (manip ios)
{
  m_stream << ios;
  return *this;
}

IndentFileStream &IndentFileStream::operator<< (omanip ostream)
{
  if (ostream == static_cast<omanip>(std::endl))
  {
    ++m_line;
    m_newline = true;
  }
  m_stream << ostream;
  return *this;
}

IndentFileStream &IndentFileStream::operator<< (ifs::control c)
{
  // execute control command
  switch (c)
  {
    // push indent level
    case ifs::push:
      m_stack += m_indent;
      break;

    // pop indent level
    case ifs::pop:
      if (m_stack.length() >= m_indent.length())
        m_stack = m_stack.substr (0, m_stack.length()-m_indent.length());
      break;

    // en-/disable indenting
    case ifs::enable:
    case ifs::disable:
      m_enabled = (c == ifs::enable);
      break;

    // done
    default:
      break;
  }

  // done
  return *this;
}

void IndentFileStream::indent (void)
{
  // after indentation we are definitly not at a new line
  if (m_newline && m_enabled)
  {
    m_stream << m_stack.c_str ();
    m_newline = false;
  }
}
