/**
 * indentfilestream.h
 * 2005-09-05
 */

#ifndef __INDENTFILESTREAM_H__
#define __INDENTFILESTREAM_H__

#include <string>
#include <fstream>
#include <sstream>

/// indent filestream namespace
namespace ifs
{
  /// control commands for pushing and popping an indent level
  typedef enum
  {
    /// push level onto the stack
    push,

    /// pop level from the stack
    pop,

    /// (re-)enable indenting
    enable,

    /// disable indenting
    disable
  } control;
}

/**
 * An output file stream with indentation support.
 */
class IndentFileStream
{
  public:
#ifdef _MSC_VER
    typedef std::ofstream::_Myios &(*manip)(std::ofstream::_Myios &);
    typedef std::ofstream::_Myos &(*omanip)(std::ofstream::_Myos &);
#else
    typedef std::ios& (*manip)(std::ios&);
    typedef std::ostream& (*omanip)(std::ostream&);
#endif

  public:
    /**
     * Open a new file stream.
     * @param filename The file to open.
     * @param indent How to indent the file. Default are two spaces per indent level.
     * @throws IOException If file cannot be created/opened.
     */
    IndentFileStream (const std::string &filename, const std::string &indent = "  ");

#ifdef _MSC_VER
    /**
     * Copy constructor.
     * @param stream Stream to copy.
     */
    IndentFileStream (const IndentFileStream &stream);
#endif

    /**
     * Close the file stream.
     */
    ~IndentFileStream (void);

  public:
#ifdef _MSC_VER
    /**
     * Assignment operator.
     * @param stream Stream to assign.
     */
    IndentFileStream &operator= (const IndentFileStream &stream);
#endif

    /**
     * Write contents of a string stream.
     * @param s String stream.
     */
    IndentFileStream &operator<< (const std::stringstream &s);

    /**
     * Write a STL-string.
     * @param s String.
     */
    IndentFileStream &operator<< (const std::string &s);

    /**
     * Write a C-string
     * @param s String.
     */
    IndentFileStream &operator<< (const char *s);

    /**
     * Write the address of a pointer.
     * @param p Pointer.
     */
    IndentFileStream &operator<< (const void *p);

    /**
     * Write a float.
     * @param f Float.
     */
    IndentFileStream &operator<< (float f);

    /**
     * Write an integer.
     * @param i Integer.
     */
    IndentFileStream &operator<< (int i);

    /**
     * Write an unsigned integer.
     * @param i Unisgned integer.
     */
    IndentFileStream &operator<< (unsigned int i);

    /**
     * Write a character.
     * @param c Character.
     */
    IndentFileStream &operator<< (char c);

    /**
     * Write an unsigned character.
     * @param c Unsigned character.
     */
    IndentFileStream &operator<< (unsigned char c);

    /**
     * Manipulate stream format.
     * @param ios Format manipulator.
     */
    IndentFileStream &operator<< (manip ios);

    /**
     * Manipulate stream.
     * @param ostream Output stream manipulater (such as std::endl).
     */
    IndentFileStream &operator<< (omanip ostream);

    /**
     * Control the indent file stream.
     * @param c Command that manipulates the stream.
     */
    IndentFileStream &operator<< (ifs::control c);

  public:
    /**
     * Access filename.
     * @return Name of the file.
     */
    inline const std::string &filename (void) const { return m_filename; }

    /**
     * Access the current line number.
     * @return Current line number.
     */
    inline unsigned int line (void) const { return m_line; }

    /**
     * Check if the stream current is at a new line.
     * @return True if the stream is at a new line.
     */
    inline bool newline (void) const { return m_newline; }

    /**
     * Checks if file stream is valid.
     * @return True if it is usable, false else.
     */
    inline bool valid (void) const { return !!m_stream; }

  private:
    /**
     * Indent the current line if at a new line.
     */
    inline void indent (void);

  private:
    /// filename to write into
    std::string m_filename;

    /// output file stream
    std::ofstream m_stream;

    /// file indentation
    std::string m_indent;

    /// indentation stack
    std::string m_stack;

    /// current line number
    unsigned int m_line;

    /// indicates if stream currently resides at a new line
    bool m_newline;

    /// indicates whether indentation is enabled or not
    bool m_enabled;
};

#endif
