/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * code.cpp
 * 2005-09-07
 */

#include "code.h"

#include "rule.h"
#include "parser.h"
#include "pattern.h"

#include <exception.h>
#include <indentfilestream.h>

#include <sstream>
#include <iostream>

#include <cstdlib>

// CODE SEGMENT

Code::Segment::Segment (const std::string &code, bool whitespace)
  : m_code (code)
  , m_whitespace (whitespace)
  , m_pattern (NULL)
  , m_local (false)
{
}

Code::Segment::Segment (const Pattern *pattern, bool local, bool unsafe)
  : m_whitespace (false)
  , m_pattern (pattern)
  , m_local (local)
  , m_unsafe (unsafe)
{
}

void Code::Segment::create (IndentFileStream &cpp) const
{
  // create pattern access
  if (m_pattern)
  {
    if (dynamic_cast<const WildcardPattern *>(m_pattern))
      cpp << "m_alternative->wildcard(" << m_pattern->index() << ")";
    else if (dynamic_cast<const ItemPattern *>(m_pattern))
    {
      // treat pattern within search pattern
      const ItemPattern *item = static_cast<const ItemPattern *>(m_pattern);
      const std::string &type = item->type ();
      bool search = m_pattern->root()->search ();
      bool cast = search || (type != item->base());
      if (cast)
      {
        cpp << "static_cast<";
        if (search)
          cpp << "const ";
        cpp << type << " *>(";
      }

      // access item from search pattern
      if (search)
      {
        if (m_local)
          cpp << "m_alternative->current()";
        else
          cpp << "m_alternative->item(" << m_pattern->index() << ")";
      }

      // access item from replace pattern
      else
      {
        if (m_local)
          cpp << "pattern";
        else
        {
          unsigned int index = m_pattern->index();
          if (Parser::instance()->flag(Parser::unsafe))
            cpp << "patterns[" << index << "]";
          else
            cpp << "access(patterns[" << index << "], " << (index+1) << ", __LINE__)";
        }
      }

      // close cast
      if (cast)
        cpp << ")";
    }
    else
    {
      // access conservated pattern
      const ItemPattern *item = static_cast<const ConservationPattern *>(m_pattern)->item();
      const std::string &type = item->type ();
      bool cast = type != item->base();
      if (cast)
        cpp << "static_cast<" << type << " *>(";
      cpp << "patterns[" << m_pattern->index() << "]";
      if (cast)
        cpp << ")";
    }

    // done
    return;
  }

  // do special indent stuff only if allowed to
  if (Parser::instance()->flag(Parser::indent))
  {
    // start a block
    if (startBlock())
    {
      // prepend newline
      if (!cpp.newline())
        cpp << std::endl;

      // begin block
      cpp << m_code
          << ifs::push << std::endl;

      // done
      return;
    }

    // end a block
    else if (endBlock())
    {
      // prepend newline
      if (!cpp.newline())
        cpp << std::endl;

      // end block
      cpp << ifs::pop
          << m_code
          << std::endl;

      // done
      return;
    }
  }

  // add as normal code
  cpp << m_code;
}

void Code::Segment::print (void) const
{
  if (m_pattern)
  {
    if (dynamic_cast<const WildcardPattern *>(m_pattern))
      std::cout << "*";
    else
      std::cout << static_cast<const ItemPattern *>(m_pattern)->type();
    std::cout << '[' << m_pattern->index() << ']';
  }
  else
    std::cout << m_code;
}

// CODE

Code::Code (const std::string &filename, unsigned int line)
  : m_filename (filename)
  , m_line (line)
  , m_level (0)
{
}

void Code::add (const Segment &segment)
{
  // check whether we really want that segment
  if (Parser::instance()->flag(Parser::indent) && !m_code.empty())
  {
    // handle whitespace
    if (segment.whitespace())
    {
      // reject whitespace at newlines or after block beginngs and endings
      const Segment &last = m_code.back ();
      if (last.newline() || last.block())
        return;
    }

    // handle newline segment
    else if (segment.newline())
    {
      // reject double newlines or newlines after blocks within functions
      const Segment &last = m_code.back ();
      if (last.newline() || (last.block() && m_level))
        return;

      // remove excess whitespace
      while (!m_code.empty() && m_code.back().whitespace())
        m_code.pop_back();
    }

    // handle block start/end
    else if (segment.block())
    {
      // remove excess whitespace and newlines
      while (!m_code.empty() && (m_code.back().whitespace() || m_code.back().newline()))
        m_code.pop_back();

      // adjust level
      m_level += segment.startBlock()?1:-1;
    }
  }

  // append code sequence
  m_code.push_back (segment);
}

void Code::normalise (void)
{
  // find the longest shared whitespace prefix
  unsigned int width = static_cast<unsigned int>(-1);
  bool newline = true;
  for (unsigned int i = 0, size = m_code.size(); i < size; ++i)
  {
    // reset newline flag when encountering a newline segment
    Segment segment = m_code[i];
    if (segment.newline())
    {
      newline = true;
      continue;
    }

    // nothing to do right now
    if (!newline)
      continue;

    // determine the greatest common width of the initial whitespace of each line
    if (segment.whitespace())
    {
      // check whether that whitespace section is smaller
      unsigned int size = segment.size ();
      if (size < width)
        width = size;
    }

    // if a line does not start with whitespace there is no shared width
    else
    {
      width = 0;
      break;
    }

    // we are within a line
    newline = false;
  }

  // now reduce the heading whitespace and remove trailing whitespace
  newline = true;
  for (std::vector<Segment>::iterator i = m_code.begin(); i != m_code.end();)
  {
    // at newline
    if (i->newline())
    {
      newline = true;
      ++i;
      continue;
    }

    // nothing to do right now
    if (!i->whitespace())
    {
      ++i;
      continue;
    }

    // completely remove this whitespace block if the next
    // segment is a newline or the code sequence ends
    std::vector<Segment>::iterator next = i+1;
    if ((next == m_code.end()) || next->newline())
      i = m_code.erase (i);

    // truncate or remove that whitespace block is possible
    else
    {
      if (width && newline)
      {
        // remove whitespace block if no more whitespace chars remain
        unsigned int size = i->size()-width;
        if (!size)
          i = m_code.erase (i);

        // reduce the size of the whitespace block
        else
          i++->setCode (std::string(size, ' '));
      }
      else
        ++i;
    }

    // done, next
    newline = false;
  }

  // remove heading newlines and correctly adjust the line number
  while (!m_code.empty() && m_code.front().newline())
  {
    m_code.erase (m_code.begin());
    ++m_line;
  }

  // remove trailing newlines
  while (!m_code.empty() && m_code.back().newline())
    m_code.pop_back ();
}

void Code::create (IndentFileStream &cpp, const std::string &prefix, const std::string &suffix) const
{
  // prepend newline before preprocessor directives
  if (!cpp.newline())
    cpp << std::endl;

  // add compiler directive
  cpp << ifs::disable << "#line " << m_line << " \"" << m_filename << '"' << ifs::enable << std::endl;

  // insert code
  cpp << prefix;
  for (unsigned int i = 0, size = m_code.size(); i < size; ++i)
    m_code[i].create (cpp);
  cpp << suffix;

  // prepend newline before preprocessor directive
  if (!cpp.newline())
    cpp << std::endl;

  // add compiler directive
  cpp << ifs::disable << "#line " << cpp.line()+1 << " \"" << cpp.filename() << '"' << ifs::enable << std::endl;
}

void Code::print (void) const
{
  for (unsigned int i = 0; i < m_code.size(); ++i)
    m_code[i].print ();
}
