/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * code.h
 * 2005-09-07
 */

#ifndef __CODE_H__
#define __CODE_H__

#include <string>
#include <vector>

// forward declarations
class Rule;
class Pattern;
class IndentFileStream;

/**
 * Encapsuled C/C++ code enriched with pattern accessors.
 */
class Code
{
  public:
    /**
     * A code segment, which is either code or a pattern accessor.
     */
    class Segment
    {
      public:
        /**
         * Init code segment.
         * @param code Pure C/C++ code.
         * @param whitespace Indicates whether the code segment is just whitespace.
         */
       Segment (const std::string &code, bool whitespace = false);

       /**
        * Init pattern accessor.
        * @param pattern Accessed pattern.
        * @param local Access to current active pattern.
        * @param unsafe Whether the access is unsafe.
        */
       Segment (const Pattern *pattern, bool local, bool unsafe);

      public:
        /**
         * Create code.
         * @param cpp Code file stream.
         */
        void create (IndentFileStream &cpp) const;

        /**
         * Print the code segment.
         */
        void print (void) const;

      public:
        /**
         * Modify the C/C++ code segent.
         * @param code New code segment.
         */
        inline void setCode (const std::string &code) { m_code = code; }

      public:
        /**
         * Get C/C++ code segment.
         * @return Encapsulated C/C++ code segment.
         */
        inline const std::string &code (void) const { return m_code; }

        /**
         * Access the size of the code section.
         * @return The code section size.
         */
        inline unsigned int size (void) const { return m_code.size (); }

        /**
         * Check whether the code segment is just whitespace.
         * @return True if it is just whitespace.
         */
        inline bool whitespace (void) const { return m_whitespace; }

        /**
         * Check whether the code segment is just a newline.
         * @return True if the code segment is a newline.
         */
        inline bool newline (void) const { return m_code == "\n"; }

        /**
         * Check whether the code segment represents the begin or end of a block.
         * @return True if the code segment is the begin or end of a block.
         */
        inline bool block (void) const { return startBlock() || endBlock(); }

        /**
         * Check whether the code segment starts a block.
         * @return True if the code segment is the beginning of a block.
         */
        inline bool startBlock (void) const { return m_code == "{"; }

        /**
         * Check whether the code segment ends a block.
         * @return True if the code segment is the ending of a block.
         */
        inline bool endBlock (void) const { return m_code == "}"; }

        /**
         * Get accessed pattern.
         * @return Accessed pattern.
         */
        inline const Pattern *pattern (void) const { return m_pattern; }

        /**
         * Pattern is the current active one?
         * @return True if the accessed pattern is the current one.
         */
        inline bool local (void) const { return m_local; }

        /**
         * Pattern access is unsafe?
         * @return True if accessing the pattern may end in a segmentation fault.
         */
        inline bool unsafe (void) const { return m_unsafe; }

      private:
        /// pure C/C++ code segment
        std::string m_code;

        /// is it a whitespace code segment
        bool m_whitespace;

        /// pattern accessor
        const Pattern *m_pattern;

        /// local pattern
        bool m_local;

        /// unsafe pattern access
        bool m_unsafe;
    };

  public:
    /**
     * Create new code wrapper.
     * @param filename The file in which this code segment has been specified.
     * @param line The original line number.
     */
    Code (const std::string &filename, unsigned int line);

  public:
    /**
     * Add code segment.
     * @param segment The segment to add.
     */
    void add (const Segment &segment);

    /**
     * Normalise the code sequence.
     * This function works like an equaliser that reduces the initial whitespace of a line to a minimum.
     * Note that this function does not care about tabs. Currently only the number of characters are
     * taken into account.
     */
    void normalise (void);

  public:
    /**
     * Create code.
     * @param cpp Code file stream.
     * @param prefix Prepend before code.
     * @param suffix Append after code.
     */
    void create (IndentFileStream &cpp, const std::string &prefix = "", const std::string &suffix = "") const;

    /**
     * Print the encapsulated code.
     */
    void print (void) const;

  public:
    /**
     * Access filename.
     * @return The name of the file this code sequence comes from.
     */
    inline const std::string &filename (void) const { return m_filename; }

    /**
     * Access line number.
     * @return The line number this code sequence has begun.
     */
    inline unsigned int line (void) const { return m_line; }

    /**
     * Access the code segments.
     * @return Stored code segments.
     */
    inline const std::vector<Segment> &code (void) const { return m_code; }

  private:
    /// the file in which the code sequence was specified
    std::string m_filename;

    /// the line number of the originating code sequence
    unsigned int m_line;

    /// the block level
    unsigned int m_level;

    /// encapsulated code
    std::vector<Segment> m_code;
};

#endif
