/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * main.cpp
 * 2005-09-01
 */

#include "parser.h"
#include "ruleset.h"

#include <util.h>
#include <clparser.h>

#include <string>
#include <vector>
#include <iostream>

/**
 * Commandline parameters.
 */
typedef enum
{
  /// for defining variables
  define,

  /// add include path
  include,

  /// issue warnings
  warn,

  /// make parser pedantic
  pedantic,

  /// enable verbose mode
  verbose,

  /// allow unsafe pattern access
  unsafe,

  /// toggle automatic indentation
  indent,

  /// display version
  version,

  /// display help
  help
} Parameter;

/**
 * Handle command line parameters.
 * Terminates the program in some special cases.
 * @param argc Argument count.
 * @param argv Argument list.
 * @param files Insert files to be processed here.
 */
static void init (int argc, char **argv, std::vector<std::string> &files)
{
  // init command line parser
  CLParser commandline (argc, argv, "Any non-parameter argument is intepreted as file that is to be compiled.");
  commandline.add (define,   'D',             1, "Define a variable.");
  commandline.add (include,  'I',             1, "Add include path.");
  commandline.add (warn,     'W',             0, "Issue warnings.");
  commandline.add (pedantic, 'P', "pedantic", 0, "Pedantic mode.");
  commandline.add (verbose,  'V', "verbose",  0, "Verbose mode.");
  commandline.add (unsafe,        "funsafe",  0, "Allow unsafe pattern access.");
  commandline.add (indent,        "indent",   0, "Enable automatic indentation.");
  commandline.add (version,  'v', "version",  0, "Print version information.");
  commandline.add (help,     'h', "help",     0, "Print this help.");

  // handle command line arguments
  Parser *parser = Parser::instance ();
  CLParser::State state;
  while ((state = commandline.next()) != CLParser::done)
    switch (state)
    {
      // parameter
      case CLParser::parameter:
      {
        const CLParameter &param = commandline.param();
        switch (param.id())
        {
          case define   : parser->define(param[0]); break;
          case include  : parser->add(param[0]); break;
          case warn     : parser->setFlag(Parser::warn, true); break;
          case pedantic : parser->setFlag(Parser::pedantic, true); break;
          case verbose  : parser->setFlag(Parser::verbose, true); break;
          case unsafe   : parser->setFlag(Parser::unsafe, true); break;
          case indent   : parser->setFlag(Parser::indent, true); break;
          case version  : std::cout << "oorg v1.1.0" << std::endl; exit(0);
          case help     : commandline.help(); exit(0);
          default: break;
        }
        break;
      }

      // simple argument
      case CLParser::argument:
        files.push_back (commandline.arg());
        break;

      // missing paramter
      case CLParser::missing:
        std::cerr << "Missing argument for parameter: '"+commandline.param().flag()+"'." << std::endl;
        commandline.help ();
        exit (1);

      // unknown parameter
      case CLParser::unknown:
        std::cerr << "Unknown command line parameter: '"+commandline.unknown_param()+"'." << std::endl;
        commandline.help ();
        exit (1);

      // never happens
      default:
        break;
    }
}

/**
 * Process the given input files.
 * @param files Input files.
 * @return Process status.
 */
static int process (const std::vector<std::string> &files)
{
  // no input files specified
  unsigned int size = files.size ();
  if (!size)
    std::cerr << "ERROR: No input files specified." << std::endl;

  // process given files
  for (unsigned int i = 0; i < size; ++i)
  {
    // process file
    std::string file = files[i];
    RuleSet *ruleset = NULL;
    try
    {
      // get parser instance
      Parser *parser = Parser::instance ();
      bool verbose = parser->flag (Parser::verbose);

      // parse ruleset
      if (verbose)
        std::cout << "Processing '"+file+"'..." << std::endl;
      ruleset = parser->parse (file);
      if (verbose)
        ruleset->print ();

      // generate pattern matcher
      if (verbose)
        std::cout << "Generating pattern matcher..." << std::endl;
      ruleset->create (prefix(file));

      // done
      if (verbose)
        std::cout << "Done." << std::endl;
      parser->clear ();
      delete ruleset;
    }

    // something went wrong
    catch (Exception &exception)
    {
      // print error message
      std::cerr << "ERROR: " << exception.msg() << std::endl;

      // cleanup
      delete Parser::instance ();
      if (ruleset) delete ruleset;
      return 2;
    }
  }

  // all went fine
  delete Parser::instance ();
  return 0;
}

/**
 * OORG main routine.
 * @param argc Argument count.
 * @param argv Argument list.
 * @return OORG status.
 */
int main (int argc, char **argv)
{
  std::vector<std::string> files;
  init (argc, argv, files);
  return process (files);
}
