/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * parser.cpp
 * 2005-09-01
 */

#include "parser.h"

#include "rule.h"
#include "profile.h"
#include "ruleset.h"
#include "pattern.h"

#include <util.h>

#include <sstream>
#include <iostream>

// singleton instance
Parser *Parser::m_instance = NULL;

// forward declaration of parse function
int yyparse (void);
yy_buffer_state *yy_create_buffer (FILE *, int);
void yy_switch_to_buffer (yy_buffer_state *);
void yy_delete_buffer (yy_buffer_state *);

// PARSER EXCEPTION

Parser::Exception::Exception (const std::string &msg, bool extend, const std::string &filename, unsigned int line)
{
  if (extend)
  {
    std::stringstream s ("");
    if (!line)
    {
      State *state = Parser::instance()->current ();
      if (state)
        s << state->filename() << ":" << state->line() << ": ";
    }
    else
      s << filename << ":" << line << ": ";
    s << "parser: " << msg;
    m_msg = s.str ();
  }
  else
    m_msg = msg;
}

// PARSER STATE

Parser::State::State (const std::string &filename, bool first)
  : m_state (NULL)
  , m_file (NULL)
  , m_filename (sanitize(filename))
  , m_path (extractPath(filename))
  , m_line (1)
{
  // open file
  if (!(m_file = fopen (filename.c_str(), "r")))
    if (first)
      throw IOException ("Cannot open '"+filename+"'.");
    else
      throw Exception ("Cannot open '"+filename+"'.");

  // create and switch to buffer
  yy_switch_to_buffer (m_state = yy_create_buffer(m_file, 16384));
}

Parser::State::~State (void)
{
  yy_delete_buffer (m_state);
  fclose (m_file);
}

void Parser::State::activate (void)
{
  yy_switch_to_buffer (m_state);
}

// PARSER

Parser::Parser (void)
  : m_flags (0)
  , m_skips (0)
  , m_line (1)
  , m_ruleset (NULL)
  , m_profile (NULL)
  , m_rule (NULL)
  , m_pattern (NULL)
{
}

Parser::~Parser (void)
{
  // clear instance
  clear ();
  m_instance = NULL;
}

RuleSet *Parser::parse (const std::string &filename)
{
  // parse from given file handle
  try
  {
    include (filename, true);
    yyparse ();
  }
  catch (Exception exception)
  {
    // walk through states
    if (m_states.size() > 1)
    {
      std::stringstream msg ("");
      for (unsigned int i = 0; i < m_states.size()-1; ++i)
      {
        State *state = m_states[i];
        if (i == 0) msg << "In file included from ";
        else msg << "                 from ";
        msg << state->filename() << ":" << state->line() << ((i < m_states.size()-2)?',':':') << std::endl;
      }
      msg << exception.msg ();
      throw Exception (msg.str(), false);
    }

    // keep message as it is
    throw exception;
  }

  // done
  RuleSet *ruleset = m_ruleset;
  m_ruleset = NULL;
  return ruleset;
}

bool Parser::beginIf (bool skip)
{
  m_ifs.push (skip);
  return m_skips += skip;
}

bool Parser::beginElse (void)
{
  // cannot begin else
  if (m_ifs.empty())
    throw Exception ("Found #else without #if.");

  // do the thing
  m_skips -= m_ifs.top();
  return m_skips += (m_ifs.top() = !m_ifs.top());
}

bool Parser::endIf (void)
{
  // cannot end if
  if (m_ifs.empty())
    throw Exception ("Found #endif without #if.");

  // do the thing
  bool skip = m_skips -= m_ifs.top();
  m_ifs.pop ();
  return skip;
}

void Parser::warning (const std::string &warning, bool extend, const std::string &filename, unsigned int line)
{
  if (flag(warn))
  {
    if (extend)
    {
      if (!line)
      {
        State *state = Parser::instance()->current ();
        if (state)
          std::cerr << state->filename() << ":" << state->line() << ": ";
      }
      else
        std::cerr << filename << ":" << line << ": ";
    }
    std::cerr << "warning: " << warning << std::endl;
  }
}

void Parser::define (const std::string &def)
{
  // do not add twice
  if (defined(def))
    return;

  // add
  m_defines.insert (def);
}

void Parser::undefine (const std::string &def)
{
  std::set<std::string>::iterator i = m_defines.find (def);
  if (i != m_defines.end())
    m_defines.erase (i);
}

void Parser::add (const std::string &path)
{
  // empty paths are invalid
  unsigned int length = path.length();
  if (!length)
    return;

  // add trailing '/' if missing
  if (path[length-1] != '/')
    m_paths.push_back (path+'/');
  else
    m_paths.push_back (path);
}

void Parser::include (const std::string &filename, bool local)
{
  // open file local to current dir
  if (local)
    try
    {
      State *state = current ();
      if (state)
        m_states.push_back (new State(state->path()+filename, false));
      else
        m_states.push_back (new State (filename, true));
      return;
    }
    catch (Exception exception)
    {
    }

  // open file using the include dirs
  for (unsigned int i = 0; i < m_paths.size(); ++i)
    try
    {
      m_states.push_back (new State (m_paths[i]+filename, m_states.empty()));
      return;
    }
    catch (Exception exception)
    {
    }

  // no way
  throw Exception ("Cannot open '"+filename+"'.");
}

bool Parser::finish (void)
{
  // kill topmost state (if any)
  if (!m_states.empty())
  {
    // kill old state
    delete m_states.back ();
    m_states.pop_back ();

    // switch to previous state
    if (!m_states.empty())
    {
      m_states.back()->activate ();
      return true;
    }
  }

  // no more states
  return false;
}

void Parser::clear (void)
{
  // clear unused obects
  if (m_rule) delete m_rule;
  if (m_ruleset) delete m_ruleset;

  // clear states
  while (!m_states.empty())
  {
    delete m_states.back ();
    m_states.pop_back ();
  }
}
