/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * parser.h
 * 2005-09-01
 */

#ifndef __PARSER_H__
#define __PARSER_H__

#include <exception.h>

#include <string>
#include <vector>
#include <stack>
#include <set>

// forward declarations
class Code;
class Rule;
class Pattern;
class Profile;
class RuleSet;
class SequencePattern;
struct yy_buffer_state;

/**
 * Tree-pattern matcher generator ruleset parser.
 */
class Parser
{
  public:
    /**
     * Parser exception.
     */
    class Exception : public ::Exception
    {
      public:
        /**
         * Constructor.
         * @param msg Exception message.
         * @param extend Extend message by filename and line number.
         * @param filename Override filenmae.
         * @param line Override line number.
         */
        Exception (const std::string &msg, bool extend = true, const std::string &filename = "", unsigned int line = 0);
    };

    /**
     * Parser state.
     */
    class State
    {
      public:
        /**
         * Create new state and switch to state.
         * @param filename Name of the file to open.
         * @param first Indicates whether this is the first file to be opened.
         */
        State (const std::string &filename, bool first);

        /**
         * Dispose state.
         */
        ~State (void);

      public:
        /**
         * Access the name of the current processed file.
         * @return Name of the current file.
         */
        inline const std::string &filename (void) const { return m_filename; }

        /**
         * Access the path where the current file resides.
         * @return Path that contains the current file.
         */
        inline const std::string &path (void) const { return m_path; }

        /**
         * Access the current line number.
         * @return Number of the current line.
         */
        inline unsigned int &line (void) { return m_line; }

      public:
        /**
         * Activate buffer state.
         */
        void activate (void);

       private:
         /// current state
         yy_buffer_state *m_state;

         /// used file
         FILE *m_file;

         /// current filename
         std::string m_filename;

         /// path
         std::string m_path;

         /// current line number
         unsigned int m_line;
    };

  public:
    /**
     * Parser flags.
     */
    typedef enum
    {
      // OPTIONS

      /// issue warnings
      warn = 1,

      /// be pedantic
      pedantic = 2,

      /// be verbose
      verbose = 4,

      /// allow unsafe pattern access
      unsafe = 8,

      /// enable automatic indentation
      indent = 16,

      // CONTROL FLAGS

      /// query define or undefine
      ifdef = 32,

      /// un-/define something
      def = 64,

      /// expect footer section (means: %% ... %%)
      footer = 128,

      /// expect function (means: { ... })
      function = 256,

      /// expect expression (means: ( ... ))
      expression = 512,

      /// within search pattern
      search = 1024,

      /// within intermediate pattern
      intermediate = 2048,

      /// within else branch
      else_branch = 4096,

      /// handle pattern accesses
      access = 8192,

      /// access local pattern
      access_local = 16384,

      /// access search pattern
      access_search = 32768,

      /// indicates that there is a current pattern
      has_current = 65536
    } Flag;

  public:
    /**
     * Access parser instance.
     * @return Parser instance.
     */
    inline static Parser *instance (void) { return m_instance?m_instance:m_instance = new Parser (); }

  public:
    /**
     * Dispose parser.
     */
    ~Parser (void);

  public:
    /**
     * Set flag.
     * @param flag The flag to set.
     * @param state The new state of the flag.
     */
    inline void setFlag (Flag flag, bool state) { state?m_flags |= flag:m_flags &= ~flag; }

    /**
     * Access remembered line number.
     * @return Remembered line number.
     */
    inline unsigned int &line (void) { return m_line; }

    /**
     * Access the current ruleset.
     * @return Current ruleset.
     */
    inline RuleSet* &ruleset (void) { return m_ruleset; }

    /**
     * Access the current profile.
     * @return Current profile.
     */
    inline Profile* &profile (void) { return m_profile; }

    /**
     * Access the current rule.
     * @return Current rule.
     */
    inline Rule* &rule (void) { return m_rule; }

    /**
     * Access the remembered pattern.
     * @return Remebered pattern.
     */
    inline Pattern* &pattern (void) { return m_pattern; }

    /**
     * Push sequence pattern.
     * @param sequence Current sequence pattern.
     */
    inline void pushSequence (SequencePattern *sequence) { m_sequences.push (sequence); }

    /**
     * Get current sequence pattern.
     * @return Current sequence pattern or NULL if there is none.
     */
    inline SequencePattern *sequence (void) { return m_sequences.empty()?NULL:m_sequences.top(); }

    /**
     * Pop sequence pattern.
     */
    inline void popSequence (void) { if (!m_sequences.empty()) m_sequences.pop(); }

  public:
    /**
     * Get given flag state.
     * @param flag The flag to look for.
     * @return True if flag is enabled.
     */
    inline bool flag (Flag flag) const { return m_flags & flag; }

    /**
     * Get amount of unmatched ifs.
     * @return Amount of unmatched ifs.
     */
    inline unsigned int ifs (void) const { return m_ifs.size(); }

    /**
     * Check whether given string is defined or not.
     * @param def String to check.
     * @return True if string is defined.
     */
    inline bool defined (const std::string &def) const { return m_defines.find(def) != m_defines.end(); }

    /**
     * Access the current state.
     * @return The current state or NULL if not existing.
     */
    inline State *current (void) const { return m_states.empty()?NULL:m_states.back(); }

  public:
    /**
     * Parse given file.
     * @param filename Name of the file to parse.
     * @return The parsed ruleset.
     */
    RuleSet *parse (const std::string &filename);

    /**
     * Begin if.
     * @param skip True if everything within if and endif should be ignored.
     * @return True if scanner should enter skip mode.
     */
    bool beginIf (bool skip);

    /**
     * Begin else.
     * @return True if scanner should enter skip mode.
     */
    bool beginElse (void);

    /**
     * End if.
     * @return True if scanner should enter skip mode.
     */
    bool endIf (void);

    /**
     * Report warning.
     * It will not be displayed if warnings are disabled.
     * @param warning The warning to print.
     * @param extend Extend message by filename and line number.
     * @param filename Override filenmae.
     * @param line Override line number.
     */
    void warning (const std::string &warning, bool extend = true, const std::string &filename = "", unsigned int line = 0);

    /**
     * Define something.
     * @param def String to define.
     */
    void define (const std::string &def);

    /**
     * Undefine something.
     * @param def String to undefine.
     */
    void undefine (const std::string &def);

    /**
     * Add include path.
     * @param path Include path to add.
     */
    void add (const std::string &path);

    /**
     * Include the given file.
     * A file will not be opened twice.
     * @param filename Name of the file to include.
     * @param local True if file should be opened local to work dir.
     */
    void include (const std::string &filename, bool local);

    /**
     * Finish the current file.
     * @return True if there is an additional file that is still to be parsed.
     */
    bool finish (void);

    /**
     * Clear parser instance to make it reuseable.
     */
    void clear (void);

  protected:
    /**
     * Protected constructor.
     * Allows only one instance.
     */
    Parser (void);

  private:
    /// parser instance
    static Parser *m_instance;

    /// parser flags
    unsigned int m_flags;

    /// amount of skips
    unsigned int m_skips;

    /// remembered line number
    unsigned int m_line;

    /// if stack
    std::stack<bool> m_ifs;

    /// all known include paths
    std::vector<std::string> m_paths;

    /// all known defines
    std::set<std::string> m_defines;

    /// parser states
    std::vector<State *> m_states;

    /// current ruleset
    RuleSet *m_ruleset;

    /// current profile
    Profile *m_profile;

    /// current rule
    Rule *m_rule;

    /// current pattern
    Pattern *m_pattern;

    /// sequence pattern stack
    std::stack<SequencePattern *> m_sequences;
};

#endif
