/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * pattern.cpp
 * 2005-09-05
 */

#include "pattern.h"

#include "code.h"
#include "rule.h"
#include "parser.h"
#include "profile.h"
#include "ruleset.h"

#include <exception.h>
#include <indentfilestream.h>

#include <iostream>
#include <sstream>

// ABSTRACT PATTERN

Pattern::Pattern (const std::string &filename, unsigned int line, const Rule *rule, SequencePattern *parent)
  : m_filename (filename)
  , m_line (line)
  , m_rule (rule)
  , m_root (parent?parent->m_root:NULL)
  , m_parent (parent)
  , m_index (0)
{
  if (m_root)
    m_index = m_root->size ();
}

Pattern::~Pattern (void)
{
}

// SEQUENCE PATTERN

SequencePattern::SequencePattern (const std::string &filename, unsigned int line, Rule *rule, SequencePattern *parent, bool ordered, bool search, bool intermediate)
  : Pattern (filename, line, rule, parent)
  , m_condition (NULL)
  , m_items (0)
  , m_ordered (ordered)
  , m_search (search)
  , m_intermediate (intermediate)
  , m_wildcard (false)
  , m_else (false)
{
  // this is an internal error
  if (search && intermediate)
    throw Exception ("A search pattern cannot be intermediate.");

  // if there is no root then we are the root
  if (!m_root)
    m_root = this;

  // set rule's search/replace pattern
  if (search)
  {
    if (!rule->search())
      rule->setSearch (this);
  }
  else if (!rule->replace())
    rule->setReplace (this);

  // insert into parent pattern
  if (parent)
    parent->insert (this, false);
}

SequencePattern::~SequencePattern (void)
{
  // dispose sequence
  while (!m_sequence.empty())
  {
    delete m_sequence.back ();
    m_sequence.pop_back ();
  }

  // dispose condition
  if (m_condition)
    delete m_condition;
}

void SequencePattern::insert (Pattern *pattern, unsigned int index, bool accessible)
{
  // check if inserted pattern is redundant
  if (dynamic_cast<WildcardPattern *>(pattern))
  {
    // check for redundant wildcard patterns in ordered sequence
    bool remove = false;
    if (m_ordered)
      remove = !pattern->index() || ((index > 0) && (index <= m_sequence.size()) && dynamic_cast<WildcardPattern *>(m_sequence[index-1]));

    // only one wildcard pattern per unordered sequence is permitted
    else
      remove = m_wildcard || !(m_wildcard = true);

    // delete the pattern to prevent memleaks and raise exception
    if (remove)
    {
      std::stringstream msg ("");
      if (m_ordered && !pattern->index())
        msg << "A search pattern must not begin with a wildcard pattern.";
      else
        msg << "Duplicate wildcard pattern at position " << (pattern->index()+1) << " detected.";
      delete pattern;
      throw Parser::Exception (msg.str());
    }
  }

  // note that one more item has to be matched
  else if (accessible && dynamic_cast<ItemPattern *>(pattern))
  {
    SequencePattern *sequence = this;
    while (sequence)
    {
      ++sequence->m_items;
      sequence = sequence->m_parent;
    }
  }

  // simply insert it at the given index
  pattern->setParent (this);
  m_sequence.insert (m_sequence.begin()+index, pattern);

  // insert it into root pattern
  if (accessible)
    m_root->m_patterns.push_back (pattern);
}

void SequencePattern::normalise (void)
{
  // this baby isn't a set if it contains at most one item
  if (m_sequence.size() <= 1)
    m_ordered = true;

  // search for list patterns
  for (unsigned int i = 0; i < m_sequence.size(); ++i)
  {
    // only sequences are interesting
    SequencePattern *pattern = dynamic_cast<SequencePattern *>(m_sequence[i]);
    if (!pattern || pattern->m_condition || pattern->m_else)
      continue;

    // sequence within a sequence or only one or no item(s) in list
    std::vector<Pattern *> &sequence = pattern->m_sequence;
    if ((m_ordered && pattern->m_ordered) || (sequence.size() <= 1) || (m_sequence.size() == 1))
    {
      // make inner sequence to outer sequence
      if (m_sequence.size() == 1)
        m_ordered = pattern->m_ordered;

      // remove the sequence and insert contents beginning at sequence position
      m_sequence.erase (m_sequence.begin()+i);
      unsigned int j = 0;
      for (std::vector<Pattern *>::iterator k = sequence.begin(); k != sequence.end(); ++j)
        try
        {
          Pattern *current = *k;
          k = sequence.erase (k);
          insert (current, i+j, false);
        }
        catch (Parser::Exception exception)
        {
          delete pattern;
          throw exception;
        }

      // dispose sequence
      i += j-1;
      delete pattern;
    }
  }
}

void SequencePattern::createInit (IndentFileStream &cpp, const std::string &name, const std::string &parent) const
{
  // create new sequence/set pattern
  const std::string &type = m_rule->profile()->ruleset()->source();
  cpp << "oors::SequencePattern<" << type << "> *" << name << " = new oors::SequencePattern<" << type
      << "> (" << parent << ", " << m_index << ", " << static_cast<unsigned int>(m_sequence.size()) << ", "
      << (m_ordered?"true":"false") << ");"
      << std::endl;

  // insert patterns
  for (unsigned int i = 0, size = m_sequence.size(); i < size; ++i)
  {
    Pattern *pattern = m_sequence[i];
    if (dynamic_cast<SequencePattern *>(pattern))
    {
      pattern->createInit (cpp, name+"_", name);
      cpp << name << "->insert (" << name << "_, " << i << ");" << std::endl;
    }
    else
    {
      cpp << name << "->insert (";
      pattern->createInit (cpp, name, name);
      cpp << ", " << i << ");" << std::endl;
    }
  }
}

void SequencePattern::createReplace (IndentFileStream &cpp, bool recursive, bool intermediate) const
{
  // bail out if generating replacement code for unordered sequences
  if (!m_ordered)
    throw Exception ("Generating replacement code for an unordered sequence.");

  // this is the root so init stuff here
  unsigned int size = m_patterns.size ();
  if (size)
  {
    // alloc pattern cache
    const RuleSet *ruleset = m_rule->profile()->ruleset();
    const std::string &source = ruleset->source(), &target = ruleset->target();
    cpp << (intermediate?source:target) << " *patterns[" << size << "];" << std::endl;
    cpp << "memset (patterns, 0, " << size << "*sizeof(" << (intermediate?source:target) << " *));" << std::endl;

    // generate code for recursive mode
    if (recursive)
    {
      // check if really operating in recursive mode
      cpp << "if (insert)" << std::endl
          << "{" << ifs::push << std::endl;

      // allocate blocks for intermediate replace pattern
      if (intermediate)
        cpp << "std::list<" << source << "*> local_source;" << std::endl;

      // allocate insert interator
      cpp << "std::list<" << source << " *>::iterator before = static_cast<const oors::ItemPattern<" << source << "> *>(*m_alternative->trace().begin())->iterator()";
      if (intermediate)
        cpp << ", local_before = local_source.end()";
      cpp << ";" << std::endl;

      // insert items
      for (unsigned int i = 0, size = m_sequence.size(); i < size; ++i)
        m_sequence[i]->createReplace (cpp, true, intermediate);

      // do intermediate step
      if (intermediate)
        cpp << "PatternMatcher (matcher).process (local_source);" << std::endl
            << "target.insert (static_cast<const oors::ItemPattern<" << source << "> *>(*m_alternative->trace().begin())->iterator(), local_source.begin(), local_source.end());" << std::endl;

      // do cleanup
      cpp << "remove (target);" << std::endl
          << "implicitPostpass (target);" << std::endl
          << ifs::pop << "}" << std::endl
          << "else" << std::endl
          << "{" << ifs::push << std::endl;
    }

    // generate code for non-recursive mode
    if (intermediate)
    {
      cpp << "std::list<" << source << " *> local_source, local_purge";
      if (source == target) cpp << ", local_target;" << std::endl;
      else cpp << ";" << std::endl << "std::list<" << target << " *> local_target;" << std::endl;
      cpp << "std::list<" << source << " *>::iterator local_before = local_source.end();" << std::endl;
    }

    // insert items
    for (unsigned int i = 0, size = m_sequence.size(); i < size; ++i)
      m_sequence[i]->createReplace (cpp, false, intermediate);

    // perform intermediate step
    if (intermediate)
      cpp << "PatternMatcher (matcher).process (local_source, local_target);" << std::endl
          << "target.insert (target.end(), local_target.begin(), local_target.end());" << std::endl
          << "oors::purge<" << source << "> (local_purge);" << std::endl;

    // finish recursive stuff
    if (recursive)
      cpp << ifs::pop << "}" << std::endl;

    // done
    return;
  }

  // add else
  bool condition = m_else || m_condition;
  if (m_else)
  {
    cpp << "else";
    if (!m_condition)
      cpp << std::endl;
  }

  // add condition
  if (m_condition)
    m_condition->create (cpp, "if (", ")");

  // open scope
  if (condition)
    cpp << "{" << ifs::push << std::endl;

  // insert items
  for (unsigned int i = 0, size = m_sequence.size(); i < size; ++i)
    m_sequence[i]->createReplace (cpp, recursive, intermediate);

  // close scop
  if (condition)
    cpp << ifs::pop << "}" << std::endl;
}

void SequencePattern::print (const std::string &offset) const
{
  std::cout << offset << m_index << " (" << m_items << "): ";
  if (m_else)
    std::cout << "else ";
  if (m_condition)
  {
    std::cout << "if (";
    m_condition->print ();
    std::cout << ") ";
  }
  std::cout << (m_ordered?'[':'{') << std::endl;
  std::string new_offset = offset+"  ";
  for (unsigned int i = 0; i < m_sequence.size(); ++i)
    m_sequence[i]->print (new_offset);
  std::cout << offset << (m_ordered?']':'}') << std::endl;
}

// WILDCARD PATTERN

WildcardPattern::WildcardPattern (const std::string &filename, unsigned int line, const Rule *rule, SequencePattern *parent)
  : Pattern (filename, line, rule, parent)
{
  if (parent)
    parent->insert (this, true);
}

void WildcardPattern::createInit (IndentFileStream &cpp, const std::string &, const std::string &parent) const
{
  cpp << "new oors::WildcardPattern<" << m_rule->profile()->ruleset()->source() << "> (" << parent << ", " << m_index << ")";
}

void WildcardPattern::createReplace (IndentFileStream &, bool, bool) const
{
  throw Exception ("Generating replacement code for a wildcard pattern.");
}

void WildcardPattern::print (const std::string &offset) const
{
  std::cout << offset << m_index << ": " << "*" << std::endl;
}

// INITIALISABLE PATTERN

InitialisablePattern::InitialisablePattern (const std::string &filename, unsigned int line, const Rule *rule, SequencePattern *parent)
  : Pattern (filename, line, rule, parent)
{
}

InitialisablePattern::~InitialisablePattern (void)
{
  // clean initialisers
  for (unsigned int i = 0, size = m_initialisers.size(); i < size; ++i)
    delete m_initialisers[i].second;
}

// ITEM PATTERN

ItemPattern::ItemPattern (const std::string &filename, unsigned int line, const Rule *rule, SequencePattern *parent, const std::string &base, const std::string &type, Code *code, bool remove)
  : InitialisablePattern (filename, line, rule, parent)
  , m_base (base)
  , m_type (type)
  , m_code (code)
  , m_remove (remove)
{
  if (parent)
    parent->insert (this, true);
}

ItemPattern::~ItemPattern (void)
{
  // clean code
  if (m_code)
    delete m_code;
}

const ItemPattern &ItemPattern::operator= (const ItemPattern &)
{
  throw Exception ("ItemPatterns cannot be assigned.");
}

void ItemPattern::createInit (IndentFileStream &cpp, const std::string &, const std::string &parent) const
{
  cpp << "new oors::ItemPattern<" << m_rule->profile()->ruleset()->source() << "> (" << parent << ", " << m_index << ", " << (m_remove?"true":"false") << ")";
}

void ItemPattern::createReplace (IndentFileStream &cpp, bool recursive, bool intermediate) const
{
  // initialise differently if there are initialisers
  unsigned int size = m_initialisers.size ();
  if (size)
    cpp << "{" << ifs::push << std::endl;

  // add compiler directive
  if (!m_code)
    cpp << ifs::disable << "#line " << m_line << " \"" << m_filename << '"' << ifs::enable << std::endl;

  // create object allocation code
  if (size)
  {
    std::stringstream prefix ("");
    prefix << m_type << " *pattern = new " << m_type << " (";
    if (m_code)
      m_code->create (cpp, prefix.str(), ");");
    else
      cpp << prefix << ");" << std::endl;

    // call initialisers
    for (unsigned int i = 0; i < size; ++i)
    {
      const std::pair<std::string, Code *> &initialiser = m_initialisers[i];
      std::stringstream prefix ("");
      prefix << "pattern->" << initialiser.first << " (";
      initialiser.second->create (cpp, prefix.str(), ");");
    }
  }

  // insert pattern
  std::stringstream prefix ("");
  std::string suffix ((!recursive && intermediate)?")));":"));");
  if (recursive)
  {
    if (intermediate)
      prefix << "local_source.insert (local_before, ";
    else
      prefix << "target.insert (before, ";
  }
  else
  {
    if (intermediate)
      prefix << "local_purge.push_back (*local_source.insert (local_before, ";
    else
      prefix << "target.push_back (";
  }

  // insert the previously allocated object
  prefix << "patterns[" << m_index << "] = ";
  if (size)
    cpp << prefix.str() << "pattern);" << std::endl;

  // err, allocate the object now then
  else
  {
    prefix << "new " << m_type << " (";
    if (m_code)
      m_code->create (cpp, prefix.str(), suffix);
    else
      cpp << prefix.str() << suffix << std::endl;
  }

  // add compiler directive
  if (!m_code)
    cpp << ifs::disable << "#line " << cpp.line()+1 << " \"" << cpp.filename() << '"' << ifs::enable << std::endl;

  // close scope
  if (size)
    cpp << ifs::pop << "}" << std::endl;
}

void ItemPattern::print (const std::string &offset) const
{
  std::cout << offset << m_index << ": " << m_type << std::endl;
}

// CONSERVATION PATTERN

ConservationPattern::ConservationPattern (const std::string &filename, unsigned int line, const Rule *rule, SequencePattern *parent, const ItemPattern *item)
  : InitialisablePattern (filename, line, rule, parent)
  , m_item (item)
{
  if (parent)
    parent->insert (this, true);
}

void ConservationPattern::createInit (IndentFileStream &, const std::string &, const std::string &) const
{
  throw Exception ("Generating initialiser code for conservation patterns.");
}

void ConservationPattern::createReplace (IndentFileStream &cpp, bool recursive, bool intermediate) const
{
  // initialise differently if there are initialisers
  const std::string &source = m_rule->profile()->ruleset()->source ();
  unsigned int size = m_initialisers.size ();
  if (size)
  {
    // get item type and determine whether to cast or not
    const std::string &type = m_item->type ();
    bool cast = type != source;

    // do the thing
    cpp << "{" << ifs::push << std::endl
        << type << " *pattern = ";
    if (cast)
      cpp << "static_cast<" << type << " *>(";
    cpp << "static_cast<oors::ItemPattern<" << source << "> *>(m_alternative->pattern(" << m_item->index() << "))->item(false)";
    if (cast)
      cpp << ")";
    cpp << ";" << std::endl;

    // call initialisers
    for (unsigned int i = 0; i < size; ++i)
    {
      const std::pair<std::string, Code *> &initialiser = m_initialisers[i];
      std::stringstream prefix ("");
      prefix << "pattern->" << initialiser.first << " (";
      initialiser.second->create (cpp, prefix.str(), ");");
    }
  }

  // how to insert
  if (intermediate)
    cpp << "local_source.push_back (";
  else if (recursive)
    cpp << "target.insert (before, ";
  else
    cpp << "target.push_back (";

  // what to insert
  cpp << "patterns[" << m_index << "] = ";
  if (size)
    cpp << "pattern);" << std::endl
        << ifs::pop << "}" << std::endl;
  else
    cpp << "static_cast<oors::ItemPattern<" << source << "> *>(m_alternative->pattern(" << m_item->index() << "))->item(false));" << std::endl;
}

void ConservationPattern::print (const std::string &offset) const
{
  std::cout << offset << m_index << ": Conservating " << m_item->index() << "." << std::endl;
}
