/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * pattern.h
 * 2005-09-05
 */

#ifndef __PATTERN_H__
#define __PATTERN_H__

#include <string>
#include <vector>

// forward declarations
class Code;
class Rule;
class SequencePattern;
class IndentFileStream;

/**
 * Abstract pattern.
 */
class Pattern
{
  public:
    /**
     * Initialize pattern.
     * @param filename The file in which this pattern has been specified.
     * @param line The original line number.
     * @param rule Associated rule.
     * @param parent Parent pattern.
     */
    Pattern (const std::string &filename, unsigned int line, const Rule *rule, SequencePattern *parent);

    /**
     * Virtual destructor.
     */
    virtual ~Pattern (void);

  public:
    /**
     * Initialize given variable with current pattern.
     * @param cpp Source code stream.
     * @param name Pattern name.
     * @param parent Parent pattern.
     */
    virtual void createInit (IndentFileStream &cpp, const std::string &name, const std::string &parent) const = 0;

    /**
     * Create replacement code that is called whenever a rule is applied.
     * @param cpp Source code stream.
     * @param recursive True if code for recursive matching should be generated.
     * @param intermediate True if code for intermediate replace patterns should be generated.
     */
    virtual void createReplace (IndentFileStream &cpp, bool recursive, bool intermediate) const = 0;

    /**
     * Print the pattern.
     * @param offset Offset string.
     */
    virtual void print (const std::string &offset = "") const = 0;

  public:
    /**
     * Set pattern index.
     * @param index New pattern index.
     */
    inline void setIndex (unsigned int index) { m_index = index; }

    /**
     * Set parent sequence.
     * In this setting no sequence will be moved to a pattern with a different root than before.
     * @param parent Parent sequence.
     */
    inline void setParent (SequencePattern *parent) { m_parent = parent; }

  public:
    /**
     * Access source filename.
     * @return The filename.
     */
    inline const std::string &filename (void) const { return m_filename; }

    /**
     * Access source line number.
     * @return The line number.
     */
    inline unsigned int line (void) const { return m_line; }

    /**
     * Access pattern index.
     * @return Pattern index.
     */
    inline unsigned int index (void) const { return m_index; }

    /**
     * Access root pattern list.
     * @return Root pattern list.
     */
    inline SequencePattern *root (void) const { return m_root; }

    /**
     * Access parent pattern list.
     * @return Parent pattern list.
     */
    inline SequencePattern *parent (void) const { return m_parent; }

  protected:
    /// the file in which the pattern was specified
    std::string m_filename;

    /// the line number in that file
    unsigned int m_line;

    /// associated rule
    const Rule *m_rule;

    /// root pattern list
    SequencePattern *m_root;

    /// parent pattern list
    SequencePattern *m_parent;

    /// pattern index
    unsigned int m_index;
};

/**
 * A sequence of patterns that can either be ordered or not.
 */
class SequencePattern : public Pattern
{
  public:
    /**
     * Create new sequence pattern.
     * This will potentially set the sequence as the rule's search or replace pattern.
     * @param filename The file in which this pattern has been specified.
     * @param line The original line number.
     * @param rule Associated rule.
     * @param parent Parent pattern.
     * @param ordered True if the sequence is ordered.
     * @param search True if this sequence is a search pattern.
     * @param intermediate True if this sequence is an intermediate replace pattern.
     */
    SequencePattern (const std::string &filename, unsigned int line, Rule *rule, SequencePattern *parent, bool ordered, bool search, bool intermediate);

    /**
     * Dispose sequence pattern.
     */
    ~SequencePattern (void);

  public:
    /**
     * Insert pattern at given index.
     * @param pattern The pattern to insert.
     * @param index Desired index.
     * @param accessible If true the pattern will be inserted in the root pattern as accessible pattern.
     * @throws Parser::Exception If a redundant wildcard pattern is going to be added.
     */
    void insert (Pattern *pattern, unsigned int index, bool accessible);

    /**
     * Normalise the list pattern.
     * This will remove useless sub-patterns but will preserve the pattern semantics.
     */
    void normalise (void);

  public:
    /**
     * Insert pattern.
     * @param pattern The pattern to insert.
     * @param accessible Pattern is accessible through pattern access operators ($i).
     */
    inline void insert (Pattern *pattern, bool accessible) { insert(pattern, m_sequence.size(), accessible); }

    /**
     * Set condition.
     * @param condition The associated condition guarding this sequence.
     * @param flag Indicates whether this baby is part of an else branch.
     */
    inline void setCondition (Code *condition, bool flag) { m_condition = condition; m_else = flag; }

  public:
    /**
     * Access the sequence.
     * @return Pattern sequence.
     */
    inline const std::vector<Pattern *> sequence (void) const { return m_sequence; }

    /**
     * Access side condition.
     * @return The sequence's side condition.
     */
    inline Code *condition (void) const { return m_condition; }

    /**
     * Access the ith accessible pattern.
     * @param index Pattern index.
     * @return The pattern with the given index or NULL if not existing.
     */
    inline Pattern *pattern (unsigned int index) const { return (index < m_patterns.size())?m_patterns[index]:NULL; }

    /**
     * Access pattern size.
     * @return Amount of accessible patterns.
     */
    inline unsigned int size (void) const { return m_patterns.size(); }

    /**
     * Access the amount of items that are to be matched.
     * @return Amount of items that are to be matched in this sequence.
     */
    inline unsigned int items (void) const { return m_items; }

    /**
     * Indicates whether repetitive matching will terminate some day.
     * @return True if matching with this pattern will terminate.
     */
    inline bool effect (void) const { return m_items != 0; }

    /**
     * Check whether the sequence is ordered.
     * @return True if this sequence of patterns is ordered.
     */
    inline bool ordered (void) const { return m_ordered; }

    /**
     * Access pattern type.
     * @return True if this pattern is a search pattern.
     */
    inline bool search (void) const { return m_search; }

    /**
     * Check whether the pattern is an intermediate pattern.
     * @return True if this pattern is an intermediate replace pattern.
     */
    inline bool intermediate (void) const { return m_intermediate; }

  public:
    void createInit (IndentFileStream &cpp, const std::string &name, const std::string &parent) const;
    void createReplace (IndentFileStream &cpp, bool recursive, bool intermediate) const;
    void print (const std::string &offset = "") const;

  private:
    /// pattern sequence
    std::vector<Pattern *> m_sequence;

    /// all available patterns (excluding sequence patterns)
    std::vector<Pattern *> m_patterns;

    /// associated condition
    Code *m_condition;

    /// how many items have to be matched in this sequence
    unsigned int m_items;

    /// ordered or not
    bool m_ordered;

    /// indicates whether this is a search or replace pattern
    bool m_search;

    /// indicates whether this is an intermediate replace pattern
    bool m_intermediate;

    /// indicates if an unordered sequence contains a wildcard pattern
    bool m_wildcard;

    /// whether the sequence is part of an else branch
    bool m_else;
};

/**
 * A wildcard pattern.
 */
class WildcardPattern : public Pattern
{
  public:
    /**
     * Initialize wildcard pattern.
     * @param filename The file in which this pattern has been specified.
     * @param line The original line number.
     * @param rule Associated rule.
     * @param parent Parent pattern.
     */
    WildcardPattern (const std::string &filename, unsigned int line, const Rule *rule, SequencePattern *parent);

  public:
    void createInit (IndentFileStream &cpp, const std::string &name, const std::string &parent) const;
    void createReplace (IndentFileStream &cpp, bool recursive, bool intermediate) const;
    void print (const std::string &offset = "") const;
};

/**
 * An initialisable pattern.
 * This cannot be directly instatiated.
 */
class InitialisablePattern : public Pattern
{
  public:
    /**
     * Create new initialisable pattern.
     * @param filename The file in which this pattern has been specified.
     * @param line The original line number.
     * @param rule Associated rule.
     * @param parent Parent pattern.
     */
    InitialisablePattern (const std::string &filename, unsigned int line, const Rule *rule, SequencePattern *parent);

    /**
     * Dispose initialisable pattern.
     */
    ~InitialisablePattern (void);

  public:
    /**
     * Add initialiser.
     * @param name Name of the thing to initialise.
     * @param code The initialiser code.
     */
    inline void addInitialiser (const std::string &name, Code *code) { m_initialisers.push_back (std::pair<std::string, Code *>(name, code)); }

  protected:
    /// associated initialisers
    std::vector< std::pair<std::string, Code *> > m_initialisers;
};

/**
 * A single item pattern.
 */
class ItemPattern : public InitialisablePattern
{
  public:
    /**
     * Create new pattern item.
     * @param filename The file in which this pattern has been specified.
     * @param line The original line number.
     * @param rule Associated rule.
     * @param parent Parent pattern.
     * @param base Object base type.
     * @param type Object type.
     * @param code Associated enriched C/C++ code.
     * @param remove Whether the matched item should be removed from the basic block.
     */
    ItemPattern (const std::string &filename, unsigned int line, const Rule *rule, SequencePattern *parent, const std::string &base, const std::string &type, Code *code = NULL, bool remove = true);

    /**
     * Dispose pattern item.
     */
    ~ItemPattern (void);

  public:
    /**
     * Assigment operator.
     * It is never used, but if so, it raises an exception.
     */
    const ItemPattern &operator= (const ItemPattern &);

  public:
    /**
     * Set code.
     * @param code Associated code.
     */
    inline void setCode (Code *code) { m_code = code; }

    /**
     * Set whether to remove the matched item frm the basic block.
     * @param remove True if to remove the item from the basic block.
     */
    inline void setRemove (bool remove) { m_remove = remove; }

  public:
    /**
     * Access object base type.
     * @return Object base type.
     */
    inline const std::string &base (void) const { return m_base; }

    /**
     * Access object type.
     * @return Object type.
     */
    inline const std::string &type (void) const { return m_type; }

    /**
     * Get code.
     * @return Associated code.
     */
    inline const Code *code (void) const { return m_code; }

    /**
     * Check whether to remove the matched item from the basic block.
     * @return True if so.
     */
    inline bool remove (void) const { return m_remove; }

  public:
    void createInit (IndentFileStream &cpp, const std::string &name, const std::string &parent) const;
    void createReplace (IndentFileStream &cpp, bool recursive, bool intermediate) const;
    void print (const std::string &offset = "") const;

  private:
    /// object base type
    const std::string &m_base;

    /// object type
    std::string m_type;

    /// associated code
    Code *m_code;

    /// whether the matched item should be removed from the basic block
    bool m_remove;
};

/**
 * A pattern that is only used in replace patterns
 * for preventing an item to be deleted within a recursive match.
 */
class ConservationPattern : public InitialisablePattern
{
  public:
    /**
     * Initialize conservation pattern.
     * @param filename The file in which this pattern has been specified.
     * @param line The original line number.
     * @param rule Associated rule.
     * @param parent The parent sequence.
     * @param item The encapsulated item pattern.
     */
    ConservationPattern (const std::string &filename, unsigned int line, const Rule *rule, SequencePattern *parent, const ItemPattern *item);

  public:
    /**
     * Access the encapsulated item pattern.
     * @return The encapsulated item pattern.
     */
    inline const ItemPattern *item (void) const { return m_item; }

  public:
    void createInit (IndentFileStream &cpp, const std::string &name, const std::string &parent) const;
    void createReplace (IndentFileStream &cpp, bool recursive, bool intermediate) const;
    void print (const std::string &offset = "") const;

  private:
    /// the item (pattern) to keep from deleting
    const ItemPattern *m_item;
};

#endif
