/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

#include "profile.h"
#include "rule.h"

#include <indentfilestream.h>

#include <iostream>

Profile::Profile (const RuleSet *ruleset)
  : m_ruleset (ruleset)
{
}

Profile::Profile (const RuleSet *ruleset, const std::string &name)
  : m_ruleset (ruleset)
  , m_name (name)
{
}

Profile::~Profile (void)
{
  while (!m_rules.empty())
  {
    delete m_rules.back ();
    m_rules.pop_back ();
  }
}

unsigned int Profile::rules (const Profile *profile) const
{
  // set current profile
  if (!profile)
    profile = this;

  // determine number of inherited rules
  unsigned int number = 0;
  for (unsigned int i = 0, size = m_extends.size(); i < size; ++i)
    number += m_extends[i]->rules (profile);

  // determine number of profile-local rules
  for (unsigned int i = 0, size = m_rules.size(); i < size; ++i)
  {
    Rule *rule = m_rules[i];
    number += !profile->omits(rule) && !rule->abstract();
  }

  // done
  return number;
}

bool Profile::extend (const Profile *profile)
{
  // do not extend twice
  for (unsigned int i = 0, size = m_extends.size(); i < size; ++i)
    if (m_extends[i] == profile)
      return false;

  // do the thing
  m_extends.push_back (profile);
  return true;
}

bool Profile::omit (const Rule *rule)
{
  // check if the rule is contained in this profile
  bool contained = false;
  for (unsigned int i = 0, size = m_rules.size(); i < size; ++i)
    if (m_rules[i] == rule)
    {
      contained = true;
      break;
    }

  // do the thing
  m_omits.insert (rule);
  return !contained;
}

bool Profile::omits (const Rule *rule) const
{
  // locally omitted
  if (m_omits.find(rule) != m_omits.end())
    return true;

  // previously omitted
  for (unsigned int i = 0, size = m_extends.size(); i < size; ++i)
    if (m_extends[i]->omits(rule))
      return true;

  // not omitted at all
  return false;
}

Rule *Profile::rule (const std::string &name) const
{
  // search
  for (unsigned int i = 0; i < m_rules.size(); ++i)
  {
    Rule *rule = m_rules[i];
    if (rule->name() == name)
      return rule;
  }

  // not found
  return NULL;
}

void Profile::create (IndentFileStream &cpp) const
{
  for (unsigned int i = 0; i < m_rules.size(); ++i)
    m_rules[i]->create (cpp);
}

void Profile::createInit (IndentFileStream &cpp, const Profile *profile) const
{
  // set current profile
  if (!profile)
    profile = this;

  // insert rules of extended profiles
  for (unsigned int i = 0, size = m_extends.size(); i < size; ++i)
  {
    const Profile *extends = m_extends[i];
    if (!extends->global() || (profile == this))
      extends->createInit (cpp, profile);
  }

  // insert own rules
  for (unsigned int i = 0, size = m_rules.size(); i < size; ++i)
  {
    Rule *rule = m_rules[i];
    if (!profile->omits(rule) && !rule->abstract())
      cpp << "m_rules.push_back (new "+rule->classname()+" ());" << std::endl;
  }
}

void Profile::print (void) const
{
  // print header
  if (global())
    std::cout << "--- GLOBAL RULES (" << rules() << "): ---" << std::endl;
  else
    std::cout << "--- PROFILE '" << m_name << "' (" << rules() << "): ---" << std::endl;

  // print rules
  for (unsigned int i = 0, size = m_rules.size(); i < size; ++i)
    m_rules[i]->print ();
}
