/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * profile.h
 * 2005-10-16
 */

#ifndef __PROFILE_H__
#define __PROFILE_H__

#include <string>
#include <vector>
#include <set>

// forward declarations
class Rule;
class RuleSet;
class IndentFileStream;

/**
 * Profile stores profile specific rules.
 */
class Profile
{
  public:
    /**
     * Create global profile.
     * @param ruleset The ruleset it belongs to.
     */
    Profile (const RuleSet *ruleset);

    /**
     * Create profile with given name.
     * @param ruleset The ruleset it belongs to.
     * @param name Profile name.
     */
    Profile (const RuleSet *ruleset, const std::string &name);

    /**
     * Dispose profile.
     */
    ~Profile (void);

  public:
    /**
     * Insert the given rule.
     * @param rule The rule to insert.
     */
    inline void insert (Rule *rule) { m_rules.push_back (rule); }

  public:
    /**
     * Get associated ruleset.
     * @return Associated ruleset.
     */
    inline const RuleSet *ruleset (void) const { return m_ruleset; }

    /**
     * Checks whether this is the global profile.
     * @return True if this is the global profile.
     */
    inline bool global (void) const { return !m_name.length(); }

    /**
     * Access profile name.
     * @return The profile's name.
     */
    inline const std::string &name (void) const { return m_name; }

    /**
     * Access extended profiles.
     * @return All extended profiles.
     */
    inline const std::vector<const Profile *> &extends (void) const { return m_extends; }

  public:
    /**
     * Determine the number of active rules contained in and inherited by this profile.
     * @param profile The profile to ask which rules are omitted.
     * @return The number of active rules.
     */
    unsigned int rules (const Profile *profile = NULL) const;

    /**
     * Extend the given profile.
     * @param profile The profile to extend.
     * @return True if the given profile has not been extended yet.
     */
    bool extend (const Profile *profile);

    /**
     * Omit the given rule.
     * @param rule The rule to omit.
     * @return True if omitting the rule does not make a previously defined rule unavailable.
     */
    bool omit (const Rule *rule);

    /**
     * Check if the given rule is omitted by this profile.
     * @param rule The rule to check.
     * @return True if it is omitted.
     */
    bool omits (const Rule *rule) const;

    /**
     * Get rule with given name.
     * @param name Name of the required rule.
     * @return The rule or NULL if not existing.
     */
    Rule *rule (const std::string &name) const;

    /**
     * Create rule classes.
     * @param cpp Code file stream.
     */
    void create (IndentFileStream &cpp) const;

    /**
     * Create tree pattern matcher initialization code.
     * @param cpp Code file stream.
     * @param profile The profile to ask which rules are to be omitted.
     */
    void createInit (IndentFileStream &cpp, const Profile *profile = NULL) const;

    /**
     * Print profile.
     */
    void print (void) const;

  private:
    /// associated ruleset
    const RuleSet *m_ruleset;

    /// profile name
    std::string m_name;

    /// the profiles that are extended by this one
    std::vector<const Profile *> m_extends;

    /// list of omitted rules
    std::set<const Rule *> m_omits;

    /// stored rules
    std::vector<Rule *> m_rules;
};

#endif
