/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * rule.cpp
 * 2005-09-05
 */

#include "rule.h"

#include "code.h"
#include "parser.h"
#include "pattern.h"
#include "ruleset.h"

#include <exception.h>
#include <indentfilestream.h>

#include <sstream>
#include <iostream>

Rule::Rule (const Profile *profile, const std::string &name, const Rule *parent, Code *mask)
  : m_profile (profile)
  , m_parent (parent)
  , m_name (name)
  , m_mask (mask)
  , m_search (NULL)
  , m_condition (NULL)
  , m_cost (NULL)
  , m_replace (NULL)
  , m_used_search (parent?parent->m_used_search:NULL)
  , m_used_condition (parent?parent->m_used_condition:NULL)
  , m_used_cost (parent?parent->m_used_cost:NULL)
  , m_used_replace (parent?parent->m_used_replace:NULL)
  , m_implicit_condition (NULL)
  , m_implicit_postpass (NULL)
{
}

Rule::~Rule (void)
{
  if (m_mask) delete m_mask;
  if (m_search) delete m_search;
  if (m_condition) delete m_condition;
  if (m_cost) delete m_cost;
  if (m_replace) delete m_replace;
  if (m_implicit_condition) delete m_implicit_condition;
  if (m_implicit_postpass) delete m_implicit_postpass;
}

const Pattern *Rule::pattern (unsigned int index, bool search) const
{
  // look in search pattern first
  const Pattern *pattern = NULL;
  if (m_used_search)
  {
    pattern = m_used_search->pattern (index);
    index -= m_used_search->size ();
  }

  // look in replace pattern if allowed to
  if (m_used_replace && !search && !pattern)
    pattern = m_used_replace->pattern (index);

  // done
  return pattern;
}

void Rule::verify (void) const
{
  // nothing to do
  if (!m_search)
    return;

  // verify condition function
  if (m_used_condition && !m_condition)
    verify ("In condition function of rule '"+m_name+"':", m_used_condition);

  // verify cost function
  if (m_used_cost && !m_cost)
    verify ("In cost function of rule '"+m_name+"':", m_used_cost);

  // verify replace function
  if (m_used_replace && !m_replace)
  {
    // check sequence side conditions
    verify (m_used_replace);

    // check item patterns
    const std::string &msg = "In replace pattern of rule '"+m_name+"':";
    for (unsigned int i = 0, size = m_used_replace->size(); i < size; ++i)
    {
      // check item pattern
      Pattern *current = m_used_replace->pattern (i);
      if (dynamic_cast<ItemPattern *>(current))
      {
        const Code *code = static_cast<ItemPattern *>(current)->code ();
        if (code) verify (msg, code);
      }

      // check conservation pattern
      else/* if (dynamic_cast<ConservationPattern *>(current))*/
      {
        const ItemPattern *item = static_cast<ConservationPattern *>(current)->item ();
        unsigned int index = item->index ();
        verify (msg, false, index, item, pattern(index, true), NULL);
      }
    }
  }
}

void Rule::verify (const SequencePattern *pattern) const
{
  const std::string &msg = "In condition within replace pattern of rule '"+m_name+"':";
  const std::vector<Pattern *> &sequence = pattern->sequence ();
  for (unsigned int i = 0, size = sequence.size(); i < size; ++i)
  {
    SequencePattern *current = dynamic_cast<SequencePattern *>(sequence[i]);
    if (current)
    {
      // first verify the side condition
      Code *condition = current->condition ();
      if (condition)
        verify (msg, condition);

      // then decent into sequence pattern
      verify (current);
    }
  }
}

void Rule::verify (const std::string &prefix, const Code *code) const
{
  // check each single segment
  const std::vector<Code::Segment> &segments = code->code ();
  for (unsigned int i = 0; i < segments.size(); ++i)
  {
    const Pattern *accessed = segments[i].pattern ();
    if (accessed)
    {
      bool search = accessed->root()->search();
      unsigned int index = (search?0:m_search->size())+accessed->index();
      const Pattern *found = pattern (index, search);
      verify (prefix, search, index, accessed, found, code);
    }
  }
}

void Rule::verify (const std::string &prefix, bool search, unsigned int index, const Pattern *accessed, const Pattern *found, const Code *code) const
{
  // the pattern cannot be accessed
  if (!found)
  {
    std::stringstream msg ("");
    msg << prefix << " Accessing unavailable pattern $" << (index+1) << ".";
    if (code) throw Parser::Exception (msg.str(), true, code->filename(), code->line());
    else throw Parser::Exception (msg.str());
  }

  // no need to go on
  if (!search)
    return;

  // mismatching pattern types
  const ItemPattern *a = dynamic_cast<const ItemPattern *>(accessed), *b = dynamic_cast<const ItemPattern *>(found);
  if (static_cast<bool>(a) ^ static_cast<bool>(b))
  {
    std::stringstream msg ("");
    msg << prefix << " Mismatching pattern types at $" << (index+1) << ": Expected " << (a?"item":"wildcard") << " pattern.";
    if (code) throw Parser::Exception (msg.str(), true, code->filename(), code->line());
    else throw Parser::Exception (msg.str());
   }

  // mismatching item pattern types
  if (a && (a->type() != b->type()))
  {
    std::stringstream msg ("");
    msg << prefix << " Mismatching item pattern types at $" << (index+1) << ": Expected '" << a->type() << "', got '" << b->type() << "'.";
    Parser *parser = Parser::instance ();
    if (parser->flag(Parser::pedantic))
    {
      if (code) throw Parser::Exception (msg.str(), true, code->filename(), code->line());
      else throw Parser::Exception (msg.str());
    }
    else
      if (code) parser->warning (msg.str(), true, code->filename(), code->line());
      else parser->warning (msg.str());
  }
}

void Rule::create (IndentFileStream &cpp) const
{
  // preparations
  const RuleSet *ruleset = m_profile->ruleset();
  bool recursive = ruleset->recursive();
  const std::string &source = ruleset->source(), &target = ruleset->target();
  std::string name = classname(), parent = (m_parent?m_parent->classname():"Rule");

  // begin class
  cpp << std::endl
      << "class " << name << " : public " << parent << std::endl
      << "{" << std::endl << ifs::push;

  // modify constructor
  cpp << "public:" << std::endl << ifs::push;

  // create prefix
  std::stringstream prefix (""), suffix ("");
  prefix << name << " (const std::string &name = \"" << fullname() << "\"";

  // create suffix
  if (m_mask)
    suffix << ", bool has_mask = true";
  else
    suffix << ", int mask = 0, bool has_mask = false";
  suffix << ", bool" << ((m_parent || m_search)?" init = true":" = false") << ")" << std::endl;

  // insert mask
  if (m_mask)
  {
    prefix << ", int mask = ";
    m_mask->create (cpp, prefix.str(), suffix.str());
  }

  // forget about the mask
  else
    cpp << prefix << suffix;

  // insert remainder of constructor
  cpp << ifs::push << ": " << parent << " (name, mask, has_mask";
  if (m_parent)
    cpp << ", " << (m_search?"false":"init");
  cpp << ")" << ifs::pop << std::endl
      << "{" << std::endl << ifs::push;

  // init pattern
  if (m_search)
  {
    cpp << "if (init)" << std::endl
        << "{" << std::endl << ifs::push;
    m_search->createInit (cpp, "pattern", "NULL");
    cpp << "m_pattern = pattern;" << std::endl
        << "m_size = " << m_search->size() << ";" << std::endl
        << "m_items = " << m_search->items() << ";" << std::endl
        << ifs::pop << "}" << std::endl;
  }

  // done with constructor
  cpp << ifs::pop << "}" << std::endl
      << ifs::pop;

  // check whether search pattern contains a wildcard pattern
  bool wildcard = false;
  if (m_search)
    for (unsigned int i = 0, size = m_search->size(); i < size; ++i)
      if (wildcard = dynamic_cast<WildcardPattern *>(m_search->pattern(i)))
        break;

  // insert side condition, cost function and replace function
  if (m_condition || wildcard || m_cost || m_replace)
  {
    // make that stuff public
    cpp << std::endl
        << "public:" << ifs::push;

    // condition function
    if (m_condition || wildcard)
    {
      // function header
      cpp << std::endl
          << "virtual bool condition (void) const" << std::endl
          << "{" << ifs::push << std::endl;

      // check implicit condition first
      if (wildcard)
        cpp << "if (!implicitCondition (m_alternative->trace()))" << std::endl
            << ifs::push << "return false;" << ifs::pop << std::endl << std::endl;

      // insert the real condition
      if (m_condition)
        m_condition->create (cpp);
      else
        cpp << "return " << parent << "::condition ();" << std::endl;

      // done with function
      cpp << ifs::pop << "}" << std::endl;
    }

    // insert cost function
    if (m_cost)
    {
      cpp << std::endl
          << "virtual int cost (void) const" << std::endl
          << "{" << ifs::push << std::endl;
      m_cost->create (cpp);
      cpp << ifs::pop << "}" << std::endl;
    }

    // replace pattern
    if (m_replace)
    {
      // function header
      bool effect = m_replace->size (), intermediate = m_replace->intermediate ();
      cpp << std::endl
          << "virtual void apply (const oors::BasePatternMatcher &" << ((effect && intermediate)?"matcher":"")
          << ", std::list<" << target << " *> &" << ((effect || recursive)?"target":"")
          << ", bool" << (recursive?" insert":"") << ") const" << std::endl
          << "{" << std::endl << ifs::push;

      // create replace pattern code
      if (effect)
        m_replace->createReplace (cpp, recursive, intermediate);

      // remove matched items if in insertion mode
      else if (recursive)
        cpp << "if (insert)" << std::endl
            << "{" << ifs::push << std::endl
            << "remove (target);" << std::endl
            << "implicitPostpass (target);" << std::endl
            << ifs::pop << "}" << std::endl;

      // done with function
      cpp << ifs::pop << "}" << std::endl;
    }

    // done with public stuff
    cpp << ifs::pop;
  }

  // check if any subpattern of the search pattern has a side condition
  bool condition = false;
  if (m_search)
    for (unsigned int i = 0; i < m_search->size(); ++i)
    {
      ItemPattern *pattern = dynamic_cast<ItemPattern *>(m_search->pattern(i));
      if (condition = (pattern && (pattern->code() || (pattern->type() != source))))
        break;
    }

  // insert item pattern side conditions and implicit functions
  if (condition || m_implicit_condition || m_implicit_postpass)
  {
    // make that stuff protected
    cpp << std::endl
        << "protected:" << ifs::push;

    // add pattern side conditions
    if (condition)
    {
      // function header
      cpp << std::endl
        << "virtual bool condition (unsigned int" << (condition?" index":"") << ") const" << std::endl
        << "{" << ifs::push << std::endl
        << "switch (index)" << std::endl
        << "{" << ifs::push << std::endl;

      // insert side conditions
      for (unsigned int i = 0, size = m_search->size(); i < size; ++i)
      {
        ItemPattern *pattern = dynamic_cast<ItemPattern *>(m_search->pattern(i));
        if (pattern)
        {
          // preparations
          const Code *code = pattern->code ();
          const std::string &type = pattern->type ();
          bool cast = type != pattern->base ();

          // create condition
          if (cast || code)
          {
            // construct prefix
            std::stringstream prefix ("");
            prefix << "case " << i << ": return ";
            if (cast)
              prefix << "dynamic_cast<" << type << " *>(m_alternative->current())";

            // insert condition
            if (code)
            {
              // construct suffix
              std::string suffix = ";";
              if (cast)
              {
                prefix << " && (";
                suffix = ");";
              }

              // now really insert the code
              code->create (cpp, prefix.str(), suffix);
            }

            // there is not condition
           else
             cpp << prefix << ";" << std::endl;
          }
        }
      }

      // done with funtion
      cpp << "default: return true;" << std::endl
          << ifs::pop << "}" << std::endl
          << ifs::pop << "}" << std::endl;
    }

    // add implicit condition function
    if (m_implicit_condition)
    {
      cpp << std::endl
          << "virtual bool implicitCondition (const std::list<const oors::Pattern<" << source << "> *> &trace) const" << std::endl
          << "{" << ifs::push << std::endl;
      m_implicit_condition->create (cpp);
      cpp << ifs::pop << "}" << std::endl;
    }

    // add implicit postpass function
    if (m_implicit_postpass)
    {
      cpp << std::endl
          << "virtual bool implicitPostpass (const std::list<" << source << " *> &block) const" << std::endl
          << "{" << ifs::push << std::endl;
      m_implicit_postpass->create (cpp);
      cpp << ifs::pop << "}" << std::endl;
    }

    // done with protected stuff
    cpp << ifs::pop;
  }

  // finish class
  cpp << ifs::pop << "};" << std::endl;
}

void Rule::print (void) const
{
  std::cout << "Rule '" << m_name << "'";
  if (m_parent) std::cout << " extends '" << m_parent->m_name << "'";
  if (m_mask)
  {
    std::cout << " (mask: ";
    m_mask->print ();
    std::cout << ")";
  }
  std::cout << ":" << std::endl;

  if (m_search)
  {
    std::cout << "Search pattern (" << m_search->size() << "):" << std::endl;
    m_search->print ();
    std::cout << std::endl;
  }
  if (m_condition)
  {
    std::cout << "Condition function:" << std::endl;
    m_condition->print ();
    std::cout << std::endl;
  }
  if (m_cost)
  {
    std::cout << "Cost function:" << std::endl;
    m_cost->print ();
    std::cout << std::endl;
  }
  if (m_replace)
  {
    std::cout << "Replace pattern (" << m_replace->size() << "):" << std::endl;
    m_replace->print ();
    std::cout << std::endl;
  }
  std::cout << std::endl;
}
