/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * rule.h
 * 2005-09-05
 */

#ifndef __RULE_H__
#define __RULE_H__

#include "profile.h"

#include <vector>
#include <string>

// forward declarations
class Code;
class Profile;
class Pattern;
class SequencePattern;
class IndentFileStream;

/**
 * Representation of a rule.
 */
class Rule
{
  public:
    /**
     * Create rule.
     * @param profile The profile the rule belongs to.
     * @param name Rule name.
     * @param parent The rule to derive from.
     * @param mask The mask to use (if any).
     */
    Rule (const Profile *profile, const std::string &name, const Rule *parent, Code *mask);

    /**
     * Dispose rule.
     */
    ~Rule (void);

  public:
    /**
     * Set the search pattern.
     * @param search Rule search pattern.
     */
    inline void setSearch (SequencePattern *search) { m_used_search = m_search = search; }

    /**
     * Set the condition function.
     * @param condition Rule condition function.
     */
    inline void setCondition (Code *condition) { m_used_condition = m_condition = condition; }

    /**
     * Set the cost function.
     * @param cost Rule cost function.
     */
    inline void setCost (Code *cost) { m_used_cost = m_cost = cost; }

    /**
     * Set the replace pattern.
     * @param replace Rule replace pattern.
     */
    inline void setReplace (SequencePattern *replace) { m_used_replace = m_replace = replace; }

    /**
     * Set the implicit condition function.
     * @param condition Implicit condition function.
     */
    inline void setImplicitCondition (Code *condition) { m_implicit_condition = condition; }

    /**
     * Set the implicit postpass function.
     * @param postpass Implicit postass function.
     */
    inline void setImplicitPostPass (Code *postpass) { m_implicit_postpass = postpass; }

  public:
    /**
     * Get profile name.
     * @return profile name.
     */
    inline const Profile *profile (void) const { return m_profile; }

    /**
     * Get the parent rule.
     * @return Parent rule.
     */
    inline const Rule *parent (void) const { return m_parent; }

    /**
     * Get the rule name.
     * @return Rule name.
     */
    inline const std::string &name (void) const { return m_name; }

    /**
     * Get the rule's full name.
     * @return Full name.
     */
    inline std::string fullname (void) const { return m_profile->name()+"::"+m_name; }

    /**
     * Build classname.
     * @return Name of the created class.
     */
    inline std::string classname (void) const { return ((m_profile->name().length())?"_"+m_profile->name():"")+"_"+m_name; }

    /**
     * Access the search pattern.
     * @return Search pattern.
     */
    inline const SequencePattern *search (void) const { return m_search; }

    /**
     * Access the condition function.
     * @return Condition function.
     */
    inline const Code *condition (void) const { return m_condition; }

    /**
     * Access the cost function.
     * @return Cost function.
     */
    inline const Code *cost (void) const { return m_cost; }

    /**
     * Access the most recent replace pattern.
     * @return Replace pattern.
     */
    inline const SequencePattern *replace (void) const { return m_replace; }

    /**
     * Access the used search pattern.
     * @return Search pattern.
     */
    inline const SequencePattern *usedSearch (void) const { return m_used_search; }

    /**
     * Access the used condition function.
     * @return Condition function.
     */
    inline const Code *usedCondition (void) const { return m_used_condition; }

    /**
     * Access the used cost function.
     * @return Cost function.
     */
    inline const Code *usedCost (void) const { return m_used_cost; }

    /**
     * Access the used replace pattern.
     * @return Replace pattern.
     */
    inline const SequencePattern *usedReplace (void) const { return m_used_replace; }

    /**
     * Indicates wheter the rule is derived or not.
     * @return True if rule is derived.
     */
    inline bool derived (void) const { return m_parent; }

    /**
     * Indicates whether the rule is abstract or not.
     * @return True if eithe search pattern, replace pattern or cost function is unspecified.
     */
    inline bool abstract (void) const { return !m_used_search || !m_used_cost || !m_used_replace; }

  public:
    /**
     * Access the ith pattern with a defined type using both patterns.
     * @param index Pattern index.
     * @param search If true only search patterns may be accessed.
     * @return The requested pattern or NULL if not existing.
     */
    const Pattern *pattern (unsigned int index, bool search) const;

    /**
     * Verify the rule.
     * Checks if condition- and cost function and replace pattern only access available patterns.
     * @throws Parser::Exception If unavailable patterns are accessed.
     */
    void verify (void) const;

    /**
     * Create rule code.
     * @param cpp Code file stream.
     */
    void create (IndentFileStream &cpp) const;

    /**
     * Print rule.
     */
    void print (void) const;

  private:
    /**
     * Verify side conditions of given sequence pattern.
     * @param pattern The sequence patten to analyse.
     * @throws Parser::Exception If something is not quite right.
     */
    void verify (const SequencePattern *pattern) const;

    /**
     * Verify the given code.
     * @param prefix Prefix for any exception message.
     * @param code The code to verify.
     * @throws Parser::Exception If something is not quite right.
     */
    void verify (const std::string &prefix, const Code *code) const;

    /**
     * Verify if accessed and found are type equivalent.
     * @param prefix Prefix for any exception message.
     * @param search True if within search pattern.
     * @param index The index of the accessed pattern.
     * @param accessed The accessed pattern.
     * @param found The found pattern.
     * @param code The associated code block (if any).
     */
    void verify (const std::string &prefix, bool search, unsigned int index, const Pattern *accessed, const Pattern *found, const Code *code) const;

  private:
    /// associated profile
    const Profile *m_profile;

    /// parent rule
    const Rule *m_parent;

    /// rule name
    std::string m_name;

    /// rule mask
    Code *m_mask;

    /// own search pattern
    SequencePattern *m_search;

    /// own condition function
    Code *m_condition;

    /// own cost function
    Code *m_cost;

    /// replace pattern sequence
    SequencePattern *m_replace;

    /// used search pattern
    const SequencePattern *m_used_search;

    /// used condition function
    const Code *m_used_condition;

    /// used cost function
    const Code *m_used_cost;

    /// used replace function
    const SequencePattern *m_used_replace;

    /// custom implicit condition function
    Code *m_implicit_condition;

    /// custom implicit postpass function
    Code *m_implicit_postpass;
};

#endif
