/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * ruleset.cpp
 * 2005-09-07
 */

#include "ruleset.h"
#include "profile.h"
#include "code.h"

#include <util.h>
#include <indentfilestream.h>

#include <iostream>

RuleSet::RuleSet (const std::string &source, const std::string &target, bool source_type, bool target_type)
  : m_source (source)
  , m_target (target)
  , m_source_type (source_type)
  , m_target_type (target_type)
  , m_recursive (source == target)
  , m_copy (NULL)
  , m_implicit_condition (NULL)
  , m_implicit_postpass (NULL)
  , m_global (new Profile (this))
{
}

RuleSet::~RuleSet (void)
{
  // delete header(s)
  for (unsigned int i = 0, size = m_header.size(); i < size; ++i)
    delete m_header[i];

  // delete footer(s)
  for (unsigned int i = 0, size = m_footer.size(); i < size; ++i)
    delete m_footer[i];

  // dispose functions
  if (m_copy) delete m_copy;
  if (m_implicit_condition) delete m_implicit_condition;
  if (m_implicit_postpass) delete m_implicit_postpass;

  // dispose profiles
  delete m_global;
  for (unsigned int i  = 0, size = m_profiles.size(); i < size; ++i)
    delete m_profiles[i];
}

Profile *RuleSet::profile (const std::string &name) const
{
  // access global profile
  if (!name.length())
    return m_global;

  // access other profile
  for (unsigned int i = 0, size = m_profiles.size(); i < size; ++i)
  {
    Profile *profile = m_profiles[i];
    if (profile->name() == name)
      return profile;
  }

  // not found
  return NULL;
}

void RuleSet::create (const std::string &prefix) const
{
  // get namespace
  std::string name = extractName (prefix);

  // HEADER
  {
    // create header file
    IndentFileStream h (prefix+".h");

    // create include guards
    h << "#ifndef __" << name << "_H__" << std::endl
      << "#define __" << name << "_H__" << std::endl << std::endl;

    // insert basic includes
    h << "#include <oors.h>" << std::endl << std::endl;

    // insert forward declarations
    h << "// forward declarations" << std::endl
      << (m_source_type?"class ":"struct ") << m_source << ";" << std::endl;
    if (m_source != m_target)
      h << (m_target_type?"class ":"struct ") << m_target << ";" << std::endl;
    h << std::endl;

    // begin namespace
    h << "// custom namespace" << std::endl
      << "namespace " << name << std::endl
      << "{" << ifs::push;

    // begin class declaration
    h << std::endl
      << "/**" << std::endl
      << " * Pattern matcher." << std::endl
      << " */" << std::endl
      << "class PatternMatcher : public oors::PatternMatcher<" << m_source << ", " << m_target << ">" << std::endl
      << "{" << ifs::push;

    // constructor/destructor declaration
    h << std::endl
      << "public:" << std::endl << ifs::push
      << "/**" << std::endl
      << " * Initialize pattern matcher." << std::endl  
      << " * @param profile The profile to use. Leave empty to use global profile." << std::endl
      << " * @throws oors::Exception If the requested profile is not supported." << std::endl
      << " */" << std::endl
      << "PatternMatcher (const std::string &profile = \"\");" << std::endl << std::endl
      << "/**" << std::endl
      << " * Initialize pattern matcher." << std::endl  
      << " * @param profile The profile to use. Leave empty to use global profile." << std::endl
      << " * @param first True to enable first-match mode." << std::endl
      << " * @param immediate True to enable immediate-match mode." << std::endl
      << " * @param local True to enable local-cost-minimum mode." << std::endl
      << " * @param level Maximum level of recursion." << std::endl
      << " * @param mask Rule selection mask." << std::endl
      << " * @throws oors::Exception If the requested profile is not supported." << std::endl
      << " */" << std::endl
      << "PatternMatcher (const std::string &profile, bool first, bool immediate, bool local, unsigned int level = 0, int mask = 0);" << std::endl << std::endl
      << "/**" << std::endl
      << " * Copy pattern matcher properties." << std::endl  
      << " * @param matcher Pattern matcher whose properties to copy." << std::endl
      << " */" << std::endl
      << "PatternMatcher (const BasePatternMatcher &matcher);" << std::endl << std::endl
      << "/**" << std::endl
      << " * Dispose tree-pattern matcher." << std::endl
      << " */" << std::endl
      << "~PatternMatcher (void);" << std::endl
      << ifs::pop;

    // setup stuff
    h << std::endl
      << "public:" << std::endl << ifs::push
      << "/**" << std::endl
      << " * Set the current active profile." << std::endl
      << " * @param profile The name of the profile. Leave empty to use global profile." << std::endl
      << " * @throws oors::Exception If the profile is not supported." << std::endl
      << " */" << std::endl
      << "void setProfile (const std::string &profile = \"\");" << std::endl
      << std::endl
      << "/**" << std::endl
      << " * Clear the used profile." << std::endl
      << " */" << std::endl
      << "void clearProfile (void);" << std::endl
      << ifs::pop;

    // processing stuff
    h << std::endl
      << "public:" << ifs::push;

    // recursive processing mode
    if (m_recursive)
      h << std::endl
        << "/**" << std::endl
        << " * Process given block until no rule can be applied." << std::endl
        << " * @param source The block to process." << std::endl
        << " */" << std::endl
        << "void process (std::list<" << m_source << " *> &source);" << std::endl;

    // sequential processing mode
    h << std::endl
      << "/**" << std::endl
      << " * Process given block." << std::endl
      << " * @param source The block to process." << std::endl
      << " * @param target The resulting block." << std::endl
      << " * @throws oors::Exception If input block cannot be matched." << std::endl
      << " */" << std::endl
      << "void process (const std::list<" << m_source << " *> &source, std::list<" << m_target << " *> &target);" << std::endl
      << ifs::pop;

    // private stuff
    if (m_copy)
      h << std::endl
        << "private:" << std::endl << ifs::push
        << "/**" << std::endl
        << " * Copy given item." << std::endl
        << " * @param item The item to copy." << std::endl
        << " * @return The copied item." << std::endl
        << " */" << std::endl
        << m_source << " *copy (const " << m_source << " *item) const;" << std::endl
        << ifs::pop;

    // end class declaration
    h << ifs::pop << "};" << std::endl;

    // end namespace
    h << ifs::pop << "}" << std::endl << std::endl;

    // close include guards
    h << "#endif" << std::endl;
  }

  // CODE
  {
    // create code file
    IndentFileStream cpp (prefix+".cpp");

    // add our header
    cpp << "#include \"" << name << ".h\"" << std::endl << std::endl
        << "#include <oors_internal.h>" << std::endl
        << "#include <cstring>" << std::endl;

    // add custom header
    for (unsigned int i = 0, size = m_header.size(); i < size; ++i)
    {
      cpp << std::endl;
      m_header[i]->create (cpp);
    }

    // begin namespace
    cpp << std::endl
        << "namespace " << name << std::endl
        << "{" << ifs::push;

    // intermediate rule class
    cpp << std::endl
        << "class Rule : public oors::Rule<"<< m_source << ", " << m_target << ">" << std::endl
        << "{" << std::endl << ifs::push
        << "public:" << std::endl << ifs::push
        << "Rule (const std::string &name, int flag, bool has_flag)" << std::endl << ifs::push
        << ": oors::Rule<" << m_source << ", " << m_target << "> (name, flag, has_flag)" << std::endl << ifs::pop
        << "{" << std::endl
        << "}" << std::endl << std::endl << ifs::pop
        << "protected:" << ifs::push;

    // add implicit side condition
    cpp << std::endl
        << "virtual bool implicitCondition (const std::list<const oors::Pattern<" << m_source << "> *> &" << (m_implicit_condition?"trace":"") << ") const" << std::endl
        << "{" << ifs::push << std::endl;
    if (m_implicit_condition)
      m_implicit_condition->create (cpp);
    else
      cpp << "return true;" << std::endl;
    cpp << ifs::pop << "}" << std::endl;

    // add implicit postpass function
    cpp << std::endl
        << "virtual void implicitPostpass (const std::list<" << m_source << " *> &" << (m_implicit_postpass?"block":"") << ") const" << std::endl
        << "{" << ifs::push << std::endl;
    if (m_implicit_postpass)
      m_implicit_postpass->create (cpp);
    cpp << ifs::pop << "}" << std::endl
        << ifs::pop;

    // finish class
    cpp << ifs::pop << "};" << std::endl;

    // create global profile rules
    m_global->create (cpp);

    // create other profile rules
    for (unsigned int i = 0, size = m_profiles.size(); i < size; ++i)
      m_profiles[i]->create (cpp);

    // constructor
    cpp << std::endl
        << "PatternMatcher::PatternMatcher (const std::string &profile)" << std::endl << ifs::push
        << ": oors::PatternMatcher<" << m_source << ", " << m_target << "> (profile)" << std::endl << ifs::pop
        << "{" << std::endl << ifs::push
        << "setProfile (profile);" << std::endl
        << ifs::pop << "}" << std::endl;

    // alternative constructor
    cpp << std::endl
        << "PatternMatcher::PatternMatcher (const std::string &profile, bool first, bool immediate, bool local, unsigned int level, int mask)" << std::endl << ifs::push
        << ": oors::PatternMatcher<" << m_source << ", " << m_target << "> (profile, first, immediate, local, level, mask)" << std::endl << ifs::pop
        << "{" << std::endl << ifs::push
        << "setProfile (profile);" << std::endl
        << ifs::pop << "}" << std::endl;

    // copy constructor
    cpp << std::endl
        << "PatternMatcher::PatternMatcher (const BasePatternMatcher &matcher)" << std::endl << ifs::push
        << ": oors::PatternMatcher<" << m_source << ", " << m_target << "> (matcher)" << std::endl << ifs::pop
        << "{" << std::endl << ifs::push
        << "setProfile (m_profile);" << std::endl
        << ifs::pop << "}" << std::endl;

    // destructor
    cpp << std::endl
        << "PatternMatcher::~PatternMatcher (void)" << std::endl
        << "{" << std::endl << ifs::push
        << "clearProfile ();" << std::endl
        << ifs::pop << "}" << std::endl;

    // setup profile
    cpp << std::endl
        << "void PatternMatcher::setProfile (const std::string &profile)" << std::endl
        << "{" << std::endl << ifs::push;

    // global profile
    cpp << "if (!profile.length())" << std::endl
        << "{" << std::endl << ifs::push;
    m_global->createInit (cpp);
    cpp << ifs::pop << "}" << std::endl;

    // custom profiles
    for (unsigned int i = 0, size = m_profiles.size(); i < size; ++i)
    {
      Profile *profile = m_profiles[i];
      cpp << "else if (profile == \""+profile->name()+"\")" << std::endl
          << "{" << std::endl << ifs::push;
      profile->createInit (cpp);
      cpp << ifs::pop << "}" << std::endl;
    }

    // unsupported profile
    cpp << "else" << std::endl << ifs::push
         << "throw oors::Exception (\"Unsupported profile: '\"+profile+\"'.\");" << ifs::pop << std::endl
         << ifs::pop << "}" << std::endl;

    // clear profile
    cpp << std::endl
        << "void PatternMatcher::clearProfile (void)" << std::endl
        << "{" << std::endl << ifs::push
        << "for (unsigned int i = 0, size = m_rules.size(); i < size; ++i)" << std::endl << ifs::push
        << "delete m_rules[i];" << ifs::pop << std::endl
        << ifs::pop << "}" << std::endl;

    // process given block recursively
    if (m_recursive)
      cpp << std::endl
          << "void PatternMatcher::process (std::list<" << m_source << " *> &source)" << std::endl
          << "{" << std::endl << ifs::push
          << "if (m_local_minimum)" << std::endl << ifs::push
          << "processLocal (source);" << ifs::pop << std::endl
          << "else" << std::endl << ifs::push
          << "processGlobal (source);" << ifs::pop << std::endl
          << ifs::pop << "}" << std::endl;

    // process given block
    cpp << std::endl
        << "void PatternMatcher::process (const std::list<" << m_source << " *> &source, std::list<" << m_target << " *> &target)" << std::endl
        << "{" << std::endl << ifs::push
        << "if (m_local_minimum)" << std::endl
        << "{" << std::endl << ifs::push
        << "if (!m_immediate_match)" << std::endl << ifs::push
        << "processLocal (source, target);" << ifs::pop << std::endl
        << "else" << std::endl << ifs::push
        << "processLocalImmediate (source, target);" << ifs::pop << std::endl
        << ifs::pop << "}" << std::endl
        << "else" << std::endl
        << "{" << std::endl << ifs::push
        << "if (m_immediate_match)" << std::endl << ifs::push
        << "throw oors::Exception (\"Immediate matching currently not supported in global-cost-minimum mode.\");" << ifs::pop << std::endl
        << "processGlobal (source, target);" << std::endl
        << ifs::pop << "}" << std::endl
        << ifs::pop << "}" << std::endl;

    // copy object function
    if (m_copy)
    {
      cpp << std::endl
          << m_source << " *PatternMatcher::copy (const " << m_source << " *item) const" << std::endl
          << "{" << std::endl << ifs::push;
      m_copy->create(cpp);
      cpp << ifs::pop << "}" << std::endl;
    }

    // end namespace
    cpp << ifs::pop << "}" << std::endl;

    // add custom footer
    for (unsigned int i = 0, size = m_footer.size(); i < size; ++i)
    {
      cpp << std::endl;
      m_footer[i]->create (cpp);
    }
  }
}

void RuleSet::print (void) const
{
  // print header
  std::cout << "--- Header ---" << std::endl;
  for (unsigned int i = 0, size = m_header.size(); i < size; ++i)
    m_header[i]->print ();
  std::cout << std::endl;

  // print global profile
  m_global->print ();

  // print remaining profiles
  for (unsigned int i = 0, size = m_profiles.size(); i < size; ++i)
    m_profiles[i]->print ();
}
