/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * ruleset.h
 * 2005-09-07
 */

#ifndef __RULESET_H__
#define __RULESET_H__

#include "profile.h"

#include <string>
#include <vector>

// forward declarations
class Code;

/**
 * Set of rules.
 */
class RuleSet
{
  public:
    /**
     * Create new ruleset.
     * @param source Pattern source.
     * @param target Pattern target.
     * @param source_type Source object type (either class or struct).
     * @param target_type Target object type (either class or struct).
     */
    RuleSet (const std::string &source, const std::string &target, bool source_type, bool target_type);

    /**
     * Dispose ruleset.
     */
    ~RuleSet (void);

  public:
    /**
     * Get profile with given name.
     * @param name Profile name. Leave empty to access global profile.
     * @return Wanted profile or NULL if not existing.
     */
    Profile *profile (const std::string &name = "") const;

    /**
     * Generate tree-pattern matcher code.
     * @param prefix Output file prefix.
     */
    void create (const std::string &prefix) const;

    /**
     * Print ruleset.
     */
    void print (void) const;

  public:
    /**
     * Add header code.
     * @param header Header code.
     */
    inline void addHeader (Code *header) { m_header.push_back (header); }

    /**
     * Add footer code.
     * @param footer Header code.
     */
    inline void addFooter (Code *footer) { m_footer.push_back (footer); }

    /**
     * Insert given profile.
     * @param profile The profile to insert.
     */
    inline void addProfile (Profile *profile) { m_profiles.push_back (profile); }

    /**
     * Set the copy object function.
     * @param copy The copy object function.
     */
    inline void setCopy (Code *copy) { m_copy = copy; }

    /**
     * Set the implicit side condition.
     * @param condition The used implicit side condition.
     */
    inline void setImplicitCondition (Code *condition) { m_implicit_condition = condition; }

    /**
     * Set the implicit postpass function.
     * @param postpass The used implicit postpass function.
     */
    inline void setImplicitPostPass (Code *postpass) { m_implicit_postpass = postpass; }

  public:
    /**
     * Get pattern source.
     * @return Pattern source.
     */
    inline const std::string &source (void) const { return m_source; }

    /**
     * Get pattern target.
     * @return Pattern target.
     */
    inline const std::string &target (void) const { return m_target; }

    /**
     * Get type depending on given flag.
     * @param source True if source type is desired.
     * @return Either source or target type.
     */
    inline const std::string &type (bool source) const { return source?m_source:m_target; }

    /**
     * Get pattern source type.
     * @return Pattern source type.
     */
    inline bool sourceType (void) const { return m_source_type; }

    /**
     * Get pattern target type.
     * @return Pattern target type.
     */
    inline bool targetType (void) const { return m_target_type; }

    /**
     * Check whether ruleset can be applied recursively.
     * @return True if so.
     */
    inline bool recursive (void) const { return m_recursive; }

    /**
     * Access the copy object function.
     * @return Copy object function.
     */
    inline const Code *copy (void) const { return m_copy; }

    /**
     * Access the implicit side condition.
     * @return Implicit side condition.
     */
    inline const Code *implicitCondition (void) const { return m_implicit_condition; }

    /**
     * Access the implicit postpass function.
     * @return Implicit postpass function.
     */
    inline const Code *implicitPostpass (void) const { return m_implicit_postpass; }

  private:
    /// C/C++ header
    std::vector<Code *> m_header;

    /// C/C++ footer
    std::vector<Code *> m_footer;

    /// pattern source type
    std::string m_source;

    /// pattern target type
    std::string m_target;

    /// source type (true: class, false: struct)
    bool m_source_type;

    /// target type (true: class, false: struct)
    bool m_target_type;

    /// indicates that ruleset can be applied recursively
    bool m_recursive;

    /// copy object function
    Code *m_copy;

    /// implicit side condition
    Code *m_implicit_condition;

    /// implicit postpass function
    Code *m_implicit_postpass;

    /// global profile
    Profile *m_global;

    /// all other profiles
    std::vector<Profile *> m_profiles;
};

#endif
