/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * scanner.ll
 * 2005-09-01
 */

D  [0-9]
L  [a-zA-Z_]
W  [ \r\t\f]

%{
#include "parser.h"
#include "yacc.hpp"

#include "code.h"
#include "rule.h"
#include "pattern.h"

#include <iostream>
#include <sstream>

#include <cstdio>

// parser instance (from yacc.yy)
extern Parser *parser;

// forward declarations
static int yyinput (void);

// current code section
static Code *code = NULL;

// current string buffer
static std::string buffer;

// the end of a block to ignore
static std::string end;

// balance level
static int balance = 1;

/**
 * Add buffer contents to the code sequence.
 * @param segment Optional code segment to add after the buffer.
 * @param whitespace Whether the segment is just whitespace.
 */
static void addBuffer (char *segment = NULL, bool whitespace = false)
{
  // append buffer contents
  if (buffer.length())
  {
    code->add (Code::Segment(buffer));
    buffer = "";
  }

  // append special code segment
  if (segment)
    code->add (Code::Segment(segment, whitespace));
}

/**
 * Add the local pattern to the code sequence.
 */
static void addLocalPattern (void)
{
  // there is no local pattern
  if (!parser->flag(Parser::access_local))
  {
    delete code;
    throw Parser::Exception ("'$$' of rule '"+parser->rule()->fullname()+"' has no declared type.");
  }

  // do the thing
  addBuffer ();
  code->add (Code::Segment (parser->pattern(), true, false));
}

/**
 * Add the pattern with the given index to the code sequence.
 * @param index The index of the pattern.
 * @param relative True if the access is relative to the current item.
 */
static void addPattern (const std::string &index, bool relative)
{
  // relative pattern accesses are not allowed within functions
  if (relative && parser->flag(Parser::function))
  {
    delete code;
    throw Parser::Exception ("Relative pattern accesses are not allowed within functions.");
  }

  // prevent invalid accesses
  unsigned int i = static_cast<unsigned int>(strtol(index.c_str(), 0, 0));
  if (!i)
  {
    delete code;
    throw Parser::Exception (relative?"Ill-formed relative pattern access: '$-0'.":"Ill-formed pattern acccess: '$0'");
  }

  // adjust pattern index for relative pattern accesses
  const Rule *rule = parser->rule ();
  const SequencePattern *search = rule->search ();
  const Pattern *current = parser->flag(Parser::has_current)?parser->pattern():NULL;
  if (relative)
  {
    // no current pattern, no relative pattern access
    if (!current)
    {
      delete code;
      throw Parser::Exception ("Cannot perform relative pattern access, because there is no current pattern.");
    }

    // compute index of current pattern
    unsigned int j = current->index ();
    if ((current->root() == rule->replace()) && search)
      j += search->size();

    // index out of bounds
    if (i > j)
    {
      delete code;
      throw Parser::Exception ("Cannt perform relative pattern access '$-"+index+"': index out of bounds.");
    }

    // do the thing
    i = j-i;
  }
  else
    i -= 1;

  // search pattern
  const Pattern *pattern = rule->pattern (i, parser->flag(Parser::access_search));
  if (!pattern || (pattern == current))
  {
    delete code;
    throw Parser::Exception ("'$"+index+"' of rule '"+rule->fullname()+"' has no declared type.");
  }

  // check if accessing the pattern is possible and safe
  bool unsafe = false;
  if (current)
  {
    // check if access is allowed and safe
    const SequencePattern *parent = pattern->parent(), *current_parent = current->parent(), *condition_parent = NULL;
    bool unordered = !current_parent->ordered();
    while (parent)
    {
      // we might have a reason for thinking that the access is unsafe
      if (!condition_parent && parent->condition())
      {
        condition_parent = parent;
        unsafe = true;
      }

      // this kind of access is not allowed
      if (unordered && (current_parent == parent))
      {
        delete code;
        throw Parser::Exception ("Cannot access '$"+index+"' of rule '"+rule->fullname()+"' at this point.");
      }

      // switch to the parent's parent
      parent = parent->parent();
    }

    // confirm whether access is unsafe
    while (current_parent)
    {
      // access is safe when any ancestor of the current pattern is the pattern with the side condition
      if (current_parent == condition_parent)
      {
        unsafe = false;
        break;
      }

      // switch to next parent
      current_parent = current_parent->parent();
    }

    // inform user about unsafe access
    if (unsafe)
    {
      Parser *parser = Parser::instance ();
      if (!parser->flag(Parser::unsafe))
      {
        std::stringstream msg ("");
        msg << "Accessing pattern $" << index << " from pattern $" << (current->index()+((current->root() == rule->usedReplace())?rule->usedSearch()->size()+1:1)) << " in rule '" << rule->fullname() << "' is considered unsafe.";
        if (parser->flag(Parser::pedantic))
        {
          delete code;
          throw Parser::Exception (msg.str());
        }
        else parser->warning (msg.str());
      }
    }
  }

  // do the thing
  addBuffer ();
  code->add (Code::Segment (pattern, false, unsafe));
}

/**
 * Finish the code block and hand control over code block to parser.
 * @param target Where to store the code block.
 */
static void finish (Code **target)
{
  // add remaining buffer and normalise the code segment
  addBuffer ();
  code->normalise ();

  // hand control over to target
  *target = code;
  code    = NULL;
}

/**
 * Read a line from yyin until the given pattern is found.
 * @param pattern The pattern to look for.
 */
static std::string readLine (const std::string &pattern)
{
  // read from yyin until pattern is found
  int c;
  std::string last, buffer;
  while ((c = yyinput()) != EOF)
  {
    // line increment
    if (c == '\n')
      ++parser->current()->line();

    // remove first character from local buffer, add it to input buffer
    if (last.length() == pattern.length())
    {
      buffer += last[0];
      last = last.substr(1, last.length()-1);
    }

    // add current character to local buffer and compare with pattern
    last += c;
    if (last == pattern)
      return buffer;
  }

  // pattern was not found
  throw Parser::Exception ("Unexpected end of line. Expected '"+pattern+"'.");
}
%}

%option noyywrap
%x PREPRO PREPRO_SKIP BRANCH DEFINE INCLUDE SKIP CODE IGNORELINE IGNOREBLOCK

%%

"class"            { return CLASS; }
"condition"        { return CONDITION; }
"copy"             { return COPY; }
"cost"             { return COST; }
"else"             { return ELSE; }
"extends"          { return EXTENDS; }
"if"               { return IF; }
"implicit"         { return IMPLICIT; }
"keep"             { return KEEP; }
"omit"             { return OMIT; }
"postpass"         { return POSTPASS; }
"profile"          { return PROFILE; }
"replace"          { return REPLACE; }
"intermediate"     { return INTERMEDIATE; }
"rule"             { return RULE; }
"ruleset"          { return RULESET; }
"search"           { return SEARCH; }
"struct"           { return STRUCT; }

{L}({L}|{D})*      { return yylval._str = strdup(yytext), IDENTIFIER; }
{D}+               { return yylval._int = strtol(yytext, 0, 0), INT_CONSTANT; }

"//"               { BEGIN (IGNORELINE); }
"/*"               {
                     BEGIN (IGNOREBLOCK);
                     end = "*/";
                   }

"%{"               {
                     BEGIN (CODE);
                     code = new Code (parser->current()->filename(), parser->current()->line());
                   }
"{"                {
                     // parser wants a function
                     if (parser->flag(Parser::function))
                     {
                       BEGIN (CODE);
                       code  = new Code (parser->current()->filename(), parser->current()->line());
                       balance = 1;
                     }

                     // do as expected
                     else
                       return '{';
                   }
"("                {
                     // parser wants an expression
                     if (parser->flag(Parser::expression))
                     {
                       BEGIN (CODE);
                       code  = new Code (parser->current()->filename(), parser->current()->line());
                       balance = 1;
                     }

                     // do as expected
                     else
                       return '(';
                   }
"%%"               {
                     // parser wants a footer section
                     if (parser->flag(Parser::footer))
                     {
                       BEGIN (CODE);
                       code  = new Code (parser->current()->filename(), parser->current()->line());
                     }

                     // do as expected
                     else
                       return SEPARATOR;
                   }

"::"               { return ACCESS; }

"*"                { return '*'; }
"."                { return '.'; }
":"                { return ':'; }
","                { return ','; }
";"                { return ';'; }
"}"                { return '}'; }
")"                { return ')'; }
"["                { return '['; }
"]"                { return ']'; }
"<"                { return '<'; }
">"                { return '>'; }
"$"                { return '$'; }
"#"                { BEGIN (PREPRO); }

<IGNORELINE>
{
  \n               {
                     // done ignoring
                     BEGIN (code?CODE:INITIAL);

                     // increment line count and consume item
                     ++parser->current()->line();
                     if (code)
                       addBuffer (yytext);
                   }
}

<CODE,IGNOREBLOCK>
{
  "\\\""           { if (code) buffer += yytext; }
  "\\'"            { if (code) buffer += yytext; }
  \n               {
                     // increment line count and consume item
                     ++parser->current()->line();
                     if (code)
                       addBuffer (yytext);
                   }
}

<IGNOREBLOCK>
{
  "*/"             {
                     // done ignoring
                     if (end == yytext)
                       BEGIN (code?CODE:INITIAL);

                     // consume item
                     if (code)
                       buffer += yytext;
                   }
  "\""             {
                     // done ignoring
                     if (end == yytext)
                       BEGIN (CODE);

                     // consume item
                     if (code)
                       buffer += yytext;
                   }
  "'"              {
                     // done ignoring
                     if (end == yytext)
                       BEGIN (CODE);

                     // consume item
                     if (code)
                       buffer += yytext;
                   }
}

<INITIAL,CODE>
{
  "//"             {
                     BEGIN(IGNORELINE);
                     if (code)
                       buffer += yytext;
                   }
  "/*"             {
                     BEGIN(IGNOREBLOCK);
                     if (code)
                       buffer += yytext;
                     end = "*/";
                   }
}

<CODE>
{
  "\""             {
                     BEGIN(IGNOREBLOCK);
                     buffer += yytext;
                     end = "\"";
                   }

  "{"              {
                     // increase balance level if function is expected
                     if (parser->flag(Parser::function))
                       ++balance;

                     // consume item
                     addBuffer (yytext);
                   }
  "}"              {
                     // decrease balance level if function is expected
                     if (parser->flag(Parser::function) && !--balance)
                     {
                       BEGIN (INITIAL);
                       finish (&yylval._code);
                       return FUNCTION;
                     }

                     // consume item
                     addBuffer (yytext);
                   }
  "("              {
                     // increase balance level if expression is expected
                     if (parser->flag(Parser::expression))
                       ++balance;

                     // consume item
                     buffer += yytext;
                   }
  ")"              {
                     // decrease balance level if expression is expected
                     if (parser->flag(Parser::expression) && !--balance)
                     {
                       BEGIN (INITIAL);
                       finish (&yylval._code);
                       return EXPRESSION;
                     }

                     // consume item
                     buffer += yytext;
                   }
  "%}"             {
                     // finish header
                     if (!parser->flag(static_cast<Parser::Flag>(Parser::footer | Parser::function | Parser::expression)))
                     {
                       BEGIN (INITIAL);
                       finish (&yylval._code);
                       return HEADER;
                     }

                     // consume item
                     buffer += yytext;
                   }
  "%%"             {
                     // finish footer
                     if (parser->flag(Parser::footer))
                     {
                       BEGIN (INITIAL);
                       finish (&yylval._code);
                       return FOOTER;
                     }

                     // consume item
                     buffer += yytext;
                   }

  "$$"             {
                     // access local pattern
                     if (parser->flag(Parser::access))
                       addLocalPattern ();

                     // consume item
                     else
                       buffer += yytext;
                   }
  "$"{D}+          {
                     // absolute pattern access
                     if (parser->flag(Parser::access))
                     {
                       std::string number = yytext;
                       addPattern (number.substr(1, number.length()-1), false);
                     }

                     // consume item
                     else
                       buffer += yytext;
                   }
  "$-"{D}+         {
                     // relative pattern access
                     if (parser->flag(Parser::access))
                     {
                       std::string number = yytext;
                       addPattern (number.substr(2, number.length()-1), true);
                     }

                     // consume item
                     else
                       buffer += yytext;
                   }
}

<CODE,IGNORELINE,IGNOREBLOCK>
{
  {W}+             { if (code) addBuffer (yytext, true); }
  .                { if (code) buffer += yytext; }
}

<*>{W}+            { /* eat up whitespace */ }

<PREPRO>
{
  \n               { throw Parser::Exception ("Expected preprocessor statement."); }
  "define"         { parser->setFlag (Parser::def, true); BEGIN (DEFINE); }
  "undef"          { parser->setFlag (Parser::def, false); BEGIN (DEFINE); }
  "include"        { BEGIN (INCLUDE); }
}

<PREPRO,PREPRO_SKIP>
{
  "ifdef"          { parser->setFlag (Parser::ifdef, true); BEGIN (BRANCH); }
  "ifndef"         { parser->setFlag (Parser::ifdef, false); BEGIN (BRANCH); }
  "else"           { BEGIN (parser->beginElse()?SKIP:INITIAL); }
  "endif"          { BEGIN (parser->endIf()?SKIP:INITIAL); }
}

<PREPRO_SKIP>
{
  .                { BEGIN (SKIP); }
}

<SKIP>
{
  "#"              { BEGIN (PREPRO_SKIP); }
  [^\n]            { /* eat up everything except newlines */ }
}

<BRANCH>
{
  {L}({L}|{D})*    { BEGIN (parser->beginIf(parser->defined(yytext) ^ parser->flag(Parser::ifdef))?SKIP:INITIAL); }
}

<DEFINE>
{
  {L}({L}|{D})*    {
                     if (parser->flag(Parser::def))
                       parser->define (yytext);
                     else
                       parser->undefine (yytext);
                     BEGIN (INITIAL);
                   }
}

<BRANCH,DEFINE>
{
  \n               { throw Parser::Exception ("Expected identifer."); }
}

<INCLUDE>
{
  "\""             {
                     BEGIN (INITIAL);
                     parser->include (readLine ("\""), true);
                   }
  "<"              {
                     BEGIN (INITIAL);
                     parser->include (readLine (">"), false);
                   }
  .                { throw Parser::Exception ("Expected include path."); }
}

<*>\n              { ++parser->current()->line(); }

<*>.               {
                     std::stringstream msg ("");
                     msg << "Unexpected character '" << (char)yytext[0] << "' (code = 0x" << std::hex << (int)yytext[0] << std::dec << ").";
                     throw Parser::Exception (msg.str());
                   }

<<EOF>>            {
                     // clean up
                     if (code)
                       delete code;

                     // unmatched ifs
                     if (parser->ifs())
                       throw Parser::Exception ("Unmatched #if.");

                     // finish current file
                     if (!parser->finish())
                       yyterminate ();
                   }

%%

#ifdef _WIN32
//int yywrap (void) { return 1; }
#  ifdef __cplusplus
int isatty (int) { return 0; }
#  else
int isatty (int a) { return 0; }
#  endif
#endif
