/* Object-Oriented Rewrite System Generator (oorg)
   Copyright (C) 2006  Gernot Gebhard

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/**
 * yacc.yy
 * 2005-09-01
 */

%{
#include "code.h"
#include "rule.h"
#include "profile.h"
#include "ruleset.h"
#include "pattern.h"
#include "parser.h"

#include <string>
#include <sstream>
#include <iostream>

// flex/bison forward declarations
int yylex (void);
int yyerror (const char *);

// parser instance
Parser *parser = NULL;
%}

%union
{
  int _int;
  bool _bool;
  char *_str;
  Rule *_rule;
  Code *_code;
};

// start symbol
%start file

// TOKENS

%token CLASS CONDITION COPY COST ELSE EXTENDS IF IMPLICIT KEEP OMIT POSTPASS PROFILE RULE RULESET REPLACE INTERMEDIATE SEARCH STRUCT
%token EXPRESSION FUNCTION HEADER FOOTER INT_CONSTANT
%token IDENTIFIER
%token ACCESS SEPARATOR

// TOKEN TYPES

// constants
%type <_int> INT_CONSTANT

// strings
%type <_str> IDENTIFIER

// code
%type <_code> EXPRESSION
%type <_code> FUNCTION
%type <_code> HEADER
%type <_code> FOOTER

// RULE TYPES

// flags
%type <_bool> ruleset_type;
%type <_bool> pattern_remove;

// strings
%type <_str> rule_name

// rules
%type <_rule> rule_extend
%type <_rule> rule_access

// code
%type <_code> expression
%type <_code> implicit_condition;
%type <_code> implicit_copy;
%type <_code> implicit_postpass;
%type <_code> ruleset_header
%type <_code> ruleset_footer
%type <_code> rule_mask

%%

// BEGIN FILE STUFF

/**
 * Start symbol, which represents the whole file.
 */
file
  : { parser = Parser::instance (); }
    file_contents
  ;

/**
 * What an oors file may consist of: rulesets.
 */
file_contents
  : ruleset
  | file_contents ruleset
  ;

// END FILE STUFF

// BEGIN RULE STUFF

/**
 * Ruleset which consists of several rules.
 */
ruleset
  : ruleset_header RULESET '<' ruleset_type IDENTIFIER ',' ruleset_type IDENTIFIER '>'
    {
      // copy names
      std::string source = $5, target = $8;
      bool source_type = $4, target_type = $7;
      free ($5);
      free ($8);

      // check types
      if ((source == target) && (source_type != target_type))
      {
        std::stringstream msg ("");
        msg << "Given type '" << source << "' cannot be class and struct at the same time.";
        throw Parser::Exception (msg.str());
      }

      // check ruleset
      RuleSet *ruleset = parser->ruleset ();
      if (ruleset)
      {
        // mismatching source types
        std::stringstream msg ("");
        if ((ruleset->source() != source) || (ruleset->sourceType() != source_type))
        {
          msg << "Invalid source type. Expected '" << (ruleset->sourceType()?"class ":"struct ") << ruleset->source() << "'.";
          throw Parser::Exception (msg.str());
        }

        // mismatching target types
        if ((ruleset->target() != target) || (ruleset->targetType() != target_type))
        {
          msg << "Invalid target type. Expected '" << (ruleset->targetType()?"class ":"struct ") << ruleset->target() << "'.";
          throw Parser::Exception (msg.str());
        }
      }

      // create first ruleset
      else
      {
        parser->ruleset() = ruleset = new RuleSet (source, target, source_type, target_type);
        parser->profile() = ruleset->profile ();
      }

      // add ruleset header
      if ($1)
        ruleset->addHeader ($1);
    }
    '{' ruleset_implicit rules '}'
    { parser->setFlag (Parser::footer, true); }
    ruleset_footer
    {
      // finish footer
      parser->setFlag (Parser::footer, false);

      // insert footer
      if ($16)
        parser->ruleset()->addFooter ($16);
    }
  ;

/**
 * The ruleset header contains C/C++ code.
 */
ruleset_header
  : HEADER
  | /* empty */
   { $$ = NULL; }
  ;

/**
 * The ruleset footer contains C/C++ code.
 */
ruleset_footer
  : FOOTER
  | /* empty */
   { $$ = NULL; }
  ;

/**
 * The type of the matched items (either class or struct).
 */
ruleset_type
  : CLASS
    { $$ = true; }
  | STRUCT
    { $$ = false; }
  | /* default: class */
    { $$ = true; }
  ;

/**
 * Implicit stuff.
 */
ruleset_implicit
  : IMPLICIT '{' ruleset_implicit_functions '}'
  | /* empty */
  ;

/**
 * A selection of implicit functions
 */
ruleset_implicit_functions
  : ruleset_implicit_condition
  | ruleset_implicit_condition ruleset_implicit_copy
  | ruleset_implicit_condition ruleset_implicit_postpass
  | ruleset_implicit_condition ruleset_implicit_copy ruleset_implicit_postpass
  | ruleset_implicit_copy
  | ruleset_implicit_copy ruleset_implicit_postpass
  | ruleset_implicit_postpass
  ;

/**
 * Copy object stuff.
 */
ruleset_implicit_copy
  : implicit_copy
    {
      // check whether the ruleset also has a copy function
      RuleSet *ruleset = parser->ruleset ();
      if (ruleset->copy())
      {
        parser->warning ("Ignoring copy function because the current ruleset already has one.");
        delete $1;
      }

      // parse and set implicit side condition
      else
        ruleset->setCopy ($1);
    }
  ;

/**
 * Implicit side condition that is called whenever a wildcard pattern is contained in a search pattern.
 */
ruleset_implicit_condition
  : implicit_condition
    {
      // check whether the ruleset already has a condition function
      RuleSet *ruleset = parser->ruleset ();
      if (ruleset->implicitCondition())
      {
        parser->warning ("Ignoring implicit side condition because the current ruleset already has one.");
        delete $1;
      }

      // parse and set implicit side condition
      else
        ruleset->setImplicitCondition ($1);
    }
  ;

/**
 * Implicit postpass function that is called whenever a rule has been applied.
 */
ruleset_implicit_postpass
  : implicit_postpass
    {
      // check whether the rulesetr already has an implicit postpass function
      RuleSet *ruleset = parser->ruleset ();
      if (ruleset->implicitPostpass())
      {
        parser->warning ("Ignoring implicit postpass function because the current ruleset already has one.");
        delete $1;
      }

      // parse and set implicit side condition
      else
        ruleset->setImplicitPostPass ($1);
    }
  ;

/**
 * Multiple rules.
 */
rules
  : rule
  | omit
  | profile
  | rules rule
  | rules omit
  | rules profile
  ;

/**
 * Profile.
 */
profile
  : profile_header '{' profile_body '}'
    { parser->profile() = parser->ruleset()->profile(); }
  | profile_header ';'
    { parser->profile() = parser->ruleset()->profile(); }
  ;

/**
 * Profile header.
 */
profile_header
  : PROFILE IDENTIFIER
    {
      // copy profile name
      std::string name = $2;
      free ($2);

      // check if profile already exists
      RuleSet *ruleset = parser->ruleset ();
      Profile *profile = ruleset->profile (name);
      if (profile)
        throw Parser::Exception ("The profile '"+name+"' already exists.");

      // create profile, extend global profile and set as active profile
      profile = new Profile (ruleset, name);
      profile->extend (parser->profile());
      ruleset->addProfile (parser->profile() = profile);
    }
    profile_extend
  ;

/**
 * Extended profiles.
 */
profile_extend
  : ':' EXTENDS profile_extend_list
  | /* empty */
  ;

/**
 * A list of extended profiles.
 */
profile_extend_list
  : profile_extend_list_item
  | profile_extend_list ',' profile_extend_list_item
  ;

/**
 * A single extended profile.
 */
profile_extend_list_item
  : IDENTIFIER
    {
      // copy profile name
      std::string name = $1;
      free ($1);

      // get profile
      Profile *profile = parser->ruleset()->profile (name);
      if (!profile || (profile == parser->profile()))
        throw Parser::Exception ("Cannot extend unknown profile '"+name+"'.");

      // do the thing
      if (!parser->profile()->extend(profile))
        parser->warning ("The profile '"+name+"' has already been extended.");
    }
  ;

/**
 * The profile body.
 * It might be empty, or not.
 */
profile_body
  : profile_rules
  | /* empty */
  ;

/**
 * Rules inside a profile.
 */
profile_rules
  : rule
  | omit
  | profile_rules rule
  | profile_rules omit
  ;

/**
 * A rule is specified through its header and body.
 */
rule
  : RULE
    { parser->setFlag (Parser::expression, true); }
    rule_header
    { parser->setFlag (Parser::expression, false); }
    '{' rule_body rule_implicit '}'
    {
      // verify and insert the created rule
      Rule* &rule = parser->rule ();
      rule->verify ();
      parser->profile()->insert (rule);
      rule = NULL;
    }
  ;

/**
 * The rule header determines the name of the rule and if it is a derived or profile
 * version of an existing rule.
 */
rule_header
  : rule_name
    {
      parser->rule() = new Rule (parser->profile(), $1, NULL, NULL);
      free ($1);
    }
  | rule_name rule_extend
    {
      parser->rule() = new Rule (parser->profile(), $1, $2, NULL);
      free ($1);
    }
  | rule_name rule_mask
    {
      parser->rule() = new Rule (parser->profile(), $1, NULL, $2);
      free ($1);
    }
  | rule_name rule_extend rule_mask
    {
      parser->rule() = new Rule (parser->profile(), $1, $2, $3);
      free ($1);
    }
  ;

/**
 * Check whezher a rule's name exists.
 */
rule_name
  : IDENTIFIER
    {
      // bail out if rule already exists
      Profile *profile = parser->profile ();
      if (profile->rule($1))
      {
        std::string name = $1;
        free ($1);
        throw Parser::Exception ("The rule '"+name+"' already exists"+(profile->global()?".":" in profile '"+profile->name()+"'."));
      }

      // all fine
      $$ = $1;
    }
  ;

/**
 * Rules can be derived from others but must not.
 */
rule_extend
  : ':' EXTENDS rule_access
    { $$ = $3; }
  ;

/**
 * Access a rule.
 */
rule_access
  : IDENTIFIER ACCESS IDENTIFIER
    {
      // copy names
      std::string profile = $1, name = $3;
      free ($1);
      free ($3);

      // get profile
      Profile *p = parser->ruleset()->profile (profile);
      if (!p)
        throw Parser::Exception ("Unknown profile: '"+profile+"'.");

      // get rule
      if (!($$ = p->rule(name)))
        throw Parser::Exception ("Unknown rule: '"+profile+"::"+name+"'.");
    }
  | ACCESS IDENTIFIER
    {
      // copy rule name
      std::string name = $2;
      free ($2);

      // get rule
      if (!($$ = parser->ruleset()->profile()->rule(name)))
        throw Parser::Exception ("Unknown rule: '::"+name+"'.");
    }
  | IDENTIFIER
    {
      // copy rule name
      std::string name = $1;
      free ($1);

      // get current and global profile
      Profile *profile = parser->profile(), *global = parser->ruleset()->profile();
      Rule *rule = profile->rule (name);
      if (!rule && (profile != global))
        rule = global->rule (name);

      // unknown rule
      if (!($$ = rule))
        throw Parser::Exception ("Unknown rule: '"+name+"'.");
    }
  ;

/**
 * The rule's flag (if any).
 */
rule_mask
  : ':' EXPRESSION
    {
      $$ = $2;
    }
  ;

/**
 * The rule body consists of four rule specific parts: search pattern, condition function, cost function
 * replace pattern. All of them are optional, which allows rules to be pure abstract.
 * Such rules cannot be matched and act simply as templates.
 */
rule_body
  : search condition cost replace
  ;

/**
 * The pattern that should be looked for.
 */
search
  : SEARCH ':'
    {
      parser->setFlag (Parser::search, true);
      parser->setFlag (Parser::intermediate, false);
    }
    pattern
    {
      // empty search rules are not allowed
      if (!parser->rule()->search()->effect())
        throw Parser::Exception ("A search pattern must match at least one item.");
    }
  | /* empty */
  ;

/**
 * The condition that must be fulfilled before the rule can be applied.
 */
condition
  : CONDITION ':'
    {
      parser->setFlag (Parser::function, true);
      parser->setFlag (Parser::access, true);
      parser->setFlag (Parser::access_local, false);
      parser->setFlag (Parser::access_search, true);
      parser->setFlag (Parser::has_current, false);
    }
    FUNCTION
    {
      parser->setFlag (Parser::function, false);
      parser->setFlag (Parser::access, false);
      parser->rule()->setCondition ($4);
    }
  | /* empty */
  ;

/**
 * The rule's cost function.
 */
cost
  : COST ':'
    {
      parser->setFlag (Parser::function, true);
      parser->setFlag (Parser::access, true);
      parser->setFlag (Parser::access_local, false);
      parser->setFlag (Parser::access_search, true);
      parser->setFlag (Parser::has_current, false);
    }
    FUNCTION
    {
      parser->setFlag (Parser::function, false);
      parser->setFlag (Parser::access, false);
      parser->rule()->setCost ($4);
    }
  | /* empty */
  ;

/**
 * The replace pattern.
 */
replace
  : REPLACE ':'
    {
      parser->setFlag (Parser::search, false);
      parser->setFlag (Parser::intermediate, false);
    }
    pattern
  | INTERMEDIATE ':'
    {
      parser->setFlag (Parser::search, false);
      parser->setFlag (Parser::intermediate, true);
    }
    pattern
  | /* empty */
  ;

/**
 * Implicit stuff.
 */
rule_implicit
  : IMPLICIT
    {
      parser->setFlag (Parser::access, true);
      parser->setFlag (Parser::access_local, false);
      parser->setFlag (Parser::access_search, true);
      parser->setFlag (Parser::has_current, false);
    }
    '{' rule_implicit_functions '}'
    {
      parser->setFlag (Parser::access, false);
    }
  | /* empty */
  ;

/**
 * A selection of implicit functions.
 */
rule_implicit_functions
  : rule_implicit_condition
  | rule_implicit_condition ruleset_implicit_postpass
  | rule_implicit_postpass
  ;

/**
 * Implicit condition function.
 */
rule_implicit_condition
  : implicit_condition { parser->rule()->setImplicitCondition ($1); }
  ;

/**
 * Implicit postpass function.
 */
rule_implicit_postpass
  : implicit_postpass { parser->rule()->setImplicitPostPass ($1); }
  ;

// END RULE STUFF

// BEGIN IMPLICIT FUNCTIONS STUFF

/**
 * Implicit condition function.
 */
implicit_condition
  : CONDITION ':'
    {
      parser->setFlag (Parser::function, true);
    }
    FUNCTION
    {
      parser->setFlag (Parser::function, false);
      $$ = $4;
    }
  ;

/**
 * Implicit copy function.
 */
implicit_copy
  : COPY ':'
    {
      parser->setFlag (Parser::function, true);
    }
    FUNCTION
    {
      parser->setFlag (Parser::function, false);
      $$ = $4;
    }
  ;

/**
 * Implicit postpass function.
 */
implicit_postpass
  : POSTPASS ':'
    {
      parser->setFlag (Parser::function, true);
    }
    FUNCTION
    {
      parser->setFlag (Parser::function, false);
      $$ = $4;
    }
  ;

// END IMPLICIT FUNCTIONS STUFF

// BEGIN PATTERN STUFF

/**
 * A pattern is either a sequence of items or a set of pattern.
 */
pattern
  : pattern_sequence
  | pattern_set
  ;

/**
 * Pattern sequence.
 */
pattern_sequence
  : '['
    {
      parser->pushSequence (new SequencePattern (parser->current()->filename(), parser->current()->line(), parser->rule(), parser->sequence(), true, parser->flag(Parser::search), parser->flag(Parser::intermediate)));
    }
    pattern_list_empty ']'
    {
      // normalise the sequence pattern
      SequencePattern *sequence = parser->sequence ();
      sequence->normalise ();

      // remember the sequence and pop it from stack
      parser->pattern() = sequence;
      parser->popSequence ();
    }
  ;

/**
 * Pattern set.
 */
pattern_set
  : '{'
    {
      // a replace pattern must be an ordered sequence
      if (!parser->flag(Parser::search))
        throw Parser::Exception ("A replace pattern must be an ordered sequence.");

      // create pattern and set it as active one
      parser->pushSequence (new SequencePattern (parser->current()->filename(), parser->current()->line(), parser->rule(), parser->sequence(), false, true, false));
    }
    pattern_list_empty '}'
    {
      // normalise the sequence pattern and pop it from stack
      parser->sequence()->normalise ();
      parser->popSequence ();
    }
  ;

/**
 * Either an empty or non-empty list of patterns.
 */
pattern_list_empty
  : pattern_list
  | /* empty */
  ;

/**
 * A non-empty list of patterns.
 */
pattern_list
  : pattern_list_item
  | pattern_list ',' pattern_list_item
  ;

/**
 * A pattern item can either be a pattern list, a wildcard list, wildcard item or an object.
 * Additionally it may be empty, allowing empty rules.
 */
pattern_list_item
  : pattern_sequence_if
  | pattern_set
  | '*'
    {
      // wildcards are not allows within replace patterns
      if (!parser->flag(Parser::search))
        throw Parser::Exception ("Wildcard patterns are not allowed within replace patterns.");

      // create wildcard pattern
      new WildcardPattern (parser->current()->filename(), parser->current()->line(), parser->rule(), parser->sequence());
    }
  | pattern_initialisable initialisers
  ;

/**
 * Sequence pattern may be tagged with a condition but only when being used in a replace pattern.
 */
pattern_sequence_if
  : IF
    {
      // must not use conditions in search patterns
      if (parser->flag(Parser::search))
        throw Parser::Exception ("Conditions are not allowed within search patterns.");

      // expect an expression
      parser->setFlag (Parser::expression, true);
      parser->setFlag (Parser::access, true);
      parser->setFlag (Parser::access_local, false);
      parser->setFlag (Parser::access_search, false);
      parser->setFlag (Parser::has_current, false);
    }
    EXPRESSION pattern_sequence
    {
      // clear flags
      parser->setFlag (Parser::expression, false);
      parser->setFlag (Parser::access, false);

      // assign condition
      static_cast<SequencePattern *>(parser->pattern())->setCondition ($3, parser->flag(Parser::else_branch));
      parser->setFlag (Parser::else_branch, false);
    }
    pattern_sequence_else
  | pattern_sequence
    {
      // set else branch flag
      static_cast<SequencePattern *>(parser->pattern())->setCondition (NULL, parser->flag(Parser::else_branch));
      parser->setFlag (Parser::else_branch, false);
    }
  ;

/**
 * Pattern sequence else.
 */
pattern_sequence_else
  : ELSE
    { parser->setFlag (Parser::else_branch, true); }
    pattern_sequence_if
  | /* empty */
  ;

/**
 * All non-wildcard and non-sequence patterns can be initialised.
 */
pattern_initialisable
  : pattern_conditional
  | '$' INT_CONSTANT
    {
      // conservation patterns cannot be used within search patterns
      if (parser->flag(Parser::search))
        throw Parser::Exception ("Conservation patterns are not allowed within search patterns.");

      // preparations
      Rule *rule = parser->rule ();
      const Pattern *access = rule->pattern ($2-1, true);
      if (!access)
      {
        std::stringstream msg ("");
        msg << "'$" << $2 << "' of rule '" << rule->fullname() << "' has no declared type.";
        throw Parser::Exception (msg.str());
      }
      const ItemPattern *item = dynamic_cast<const ItemPattern *>(access);
      if (!item)
      {
        std::stringstream msg ("");
        msg << "'$" << $2 << "' of rule '" << rule->fullname() << "' is not an item pattern and thus cannot be conservated.";
        throw Parser::Exception (msg.str());
      }

      // if the consvered item is to be used in an intermediate replace pattern its survival cannot be guaranteed
      if (parser->flag(Parser::intermediate) && !item->remove())
      {
        // if we are pedantic bail out
        std::stringstream msg ("");
        msg << "'$" << $2 << "' of rule '" << rule->fullname() << "' might be deleted by subsequent matches.";
        if (parser->flag(Parser::pedantic))
          throw Parser::Exception (msg.str());

        // otherwise just warn
        else
          parser->warning (msg.str());
      }

      // create and remember conservation pattern
      parser->pattern() = new ConservationPattern (parser->current()->filename(), parser->current()->line(), parser->rule(), parser->sequence(), item);
    }
  ;

/**
 * Conditional item patterns.
 */
pattern_conditional
  : {
      parser->setFlag (Parser::expression, true);
      parser->setFlag (Parser::access, true);
      parser->setFlag (Parser::access_local, parser->flag (Parser::search));
      parser->setFlag (Parser::access_search, parser->flag (Parser::search));
      parser->setFlag (Parser::has_current, true);
    }
    pattern_remove pattern_item expression
    {
      // clear flags
      parser->setFlag (Parser::expression, false);
      parser->setFlag (Parser::access, false);

      // set condition/replace code
      ItemPattern *pattern = static_cast<ItemPattern *>(parser->pattern());
      if ($4)
        pattern->setCode ($4);

      // set remove flag
      pattern->setRemove ($2);
    }
  ;

/**
 * Check whether the matched item may be deleted after applying the rule.
 */
pattern_remove
  : KEEP ':'
    {
      if (!parser->flag(Parser::search))
      {
        parser->warning ("Item patterns within a replace pattern cannot be kept. Ignoring keep token.");
        $$ = true;
      }
      else if (!parser->ruleset()->recursive())
      {
        parser->warning ("The ruleset cannot be applied recursively. Ignoring keep token.");
        $$ = true;
      }
      else
        $$ = false;
    }
  | /* empty */
    { $$ = true; }
  ;

/**
 * There are two different types of item patterns.
 */
pattern_item
  : '.'
    {
      // create new base pattern item
      const std::string &type = parser->ruleset()->type(parser->flag(Parser::search));
      parser->pattern() = new ItemPattern (parser->current()->filename(), parser->current()->line(), parser->rule(), parser->sequence(), type, type);
    }
  | IDENTIFIER
    {
      // create new item pattern
      parser->pattern() = new ItemPattern (parser->current()->filename(), parser->current()->line(), parser->rule(), parser->sequence(), parser->ruleset()->type(parser->flag(Parser::search) || parser->flag(Parser::intermediate)), $1);
      free ($1);
    }
  ;

/**
 * A code section that may also be empty.
 */
expression
  : EXPRESSION
  | /* empty */
    { $$ = NULL; }
  ;

/**
 * A possible empty set of initialisers.
 * Initialisers are used in replace patterns for initialising the newly created objects.
 */
initialisers
  : ':'
    {
      // initialisers cannot be used in search patterns
      if (parser->flag(Parser::search))
        throw Parser::Exception ("Initialisers may only be used within replace patterns.");

      // expect expressions
      parser->setFlag (Parser::expression, true);
      parser->setFlag (Parser::access, true);
      parser->setFlag (Parser::access_local, true);
      parser->setFlag (Parser::access_search, false);
      parser->setFlag (Parser::has_current, true);
    }
    initialiser_list
    {
      parser->setFlag (Parser::expression, false);
      parser->setFlag (Parser::access, false);
    }
  | /* empty */
  ;

/**
 * A non-empty list of initialisers.
 */
initialiser_list
  : initialiser
  | initialiser_list ':' initialiser
  ;

/**
 * A single initialiser.
 */
initialiser
  : IDENTIFIER EXPRESSION
    {
      Pattern *pattern = parser->pattern ();
      static_cast<InitialisablePattern *>(pattern)->addInitialiser ($1, $2);
      free ($1);
    }
  ;

/**
 * Omit one or more rules.
 */
omit
  : OMIT omit_list ';'
  ;

/**
 * A list of omitted rules.
 */
omit_list
  : omit_list_item
  | omit_list ',' omit_list_item
  ;

/**
 * Omit a single rule.
 */
omit_list_item
  : rule_access
    {
      if (!parser->profile()->omit($1))
        parser->warning ("Omitting rule '"+$1->fullname()+"' at this point makes it unavailable.");
    }
  ;

// END PATTERN STUFF

%%

/**
 * Handle parse error.
 */
int yyerror (const char *)
{
  // unexpected end of file
  if ((yychar == 0) || (yychar == EOF))
    throw Parser::Exception ("Unexpected end of file.");

  // handle syntax error
  bool quote = (yychar != HEADER) && (yychar != FOOTER) && (yychar != FUNCTION) && (yychar != EXPRESSION);
  std::stringstream msg ("");
  msg << "Parse error before ";
  if (quote)
    msg << "'";

  // complete error message
  switch (yychar)
  {
    // code
    case HEADER         : msg << "header"; break;
    case FOOTER         : msg << "footer"; break;
    case FUNCTION       : msg << "function"; break;
    case EXPRESSION     : msg << "expression"; break;

    // keywords
    case CLASS          : msg << "class"; break;
    case CONDITION      : msg << "condition"; break;
    case COPY           : msg << "copy"; break;
    case COST           : msg << "cost"; break;
    case ELSE           : msg << "else"; break;
    case EXTENDS        : msg << "extends"; break;
    case KEEP           : msg << "keep"; break;
    case IF             : msg << "if"; break;
    case IMPLICIT       : msg << "implicit"; break;
    case OMIT           : msg << "omit"; break;
    case POSTPASS       : msg << "postpass"; break;
    case PROFILE        : msg << "profile"; break;
    case REPLACE        : msg << "replace"; break;
    case RULE           : msg << "rule"; break;
    case RULESET        : msg << "ruleset"; break;
    case SEARCH         : msg << "search"; break;
    case STRUCT         : msg << "struct"; break;

    // stored stuff
    case IDENTIFIER     : msg << yylval._str; break;
    case INT_CONSTANT   : msg << yylval._int; break;

    // special tokens
    case ACCESS         : msg << "::"; break;
    case SEPARATOR      : msg << "%%"; break;

    // default stuff
    default             : msg << static_cast<char>(yychar); break;
  }

  // doneSEARCH
  if (quote)
    msg << "'";
  msg << ".";
  throw Parser::Exception (msg.str());
}
